﻿<#
.SYNOPSIS
  Vadlidate an XML file against XSD schema.
.DESCRIPTION
  Check the validity of an XML file against an XSD schema file, throw exception if not valid
.EXAMPLE
  ValidateXmlAgainstXsd .\HPCIaaSClusterConfig.xml .\HPCIaaSClusterConfig.xsd
#>
function ValidateXmlAgainstXsd
{
    Param
    (
        # The xml file path
        [Parameter(Mandatory=$true)]
        [String] $XmlFile,
		
        # The schema file path
        [Parameter(Mandatory=$true)]
        [String] $SchemaFile		
    )
	

    [string[]] $Script:XmlValidationErrorLog = @()
    [scriptblock] $xmlValidationEventHandler = {
        $Script:XmlValidationErrorLog += $args[1].Exception.Message
    }

    $xml = New-Object System.Xml.XmlDocument
    $schemaReader = New-Object System.Xml.XmlTextReader $SchemaFile
    try
    {
        $schema = [System.Xml.Schema.XmlSchema]::Read($schemaReader, $xmlValidationEventHandler)
        $xml.Schemas.Add($schema) | Out-Null
        $xml.Load($XmlFile)
        $xml.Validate($xmlValidationEventHandler)
    }
    catch
    {
        throw
    }
    finally
    {
        $schemaReader.Dispose()
    }

    if($Script:XmlValidationErrorLog.Length -gt 0)
    {
        throw ($Script:XmlValidationErrorLog -join "`r`n")
    }
}

function ValidatePasswordComplexity
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Password
    )

    if($Password.Length -lt 8)
    {
        throw "The password must contain at least 8 characters"
    }

    $chCategories = 0
    if($Password -cmatch "[^0-9a-zA-Z]")
    {
        $chCategories++
    }

    if($Password -cmatch "[0-9]")
    {
        $chCategories++
    }

    if($Password -cmatch "[a-z]")
    {
        $chCategories++
    }

    if($Password -cmatch "[A-Z]")
    {
        $chCategories++
    }

    if($chCategories -lt 3)
    {
        throw "The password must contain 3 of the following: `n a lowercase character `n an uppercase character `n a number `n a special character"
    }
}

function ValidateUserName
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $UserName
    )

    $valid = $false
    do
    {
        if($UserName -match "^admin[1-2]{0,1}$")
        {
            break
        }

        if($UserName -match "^user[1-5]{0,1}$")
        {
            break
        }

        if($UserName -eq "a")
        {
            break
        }

        if($UserName -eq "1")
        {
            break
        }

        if($UserName -match '[\s, \\, /, \[, \], :, |, <, >, +, =, ;, \,, ?, *, @]')
        {
            break
        }

        $valid = $true
    } while ($false)

    if(-not $valid)
    {
        throw "The user name $UserName is not allowed. Please specify another one"
    }
}

function ValidateVNetNaming
{
    Param
    (
        # The VNet Name
        [Parameter(Mandatory=$true)]
        [String] $VNetName,

        # The Subnet Name
        [Parameter(Mandatory=$true)]
        [String] $SubnetName
    )

    $regex = "^[^0-9].{1,62}$"
    if($VNetName -notmatch $regex)
    {
        throw "The virtual network name $VNetName is invalid. It must contain between 2 and 63 characters, and cannot start with a number."
    }
    if($SubnetName -notmatch $regex)
    {
        throw "The subnet Name $SubnetName is invalid. It must contain between 2 and 63 characters, and cannot start with a number."
    }
}

function ValidateServiceNaming
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $ServiceName
    )
    
    $matched = $false
    do 
    {
        $regex = "^[0-9a-zA-Z-]{1,63}$"
        if($ServiceName -cnotmatch $regex)
        {
            break;
        }
        if($ServiceName[0] -eq '-')
        {
            break;
        }
        if($ServiceName[-1] -eq '-')
        {
            break;
        }

        $matched = $true
    } while($false)
    
    if(-not $matched)
    {
        throw "The cloud service name $ServiceName is invalid, It must meet the following criteria: `n 1. It must contain fewer than 63 characters;`n 2. It can contain only letters, numbers, and hyphens, and the first and last character must be a letter or number."
    }
}

function ValidateServiceNamingPattern
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $ServiceNamePattern,

        [Parameter(Mandatory=$true)]
        [int] $Count      
    )

    $matched = $false
    do 
    {
        $regex = "^[0-9a-zA-Z][0-9a-zA-Z-]{0,61}%[0-9]{1,62}%$"
        if($ServiceNamePattern -cnotmatch $regex)
        {
            break
        }
        if($ServiceNamePattern.Length -gt 65)
        {
            break
        }
        $matched = $true
    } while ($false)

    if(-not $matched)
    {
        throw "The cloud service naming pattern $ServiceNamePattern is invalid. It should be <root_name>%<start_number>% and meet the following criteria:`n 1. It must contain between 4 and 65 characters including the percent signs;`n 2. The root_name can contain only letters, numbers, and hyphens, and it must not start with a hyphen; `n 3. The start_number can contain only numbers; `nExample: AzureService%100%"
    }

    $result = ParseNamePattern -NamePattern $ServiceNamePattern
    $availNameCount = [System.Math]::Pow(10, $result[1]) - $result[2]
    if($availNameCount -lt $Count)
    {
        throw "Not enough name space in service naming pattern $ServiceNamePattern, $Count names required, only $availNameCount available"
    }
}

function ValidateStorageAccountNaming
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $StorageAccountName
    )
    
    $regex = "^[0-9a-z]{3,24}$"
    if($StorageAccountName -cnotmatch $regex)
    {
        throw "The storage account name $StorageAccountName is invalid. It must contain between 3 and 24 characters, and can contain only lowercase letters and numbers."
    }
}

function ValidateStorageAccountNamingPattern
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $StorageAccountNamePattern,

        [Parameter(Mandatory=$true)]
        [int] $Count      
    )

    $matched = $false
    do 
    {
        $regex = "^[0-9a-z]{1,23}%[0-9]{1,23}%$"
        if($StorageAccountNamePattern -cnotmatch $regex)
        {
            break
        }
        if(($StorageAccountNamePattern.Length -gt 26) -or ($StorageAccountNamePattern.Length -lt 5))
        {
            break
        }
        $matched = $true
    } while ($false)

    if(-not $matched)
    {
        throw "The storage account naming pattern $StorageAccountNamePattern is invalid. It should be <root_name>%<start_number>% and meet the following criteria:`n 1. It must contain between 5 and 26 characters including the percent signs;`n 2. The root_name can contain only lower-case letters and numbers; `n 3. The start_number can contain only numbers; `nExample: azurestorage%100%"
    }

    $result = ParseNamePattern -NamePattern $StorageAccountNamePattern
    $availNameCount = [System.Math]::Pow(10, $result[1]) - $result[2]
    if($availNameCount -lt $Count)
    {
        throw "Not enough name space in storage account naming pattern $StorageAccountNamePattern, $Count names required, only $availNameCount available"
    }
}

function ValidateAffinityGroupNaming
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $AffinityGroupName
    )
    
    $matched = $false
    do 
    {
        $regex = "^[0-9a-zA-Z][0-9a-zA-Z-]{0,62}$"
        if($AffinityGroupName -cnotmatch $regex)
        {
            break;
        }
        if($AffinityGroupName[-1] -eq '-')
        {
            break;
        }

        $matched = $true
    } while($false)
    
    if(-not $matched)
    {
        throw "The affinity group name $AffinityGroupName is invalid. It must meet the following criteria: `n 1. It must contain fewer than 63 characters;`n 2. It can contain only letters, numbers, and hyphens, and the first and last character must be a letter or number."
    }
}

function ValidateVMNaming
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $VMName
    )

    $regex = "^[a-zA-Z][a-zA-Z0-9-]{1,13}[a-zA-Z0-9]$"
    if($VMName -cnotmatch $regex)
    {
        throw "The VM Name $VMName is invalid. it must meet the following criteria: `n 1. It must contain between 3 and 15 characters; `n 2. It can contain only letters, numbers, and hyphens. `n 3. It must start with a letter and must end with a letter or a number."
    }
}


function ValidateVMPatternNaming
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $VMNamePattern,

        [Parameter(Mandatory=$true)]
        [int] $NodeCount      
    )

    $matched = $false
    do 
    {
        $regex = "^[a-zA-Z][a-zA-Z0-9-]{0,13}%[0-9]{1,14}%$"
        if($VMNamePattern -cnotmatch $regex)
        {
            break
        }
        if(($VMNamePattern.Length -gt 17) -or ($VMNamePattern.Length -lt 5))
        {
            break
        }
        $matched = $true
    } while ($false)

    if(-not $matched)
    {
        throw "The VM naming pattern $VMNamePattern is invalid. It should be <root_name>%<start_number>% and meet the following criteria:`n 1. It must contain between 5 and 17 characters including the percent signs;`n 2. The root_name can contain only letters, numbers, and hyphens, and it must start with a letter `n 3. The start_number can contain only numbers; `nExample: AzureCN%1000%"
    }

    $result = ParseNamePattern -NamePattern $VMNamePattern
    $availNameCount = [System.Math]::Pow(10, $result[1]) - $result[2]
    if($availNameCount -lt $NodeCount)
    {
        throw "Not enough name space in VM naming pattern $VMNamePattern, $NodeCount names required, only $availNameCount available"
    }
}

function ValidateVMSize
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $VMName,

        [Parameter(Mandatory=$true)]
        [String] $VMSize,
        
        [Parameter(Mandatory=$true)]
        [String[]] $AvailableSizes              
    )

    if($AvailableSizes -cnotcontains $VMSize)
    {
        throw "The VM size '$VMSize' of VM $VMName is invalid. Valid sizes are: $([String]::Join(',', $AvailableSizes))"
    }
}


function GetRegionalVNetLocation
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $VNetName
    )

    $uniqueID = [Guid]::NewGuid().ToString()
    $vnetCfgFile = "$env:temp\VNetCfg_$uniqueID.xml"

    InvokeAzureCmdWithRetry -Command "Get-AzureVNetConfig -ExportToFile $vnetCfgFile" | Out-Null
    if(Test-Path -Path $vnetCfgFile)
    {
        $xmlDoc = [xml](Get-Content $vnetCfgFile)
        Remove-Item $vnetCfgFile -Force

        $vnetSite = $xmlDoc.GetElementsByTagName("VirtualNetworkSite") | Where-Object {$_.name -eq $VNetName}
        if($vnetSite -ne $null)
        {
            if($vnetSite.HasAttribute("Location"))
            {
                return $vnetSite.Location
            }
            else
            {
                throw "Virtual network $VNetName is not a regional virtual network." 
            }
        }
    }

    throw "Virtual network $VNetName doesn't exist"
}

function CheckService
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String[]] $ServiceName,

        [Parameter(Mandatory=$true)]
        [String] $VNetName,

        [Parameter(Mandatory=$true)]
        [String] $Location,

        [Parameter(Mandatory=$true)]
        [HashTable] $ServiceAffinityGroupTable,
        
        [Parameter(Mandatory=$true)]
        [String] $DeploymentLabel
    )

    $svcStateTable = @{}
    $errmsgs = @()
    $existingServices = @(InvokeAzureCmdWithRetry -Command "Get-AzureService")
    foreach($svcName in $ServiceName)
    {
        try
        {
            TraceInfo "Checking cloud service $svcName ..."
            $service = $existingServices | ?{$_.ServiceName -eq $svcName}
            if($service -ne $null)
            {
                $agName = $ServiceAffinityGroupTable[$svcName]
                if([String]::IsNullOrEmpty($agName))
                {
                    if($service.AffinityGroup -ne $null)
                    {
                        throw "The cloud service $svcName already exists but in an affinity group $($service.AffinityGroup), not directly in the region $Location."
                    }
                    elseif($service.Location -ne $Location)
                    {
                        throw "The cloud service $svcName already exists but in the region $($service.Location), not in the region $Location."
                    }
                }
                else
                {
                    if($service.AffinityGroup -eq $null)
                    {
                        throw "The cloud service $svcName already exists but not in the affinity group $agName."
                    }
                    if($service.AffinityGroup -ne $agName)
                    {
                        throw "The cloud service $svcName already exists but in affinity group $($service.AffinityGroup), not in affinity group $agName."
                    }
                }

                $deployment = InvokeAzureCmdWithRetry -Command "Get-AzureDeployment -ServiceName $svcName -WarningAction SilentlyContinue"
                if($deployment -ne $null)
                {
                    # Check VNet
                    if($deployment.VNetName -ne $VNetName)
                    {
                        throw "The production deployment of cloud service $ServiceName is not in virtual network $VNetName, but in virtual network $($deployment.VNetName)."
                    }
            
                    if((DecodeLabelIfBase64Encoded $deployment.Label) -ne $DeploymentLabel)
                    {
                         throw "The cloud service $svcName is already in use by other production deployment."
                    }

                    TraceInfo "The cloud service $svcName already exists."
                    $svcStateTable[$svcName] = "Deployed"
                }
                else
                {
                    TraceInfo "The cloud service $svcName already exists."
                    $svcStateTable[$svcName] =  "Created"
                }
            }
            else
            {
                if((InvokeAzureCmdWithRetry -Command "Test-AzureName -Service -Name $svcName") -eq $true)
                {
                    throw "The cloud service name $svcName is not available, it is already in use."
                }

                TraceInfo "Cloud service $svcName doesn't exist"
                $svcStateTable[$svcName] = "NotCreated"
            }
        }
        catch
        {
            $errmsgs += $_.Exception.Message
        }
    }

    if($errmsgs.Count -ne 0)
    {
        throw ($errmsgs -join "`r`n")
    }

    return $svcStateTable
}

function ValidateVMConfig
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName,

        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [String] $SubnetName
    )

    $vmexists = $false
    $vm = InvokeAzureCmdWithRetry -Command "Get-AzureVM -Name $VMName -ServiceName $ServiceName"
    if($vm -ne $null)
    {
        $vmexists = $true
        $vmInSubnet = Get-AzureSubnet -VM $vm -ErrorAction SilentlyContinue
        if($vmInSubnet -ne $SubnetName)
        {
            throw "The virtual machine with same name $VMName already exists in cloud service $ServiceName, but not in Subnet $SubnetName"
        }
        TraceInfo "The virtual machine $VMName already exists"
    }
    else
    {
        TraceInfo "The virtual machine $VMName doesn't exist"
    }

    return $vmexists
}

<#
.SYNOPSIS
  Vadlidate input Cluster Config file
.DESCRIPTION
  Validate whether the input cluster config file is valid, throw exception if not valid
.EXAMPLE
  ConfigFileSemanticsCheck .\HPCIaaSClusterConfig.xml
#>
function ConfigFileSemanticsCheck
{
    Param
    (
        # Prefix name to name resources
        [Parameter(Mandatory=$true)]
        [String] $ConfigFile
    )
	
    if (-not (Test-Path -Path $ConfigFile))
    {
        throw "Config File $ConfigFile does not exist"
    }

    if(-not (Test-Path -Path "$PSScriptRoot\HPCIaaSClusterConfig.xsd"))
    {
        throw "XML Schema File HPCIaaSClusterConfig.xsd does not exist"
    }
    
    $xmlFullPath = (Resolve-Path -Path $ConfigFile).Path

    ## 1. Validate the config file against schema
    ValidateXmlAgainstXsd -XmlFile $xmlFullPath -SchemaFile "$PSScriptRoot\HPCIaaSClusterConfig.xsd"

    $errormsgs = @()
    ## 2. Validate Domain options
    $config = ([Xml] (Get-Content $xmlFullPath)).IaaSClusterConfig
    $configDCOption = $config.Domain.DCOption.Trim()
    if($configDCOption -eq "NewDC")
    {
        if($config.Domain["DomainController"] -eq $null)
        {
            $errormsgs += "'DomainController' must be specified when 'DCOption' is specified as 'NewDC'"
        }
    }
    else
    {
        if($config.Domain["DomainController"] -ne $null)
        {
            $errormsgs += "'DomainController' must not be specified when 'DCOption' is specified as 'ExistingDC' or 'HeadNodeAsDC'"
        }
    }

    ## 3. Validate Database options
    $configDBOption = $config.Database.DBOption.Trim()
    if ($configDBOption -eq "ExistingDB")
    {
        if($configDCOption -ne "ExistingDC")
        {
            $errormsgs += "'ExistingDB' can only be specified when you already have an existing domain forest"
        }
        if($config.Database["DBVersion"] -ne $null)
        {
            $errormsgs += "'DBVersion' must not be specified when 'DBOption' is specified as 'ExistingDB'"
        }
        if($config.Database["DBConnection"] -eq $null)
        {
            $errormsgs += "'DBConnection' must be specified when 'DBOption' is specified as 'ExistingDB'"
        }
        else
        {
            $configDBServerInstance = $config.Database.DBConnection.ServerInstance.Trim()
            if([String]::IsNullOrEmpty($configDBServerInstance))
            {
                $errormsgs += "'ServerInstance' must be specified when 'DBOption' is specified as 'ExistingDB'"
            }
        }
    }
    elseif($configDBOption -eq "NewRemoteDB")
    {
        if($config.Database["DBServer"] -eq $null)
        {
            $errormsgs += "'DBServer' must be specified when 'DBOption' is specified as 'NewRemoteDB'"
        }
        else
        {
            if($config.Database.DBServer["DataDiskSizeInGB"] -ne $null)
            {
                $configRemoteDBDataDiskSize = [int]$config.Database.DBServer.DataDiskSizeInGB.Trim()
                if(($configRemoteDBDataDiskSize -lt 100) -or ($configRemoteDBDataDiskSize -gt 1000))
                {
                    $errormsgs += "The allowed value range of 'DataDiskSizeInGB' is 100 - 1000."
                }
            }
        }

        if($config.Database["DBVersion"] -eq $null)
        {
            $errormsgs += "'DBVersion' must be specified when 'DBOption' is specified as 'NewRemoteDB'"
        }
    }
    elseif($configDBOption -eq "LocalDB")
    {
        if($config.Database["DBVersion"] -ne $null)
        {
            $errormsgs += "'DBVersion' must not be specified when 'DBOption' is specified as 'LocalDB'"
        }
        if($config.Database["DBServer"] -ne $null)
        {
            $errormsgs += "'DBServer' must not be specified when 'DBOption' is specified as 'LocalDB'"
        }
        if($config.Database["DBConnection"] -ne $null)
        {
            $errormsgs += "'DBConnection' must not be specified when 'DBOption' is specified as 'LocalDB'"
        }
    }

    ## 4. Validate the VM Size Combination in Affinity Group
    $AGVMSizeMap = @{}
    $configDefaultAGName = ""
    if($config["AffinityGroup"] -ne $null)
    {
        $configDefaultAGName = $config.AffinityGroup.Trim()
    }
    if($config.Domain["DomainController"] -ne $null)
    {            
        $configDCSize = $config.Domain.DomainController.VMSize.Trim()
        $configDCAGName = $configDefaultAGName
        if($config.Domain.DomainController["AffinityGroup"] -ne $null)
        {
            $configDCAGName = $config.Domain.DomainController.AffinityGroup.Trim()
        }

        if(-not $AGVMSizeMap.ContainsKey($configDCAGName))
        {
            $AGVMSizeMap[$configDCAGName] = @($configDCSize)
        }
        else
        {
            if($AGVMSizeMap[$configDCAGName] -notcontains $configDCSize)
            {
                $AGVMSizeMap[$configDCAGName] += $configDCSize
            }        
        }
    }
    if($config.Database["DBServer"] -ne $null)
    {            
        $configRemoteDBSize = $config.Database.DBServer.VMSize.Trim()
        $configRemoteDBAGName = $configDefaultAGName
        if($config.Database.DBServer["AffinityGroup"] -ne $null)
        {
            $configRemoteDBAGName = $config.Database.DBServer.AffinityGroup.Trim()
        }

        if(-not $AGVMSizeMap.ContainsKey($configRemoteDBAGName))
        {
            $AGVMSizeMap[$configRemoteDBAGName] = @($configRemoteDBSize)
        }
        else
        {
            if($AGVMSizeMap[$configRemoteDBAGName] -notcontains $configRemoteDBSize)
            {
                $AGVMSizeMap[$configRemoteDBAGName] += $configRemoteDBSize
            }      
        }
    }
    if($config["HeadNode"] -ne $null)
    {
        $configHNSize = $config.HeadNode.VMSize.Trim()
        $configHNAGName = $configDefaultAGName
        if($config.HeadNode["AffinityGroup"] -ne $null)
        {
            $configHNAGName = $config.HeadNode.AffinityGroup.Trim()
        }

        if(-not $AGVMSizeMap.ContainsKey($configHNAGName))
        {
            $AGVMSizeMap[$configHNAGName] = @($configHNSize)
        }
        else
        {
            if($AGVMSizeMap[$configHNAGName] -notcontains $configHNSize)
            {
                $AGVMSizeMap[$configHNAGName] += $configHNSize
            }      
        }
    }

    if($config["ComputeNodes"] -ne $null)
    {
        $configCNSize = $config.ComputeNodes.VMSize.Trim()
        $configCNAGName = $configDefaultAGName
        if($config.ComputeNodes["AffinityGroup"] -ne $null)
        {
            $configCNAGName = $config.ComputeNodes.AffinityGroup.Trim()
        }

        if(-not $AGVMSizeMap.ContainsKey($configCNAGName))
        {
            $AGVMSizeMap[$configCNAGName] = @($configCNSize)
        }
        else
        {
            if($AGVMSizeMap[$configCNAGName] -notcontains $configCNSize)
            {
                $AGVMSizeMap[$configCNAGName] += $configCNSize
            }      
        }
    }

    if($config["BrokerNodes"] -ne $null)
    {
        $configBNSize = $config.BrokerNodes.VMSize.Trim()
        $configBNAGName = $configDefaultAGName
        if($config.BrokerNodes["AffinityGroup"] -ne $null)
        {
            $configBNAGName = $config.BrokerNodes.AffinityGroup.Trim()
        }

        if(-not $AGVMSizeMap.ContainsKey($configBNAGName))
        {
            $AGVMSizeMap[$configBNAGName] = @($configBNSize)
        }
        else
        {
            if($AGVMSizeMap[$configBNAGName] -notcontains $configBNSize)
            {
                $AGVMSizeMap[$configBNAGName] += $configBNSize
            }      
        }
    }

    $agErr = @()
    $RDMASizes = @(GetVMSizesWithRDMA)
    foreach($agname in $AGVMSizeMap.Keys)
    {
        if($agname -eq "")
        {
            continue
        }

        # All in RDMASizes or none in RDMASizes
        if(@($AGVMSizeMap[$agname] | % { $_ -in $RDMASizes } | select -Unique).Count -ne 1)
        {
            $agErr += ("$agname : {0}`n" -f ($AGVMSizeMap[$agname] -join " "))
        }
    }

    if($agErr.Count -gt 0)
    {
        $errormsgs += "VM size combination error in the following affinity group(s), A8/A9 sizes cannot co-exist with other sizes: `n$agErr"
    }

    ## 5. Validate the naming of the config file
    foreach($agname in $AGVMSizeMap.Keys)
    {
        if($agname -eq "")
        {
            continue
        }

        try
        {
            ValidateAffinityGroupNaming -AffinityGroupName $agname
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }
    }

    try
    {
        ValidateStorageAccountNaming -StorageAccountName $config.Subscription.StorageAccount.Trim()
    }
    catch
    {
        $errormsgs += $_.Exception.Message
    }

    if($config["VNet"] -ne $null)
    {
        $configVNetName = $config.VNet.VNetName.Trim()
        $configSubnetName = $config.VNet.SubnetName.Trim()
        try
        {
            ValidateVNetNaming -VNetName $configVNetName -SubnetName $configSubnetName
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }
    }

    if($config.Domain["DomainController"] -ne $null)
    {
        $configDCName = $config.Domain.DomainController.VMName.Trim();
        $configDCSvcName = $config.Domain.DomainController.VMName.Trim();

        try
        {
            ValidateVMNaming -VMName $configDCName
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }

        try
        {
            ValidateServiceNaming -ServiceName $configDCSvcName
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }
    }

    if($config.Database["DBServer"] -ne $null)
    {
        $configRemoteDBName = $config.Database.DBServer.VMName.Trim()
        $configRemoteDBSvcName = $config.Database.DBServer.ServiceName.Trim()

        try
        {
            ValidateVMNaming -VMName $configRemoteDBName
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }

        try
        {
            ValidateServiceNaming -ServiceName $configRemoteDBSvcName
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }
    }

    if($config["HeadNode"] -ne $null)
    {
        $configHNName = $config.HeadNode.VMName.Trim()
        $configHNSvcName = $config.HeadNode.ServiceName.Trim()

        try
        {
            ValidateVMNaming -VMName $configHNName
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }

        try
        {
            ValidateServiceNaming -ServiceName $configHNSvcName
        }
        catch
        {
            $errormsgs += $_.Exception.Message
        }

        if($config.HeadNode["PreConfigScript"] -ne $null)
        {
            $configHNPreConfigScript = $config.HeadNode.PreConfigScript.Trim()
            if (-not (Test-Path -Path $configHNPreConfigScript -PathType Leaf))
            {
                $errormsgs += "The 'PreConfigScript' file does not exist: $configHNPreConfigScript"
            }
            elseif(-not $configHNPreConfigScript.ToLower().EndsWith(".ps1"))
            {
                $errormsgs += "The 'PreConfigScript' file is not a PowerShell script file: $configHNPreConfigScript"
            }
        }

        if($config.HeadNode["PostConfigScript"] -ne $null)
        {
            $configHNPostConfigScript = $config.HeadNode.PostConfigScript.Trim()
            if (-not (Test-Path -Path $configHNPostConfigScript -PathType Leaf))
            {
                $errormsgs += "The 'PostConfigScript' file does not exist: $configHNPostConfigScript"
            }
            elseif(-not $configHNPostConfigScript.ToLower().EndsWith(".ps1"))
            {
                $errormsgs += "The 'PostConfigScript' file is not a PowerShell script file: $configHNPostConfigScript"
            }
        }
    }

    if($config["ComputeNodes"] -ne $null)
    {
        $configCNNamePattern = $config.ComputeNodes.VMNamePattern.Trim()
        $configCNCount = [int]$config.ComputeNodes.NodeCount.Trim()
        try
        {
            ValidateVMPatternNaming -VMNamePattern $configCNNamePattern -NodeCount $configCNCount
        }
        catch
        {
            $errormsgs += $_.Exception.Message            
        }

        if($config.ComputeNodes["ServiceName"] -ne $null)
        {
            $configCNSvcName = $config.ComputeNodes.ServiceName.Trim()
            try
            {
                ValidateServiceNaming -ServiceName $configCNSvcName
            }
            catch
            {
                $errormsgs += $_.Exception.Message            
            }

            if($configCNCount -gt 50)
            {
                $errormsgs += "The NodeCount of ComputeNodes has exceeded the maximum value 50"
            }
        }
        else
        {
            $configCNSvcNamePattern = $config.ComputeNodes.ServiceNamePattern.Trim()
            $configCNMaxNodeCountPerSvc = [int]$config.ComputeNodes.MaxNodeCountPerService.Trim()
            $configCNStorageNamePattern = $config.ComputeNodes.StorageAccountNamePattern.Trim()
            if(($configCNMaxNodeCountPerSvc -eq 0) -or ($configCNMaxNodeCountPerSvc -gt 50))
            {
                $errormsgs += "The MaxNodeCountPerService of ComputeNodes must be between 1 and 50"
            }
            else
            {
                $configCNSetCount = [Math]::Ceiling($configCNCount/$configCNMaxNodeCountPerSvc)
                try
                {
                    ValidateServiceNamingPattern -ServiceNamePattern $configCNSvcNamePattern -Count $configCNSetCount
                }
                catch
                {
                    $errormsgs += $_.Exception.Message            
                }
                try
                {
                    ValidateStorageAccountNamingPattern -StorageAccountNamePattern $configCNStorageNamePattern -Count $configCNSetCount
                }
                catch
                {
                    $errormsgs += $_.Exception.Message            
                }
            }
        }
    }

    if($config["BrokerNodes"] -ne $null)
    {
        $configBNNamePattern = $config.BrokerNodes.VMNamePattern.Trim()
        $configBNCount = [int]$config.BrokerNodes.NodeCount.Trim()

        try
        {
            ValidateVMPatternNaming -VMNamePattern $configBNNamePattern -NodeCount $configBNCount
        }
        catch
        {
            $errormsgs += $_.Exception.Message            
        }

        if($config.BrokerNodes["ServiceName"] -ne $null)
        {
            $configBNSvcName = $config.BrokerNodes.ServiceName.Trim()
            if($configBNCount -gt 50)
            {
                $errormsgs += "The NodeCount of BrokerNodes has exceeded the maximum value 50"
            }
            try
            {
                ValidateServiceNaming -ServiceName $configBNSvcName
            }
            catch
            {
                $errormsgs += $_.Exception.Message            
            }
        }
        else
        {
            $errormsgs += "'ServiceNamePattern' is not supported in BrokerNodes, please specify 'ServiceName' instead."
        }

        if($config.BrokerNodes["OSVersion"] -ne $null)
        {
            if($config.BrokerNodes.OSVersion.Trim() -eq "WindowsServer2008R2")
            {
                $errormsgs +=  "Broker nodes do not support 'WindowsServer2008R2'. Please specify 'OSVersion' as 'WindowsServer2012' or 'WindowsServer2012R2'."
            }
        }
    }

    #6. Validate the Service conflict
    $services = @{}
    if($config.Domain["DomainController"] -ne $null)
    {
        $services[$configDCSvcName] = @($configDCAGName)
    }

    if($config.Database["DBServer"] -ne $null)
    {            
        if($services.Keys -notcontains $configRemoteDBSvcName)
        {
            $services[$configRemoteDBSvcName] = @($configRemoteDBAGName)
        }
        else
        {
            if($services[$configRemoteDBSvcName] -notcontains $configRemoteDBAGName)
            {
                $services[$configRemoteDBSvcName] += $configRemoteDBAGName
            }
        }
    }
    if($config["HeadNode"] -ne $null)
    {
        if($services.Keys -notcontains $configHNSvcName)
        {
            $services[$configHNSvcName] = @($configHNAGName)
        }
        else
        {
            if($services[$configHNSvcName] -notcontains $configHNAGName)
            {
                $services[$configHNSvcName] += $configHNAGName
            }
        }
    }

    if($config["BrokerNodes"] -ne $null)
    {
        if($config.BrokerNodes["ServiceName"] -ne $null)
        {
            if($services.Keys -contains $configBNSvcName)
            {
                $errormsgs += "Broker nodes cannot share cloud service with other VMs. You must specifiy a dedicated cloud service for broker nodes"
            }
            else
            {
                $services[$configBNSvcName] = @($configBNAGName)
            }
        }
        else
        {
            $errormsgs += "'ServiceNamePattern' is not supported for BrokerNodes."
        }
    }

    foreach($svc in $services.Keys)
    {
        if($services[$svc].Count -gt 1)
        {
            $errormsgs += ("The cloud service $svc cannot be specified in multiple affnity groups: " + $services[$svc])
        }
    }

    if($config["ComputeNodes"] -ne $null)
    {
        if($config.ComputeNodes["ServiceName"] -ne $null)
        {
            if($services.Keys -contains $configCNSvcName)
            {
                $errormsgs += "Compute nodes cannot share cloud service with other VMs. You must specifiy a dedicated cloud service for compute nodes."
            }
        }
        else
        {
            foreach($svc in $services.Keys)
            {
                if(NameMatchesPattern -Name $svc -NamePattern $configCNSvcNamePattern)
                {
                    $errormsgs += "The cloud service name $svc is conflicted with ServiceNamePattern for ComputeNodes $configCNSvcNamePattern."
                }
            }
        }
    }


    if($config["AzureBurst"] -ne $null)
    {
        $errormsgs += CheckAzureBustConfig $config["AzureBurst"]
    }

    ### Conclusion
    if($errormsgs.Count -gt 0)
    {
        $msg = ""
        $i = 1
        while($i -le $errormsgs.Count)
        {
            $msg += ("[Error $i] {0}`r`n" -f $errormsgs[$i-1])
            $i++
        }

        throw $msg
    }
}

function CheckStorageAccount
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String[]] $StorageAccountName,

        [Parameter(Mandatory=$true)]
        [String] $Location
    )

    $accountStatus = @{}
    $errMsg = @()
    $existingStorageAccounts = @(InvokeAzureCmdWithRetry -Command "Get-AzureStorageAccount")
    foreach($stg in $StorageAccountName)
    {
        TraceInfo "Checking storage account $stg"
        $storageAccount = $existingStorageAccounts | ?{$_.StorageAccountName -eq $stg}
        if($storageAccount -ne $null)
        {
            if($storageAccount.GeoPrimaryLocation -ne $Location)
            {
                $errMsg += "The storage account $stg already exists, but in region $($storageAccount.GeoPrimaryLocation), not in region $Location."
                continue
            }

            if($storageAccount.AffinityGroup -eq $null)
            {
                TraceInfo "The storage account $stg already exists in region $Location."
            }
            else
            {
                TraceInfo "The storage account $stg already exists in Affinity Group $($storageAccount.AffinityGroup)."
            }
        }
        else
        {
            if((InvokeAzureCmdWithRetry -Command "Test-AzureName -Storage -Name $stg") -eq $true)
            {
                $errMsg += "The storage account name $stg is not available. It is already in use."
                continue
            }
        }

        $accountStatus[$stg] = ($storageAccount -ne $null)
    }

    if($errMsg.Count -gt 0)
    {
        throw ($errMsg -join "`r`n")
    }

    return $accountStatus
}

function CheckLocationAndAffinityGroup
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $Location,

        [Parameter(Mandatory=$false)]
        [String[]] $AGNames
    )

    $locations = InvokeAzureCmdWithRetry -Command "Get-AzureLocation" | ForEach-Object {$_.Name}
    if($locations -notcontains $Location)
    {
        throw "Invalid Location: $Location."
    }
    else
    {
        TraceInfo "$Location is a valid Location."
    }

    $agStates = @{}
    if(($AGNames -ne $null) -and ($AGNames.Count -gt 0))
    {
        $errmsgs = @()
        foreach($agname in $AGNames)
        {
            $ag = InvokeAzureCmdWithRetry -Command "Get-AzureAffinityGroup -Name $agname"
            if($ag -ne $null)
            {
                if($ag.Location -ne $Location)
                {
                    $errmsgs += "The affinity group $agname already exists in region $($ag.Location), but not in region $Location."
                }
                else
                {
                    TraceInfo "The affinity group $agname already exists."
                    $agStates[$agname] = $true
                }
            }
            else
            {
                TraceInfo "The affinity group $agname doesn't exist yet."
                $agStates[$agname] = $false
            }
        }
        if($errmsgs.Count -gt 0)
        {
            throw ($errmsgs -join "`r`n")
        }
    }

    return $agStates
}

function CheckVNet
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $VNetName,

        [Parameter(Mandatory=$true)]
        [String] $SubnetName,

        [Parameter(Mandatory=$true)]
        [String] $Location
    )

    $vnetExists = $false
    $vnet = Get-AzureVNetSiteWithRetry -VNetName $VNetName -ErrorAction SilentlyContinue
    if ($vnet -ne $null)
    {
        $vnetExists = $true
        # The VNet already exists, check whether the settings are consistent
        # Decide whether it is a Regional VNet by Property AffinityGroup, for Regional VNet, AffinityGroup is $null
        if($vnet.AffinityGroup -eq $null)
        {
            # Check the location information for Regional VNet
            # Currently, we cannot get the Location by Get-AzureVNetSite command for Regional VNet
            $vnetLocation = GetRegionalVNetLocation -VNetName $VNetName
            if($vnetLocation -ne $Location)
            {
                throw "The virtual network $VNetName already exists but in region $vnetLocation, not in $Location."
            }

            TraceInfo "The virtual network $VNetName already exists in region $Location."                    
        }
        else
        {
            $vnetAG = $vnet.AffinityGroup
            throw "The virtual network $VNetName already exists but is assiciated with an affinity group $vnetAG."
        }

        ### Validate Subnet
        $subnet = $vnet.Subnets | Where-Object Name -eq $SubnetName
        if($subnet -eq $null)
        {
            throw "The virtual network $VNetName already exists but no subnet $SubnetName configured in it."
        }
        else
        {
            TraceInfo "The subnet $SubnetName already configured in $VNetName."
        }
    }
    else
    {
        TraceInfo "The virtual network $VNetName doesn't exist."
    }

    return $vnetExists
}

function CheckAzureVMImage
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $ImageName,

        [Parameter(Mandatory=$false)]
        [String] $Location = "",

        # The image list to be searched, if not specified, use Get-AzureVMImage to get the image.
        [Parameter(Mandatory = $false)]
        [Object[]] $FromImageList = @()
    )

    if(($FromImageList -eq $null) -or ($FromImageList.Count -eq 0))
    {
        $image = InvokeAzureCmdWithRetry -Command "Get-AzureVMImage -ImageName $ImageName"
    }
    else
    {
        $image = $FromImageList | ?{$_.ImageName -eq $ImageName}
    }

    if($image -eq $null)
    {
        throw "The image $ImageName does not exist"
    }
    
    if(-not [String]::IsNullOrEmpty($Location))
    {
        $imgLocations = @()        
        if($image.Location -ne $null)
        {
            $imgLocations += $image.Location.Split(';')
        }
        elseif($image.AffinityGroup -ne $null)
        {
            $agName = $image.AffinityGroup
            $ag = InvokeAzureCmdWithRetry -Command "Get-AzureAffinityGroup -Name $agName"
            if($ag -ne $null)
            {
                $imgLocations += $ag.Location
            }
        }

        if($imgLocations -notcontains $Location)
        {
            throw "The image $ImageName found, but not in region $Location"
        }
    }
}


function CheckNodesInService
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [String[]] $TargetNodeNames,
        
        [Parameter(Mandatory=$true)]
        [ValidateSet("ComputeNode","BrokerNode")]
        [String] $NodeType,

        [Parameter(Mandatory=$false)]
        [String] $ImageName = "",
                
        [Parameter(Mandatory=$false)]
        [Switch] $ImageHPCInstalled
    )

    TraceInfo "Checking virtual machines in cloud service $ServiceName"
    $VMsInService = @(InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $ServiceName")
    # Check whether there is any VM other than the target VMs in the deployment
    $otherVMNames = @($VMsInService | ?{$TargetNodeNames -notcontains $_.Name} | %{$_.Name})
    if($otherVMNames.Count -gt 0)
    {
        TraceWarning "There are some other virtual machines in the cloud service $ServiceName : $otherVMNames"
    }

    $existingNodes = @($VMsInService | ?{$TargetNodeNames -contains $_.Name})
    # Check the source image of the existing VMs
    $imgNotMatch = @()
    $srcImgNames = @()
    foreach($vm in $existingNodes)
    {
        # There are two types of VM images: VMImage and OSImage
        # If the VM was created from VMImage, the SourceImageName is VMImageContext.OSDiskConfiguration.Name (typically as <ImageName>-os-yyyy-MM-dd)
        # If the VM was created from OSImage, the SourceImageName is OSImageContext.ImageName
        $srcImgName = $vm.VM.OSVirtualHardDisk.SourceImageName
        $srcImgNames += $srcImgName
        if(-not [String]::IsNullOrEmpty($ImageName) -and $ImageHPCInstalled.IsPresent)
        {
            # if the given image is HPC installed, the source image name shall be <ImageName>[_<StorageAccountName>] or <ImageName>-os-yyyy-MM-dd
            if(($srcImgName -notmatch "^$ImageName(_[0-9a-z]{3,24})?$") -and ($srcImgName -notmatch "^$ImageName-os-"))
            {
                $imgNotMatch += $vm.Name
            }
        }
        else
        {
            # if the given image is not HPC installed, means we will create the base HPC node image and create nodes with this base image.
            # The source image name shall be HPC-Pack-<NodeType>_<DateTime>[_<StorageAccountName>] or HPC-Pack-<NodeType>_<DateTime>-os-yyyy-MM-dd
            if(-not (IsBaseHPCNodeImage -VMSrcImageName $srcImgName -NodeType $NodeType))
            {
                $imgNotMatch += $vm.Name
            }
        }
    }
    if($imgNotMatch.Count -gt 0)
    {
        throw ("The following virtual machines already exist in cloud service $ServiceName but not created by this script: " + ($imgNotMatch -join ","))
    }

    $readyVMNames = @()
    if($existingNodes.Count -gt 0)
    {
        $srcImgNames = @($srcImgNames | select -Unique | Sort-Object -Descending)
        $readyVMNames = @($existingNodes | %{$_.Name})
    }

    return @{"ReadyVMNames"=$readyVMNames;"VMImages"=$srcImgNames}
}

function CheckNodeGroup
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $NamePattern, 

        [Parameter(Mandatory=$true)]
        [int] $NodeCount, 

        [Parameter(Mandatory=$true)]
        [String] $VMSize,

        [Parameter(Mandatory=$true)]
        [String[]] $AvailableVMSizes,
        
        [Parameter(Mandatory=$true)]
        [Object[]] $AvailableVMImages,
                
        [Parameter(Mandatory=$true, ParameterSetName="SingleService")]
        [String] $ServiceName, 
        
        [Parameter(Mandatory=$true, ParameterSetName="ServicePattern")]
        [String] $ServiceNamePattern, 

        [Parameter(Mandatory=$true, ParameterSetName="ServicePattern")]
        [String] $StorageAccountNamePattern,

        [Parameter(Mandatory=$true, ParameterSetName="ServicePattern")]
        [int] $MaxNodeCountPerService,

        [Parameter(Mandatory=$true)]
        [HashTable] $ServiceStateTable, 
        
        [Parameter(Mandatory=$true)]
        [ValidateSet("ComputeNode","BrokerNode")]
        [String] $NodeType,

        [Parameter(Mandatory=$true)]
        [String] $ClusterName,
                
        [Parameter(Mandatory=$false)]
        [String] $ImageName = "",

        [Parameter(Mandatory=$true)]
        [String] $Location,
                
        [Parameter(Mandatory=$false)]
        [Switch] $ImageHPCInstalled
    )

    try
    {
        ValidateVMSize -VMName $NamePattern -VMSize $VMSize -AvailableSizes $AvailableVMSizes
        if(-not [String]::IsNullOrEmpty($ImageName))
        {
            CheckAzureVMImage -ImageName $ImageName -Location $Location -FromImageList $AvailableVMImages
        }

        $srcImageNames = @()
        $serviceVMTable = @{}
        if($PsCmdlet.ParameterSetName -eq "ServicePattern")
        {
            $nodeNameSets = @(GetNameSetsByPattern -NamePattern $NamePattern -Count $NodeCount -MaxCountPerSet $MaxNodeCountPerService)
            $nodeSetCount = $nodeNameSets.Count
            $serviceNames = @(GetNamesByPattern -NamePattern $ServiceNamePattern -Count $nodeSetCount)
            $storageNames = @(GetNamesByPattern -NamePattern $StorageAccountNamePattern -Count $nodeSetCount)
        }
        else
        {
            $nodeNameSets = @(, @(GetNamesByPattern -NamePattern $NamePattern -Count $NodeCount))
            $nodeSetCount = 1
            $serviceNames = @($ServiceName)
        }
        $totalReadyVMCount = 0
        for($i=0;$i -lt $nodeSetCount; $i++)
        {
            $serviceName = $serviceNames[$i]
            $targetNodeNames = $nodeNameSets[$i]
            if($ServiceStateTable[$serviceName] -eq "Deployed")
            {
                $result = CheckNodesInService -ServiceName $serviceName -TargetNodeNames $targetNodeNames -NodeType $NodeType `
                                -ImageName $ImageName -ImageHPCInstalled:$ImageHPCInstalled
                $nodeNamesToCreate = @($targetNodeNames | ?{$result["ReadyVMNames"] -notcontains $_})
                $srcImageNames += $result["VMImages"]
                if($nodeNamesToCreate.Count -gt 0)
                {
                    $serviceVMTable[$serviceName] = $nodeNamesToCreate
                }

                $totalReadyVMCount += $result["ReadyVMNames"].Count
            }
            else
            {
                $serviceVMTable[$serviceName] = $targetNodeNames
            }
        }

        $hpcNodeImageName = ""
        if($ImageHPCInstalled)
        {
            $hpcNodeImageName = $ImageName
        }
        elseif($srcImageNames.Count -gt 0)
        {
            $imageList = @(Get-BaseHPCNodeImage -NodeType $NodeType -Location $Location)
            $srcImages = @()
            if($imageList.Count -gt 0)
            {
                $srcImageNames = $srcImageNames | select -Unique
                foreach($srcImgName in $srcImageNames)
                {
                    $img = GetVMImageByVMSourceImageName -SourceImageName $srcImgName -FromImageList $imageList
                    if(($img -ne $null) -and (IsBaseHPCNodeImage -ImageName $img.ImageName -NodeType $NodeType))
                    {
                        $srcImages += $img
                    }
                }
            }

            if($srcImages.Count -gt 0)
            {
                # The format of the hpc node base image name is HPC-Pack-<NodeType>_<DateTime>[_<StorageAccountName>]
                # Sort the images in descending order by name(<DateTime>[_<StorageAccountName>] part), so that the latest image can be selected.
                $srcImages = $srcImages | Sort-Object -Property ImageName -Descending
                # The first image is probably with tailing _<StorageAccountName>, try to find an image with same <DateTime> but without _<StorageAccountName>.
                $firstImgStg = GetStorageAccountFromMediaLink -MediaLink (GetVMImageMediaLink $srcImages[0])
                $hpcNodeImageName = $srcImages[0].ImageName
                if($hpcNodeImageName -match "_$firstImgStg$")
                {
                    $firstImgNameRoot = $hpcNodeImageName.Substring(0, $hpcNodeImageName.Length - $firstImgStg.Length - 1)
                    $imgWithRootName = $srcImages | ?{$_.ImageName -eq $firstImgNameRoot}
                    if($imgWithRootName -ne $null)
                    {
                        $hpcNodeImageName = $imgWithRootName.ImageName
                    }                
                }
            }
        }

        TraceInfo "$totalReadyVMCount node(s) already exist, $($NodeCount - $totalReadyVMCount) node(s) will be created."

        return @($serviceVMTable, $hpcNodeImageName)
    }
    catch
    {
        throw
    }
}


function GetVMImageByVMSourceImageName
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String]$SourceImageName,

        [Parameter(Mandatory=$true)]
        [Object[]] $FromImageList
    )
  
    if($FromImageList -ne $null -and $FromImageList.Count -gt 0)
    {
        foreach($img in $FromImageList)
        {
            if($img.ImageName -eq $SourceImageName)
            {
                return $img
            }
            if($img.PSObject.Properties.Name -contains "OSDiskConfiguration" -and $img.OSDiskConfiguration.PSObject.Properties.Name -contains "Name")
            {
                if($img.OSDiskConfiguration.Name -eq $SourceImageName)
                {
                    return $img
                }
            }
        }
    }

    return $null
}


function CheckAzureBustConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $AzureBustConfig
    )

    $errMsgs = @()
    $certIds = @()
    foreach($cert in $AzureBustConfig.Certificate)
    {        
        $certId = [Int]$cert.Id.Trim()
        $certIds += $certId

        $pfxFile = $cert.PfxFile.Trim()
        if(-not (Test-Path -Path $pfxFile))
        {
            $errMsgs += "[Certificate $certId] The certificate file $pfxFile doesn't exist."
        }
    }

    foreach($group in ($certIds | group))
    {
        if($group.Count -gt 1)
        {
            $errMsgs += "[Certificate] There are two or more certificates with the same Id: " + $group.Name
        }
    }

    $tempNames = @()
    $svcNames = @()
    foreach($xmlTemp in $AzureBustConfig.AzureNodeTemplate)
    {
        $tempErrMsgs = @()
        $tempName = $xmlTemp.TemplateName.Trim()
        $tempNames += $tempName

        $svcName = $xmlTemp.ServiceName.Trim()
        try
        {
            ValidateServiceNaming -ServiceName $svcName | Out-Null
        }
        catch
        {
            $tempErrMsgs += $_.Exception.Message
        }

        try
        {
            ValidateStorageAccountNaming -StorageAccountName $xmlTemp.StorageAccount.Trim() | Out-Null
        }
        catch
        {
            $tempErrMsgs += $_.Exception.Message
        }

        $subId = $xmlTemp.SubscriptionId.Trim()
        try
        {
            [System.Guid]::Parse($subId) | Out-Null
        }
        catch
        {
            $tempErrMsgs += "The 'SubscriptionId' is invalid: $subId"
        }

        $svcNames += $svcName

        $certId = [Int]$xmlTemp.CertificateId.Trim()
        if($certIds -notcontains $certId)
        {
            $tempErrMsgs += "The 'CertificateId' $certId cannot be found."
        }

        if(($xmlTemp["NodeCount"] -ne $null) -and ($xmlTemp["RoleSize"] -eq $null))
        {
            $tempErrMsgs += "'RoleSize' must be specified when 'NodeCount' is specified."
        }

        if($xmlTemp["Proxy"] -ne $null)
        {
            if($xmlTemp.Proxy.UsesStaticProxyCount.Trim() -eq "true")
            {
                if($xmlTemp.Proxy["ProxyMultiplicity"] -ne $null)
                {
                    $proxyNum = [int]$xmlTemp.Proxy.ProxyMultiplicity.Trim()
                    if($proxyNum -gt 10 -or $proxyNum -lt 2)
                    {
                        $tempErrMsgs += "The value of 'ProxyMultiplicity' is invalid: $proxyNum (allowed range 2 - 10)."
                    }
                }
            }
            else
            {
                if($xmlTemp.Proxy["ProxyRatio"] -ne $null)
                {
                    $prxoyRatio = [int]$xmlTemp.Proxy.ProxyRatio.Trim()
                    if($prxoyRatio -gt 400 -or $prxoyRatio -lt 100)
                    {
                        $tempErrMsgs += "The value of 'ProxyRatio' is invalid: $prxoyRatio (allowed range 100 - 400)."
                    }
                }
            
                if($xmlTemp.Proxy["ProxyRatioBase"] -ne $null)
                {
                    $prxoyRatioBase = [int]$xmlTemp.Proxy.ProxyRatioBase.Trim()
                    if($prxoyRatioBase -gt 600 -or $prxoyRatioBase -lt 200)
                    {
                        $tempErrMsgs += "The value of 'ProxyRatioBase' is invalid: $prxoyRatioBase (allowed range 200 - 600)."
                    }
                }
            }
        }

        foreach($msg in $tempErrMsgs)
        {
            $errMsgs += "[AzureNodeTemplate $tempName] $msg"
        }

        $tempErrMsgs.Clear()
    }

    foreach($group in ($tempNames | group))
    {
        if($group.Count -gt 1)
        {
            $errMsgs += "[AzureNodeTemplate] There are two or more AzureNodeTemplates with the same 'TemplateName': " + $group.Name
        }
    }

    foreach($group in ($svcNames | group))
    {
        if($group.Count -gt 1)
        {
            $errMsgs += "[AzureNodeTemplate] The cloud service name $($group.Name) is specified in two or more AzureNodeTemplates."
        }
    } 
    
    return $errMsgs
}

function ValidateCertificates
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $AzureBustConfig
    )

    $certPasswords = @{}
    foreach($cert in $AzureBustConfig.Certificate)
    {        
        $certId = [Int]$cert.Id.Trim()
        $pfxFile = $cert.PfxFile.Trim()
        $password = ""
        if($cert["Password"] -ne $null)
        {
            $password = $cert.Password.Trim()
        }

        $retry = 0
        while($true)
        {
            try
            {
                $pfxCert = GetCertificate $pfxFile $password
                if(-not [String]::IsNullOrEmpty($password))
                {
                    $certPasswords[$certId] = $password
                }

                break
            }
            catch
            {
                if($_.Exception.HResult -eq 0x80070056)
                {
                    if([String]::IsNullOrEmpty($password))
                    {
                        $prompt = "The certificate file $pfxFile is password protected. Please input the password"
                    }
                    else
                    {
                        $prompt = "The password for the certificate file $pfxFile is incorrect. Please input the correct password"
                    }
                    
                    if($retry -lt 3)
                    {
                        $secPsw = Read-Host -Prompt $prompt -AsSecureString
                        $password = ConvertSecureStrToPlain -SecurePassword $secPsw 
                        $retry++
                    }
                    else
                    {
                        throw "The password for the certificate file $pfxFile is incorrect."
                    }
                }
                elseif($_.Exception.HResult -eq 0x80092009)
                {
                    throw "The file $pfxFile is not a valid PFX file."
                }
                else
                {
                    throw
                }
            }
        }

        if(-not $pfxCert.HasPrivateKey)
        {
            throw "The certificate file $pfxFile doesn't contain a private key."
        }
    }

    return $certPasswords
}
# SIG # Begin signature block
# MIIavwYJKoZIhvcNAQcCoIIasDCCGqwCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUKttmIcGTNRnLXPN2rOB7r6l0
# mLKgghV6MIIEuzCCA6OgAwIBAgITMwAAAFwJq3ADEfxcFQAAAAAAXDANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTQwNTIzMTcxMzE2
# WhcNMTUwODIzMTcxMzE2WjCBqzELMAkGA1UEBhMCVVMxCzAJBgNVBAgTAldBMRAw
# DgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24x
# DTALBgNVBAsTBE1PUFIxJzAlBgNVBAsTHm5DaXBoZXIgRFNFIEVTTjpCQkVDLTMw
# Q0EtMkRCRTElMCMGA1UEAxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZTCC
# ASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBANYJa/7h+ToIHrYOs0PLI95O
# bOJZcc75TxmjL8QLy/H1Xyl5ielMpmU8xUtMXQHp/i9QvCCDC5hHW5nqTZ/eqcsj
# lYzk07XcemMgs/L/r8dp/5K7jQpF7punPfdkmj6O6rRaABbNqeXVTJUmQH8DmFyk
# 6dBH1jblvqVRQb9b9uyuNN9K8gXa/fw9YY6yEa11unZrY2JOJSHPpz88ub9uBEMk
# HrvplbnGydEgydhVI1Xnsr+vLVxathcwcTMQJH5xnrFl4ma2BdMyV0nEYhPYkoyB
# kY8zQ5EZbBL0eRTm5Z/F5anhM0C22y9X05U2It3yQqmUtGTTgSMUlMPZRhdm3r0C
# AwEAAaOCAQkwggEFMB0GA1UdDgQWBBR/kUhDg4fJeBzpVcW521b4F9FHBDAfBgNV
# HSMEGDAWgBQjNPjZUkZwCu1A+3b7syuwwzWzDzBUBgNVHR8ETTBLMEmgR6BFhkNo
# dHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNyb3Nv
# ZnRUaW1lU3RhbXBQQ0EuY3JsMFgGCCsGAQUFBwEBBEwwSjBIBggrBgEFBQcwAoY8
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNyb3NvZnRUaW1l
# U3RhbXBQQ0EuY3J0MBMGA1UdJQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEBBQUA
# A4IBAQCCCVb4b+5GzEwp9r7Tx2pf+V2EABkXgYloZBU/wUQ2OxhtBMusE7eOlOMt
# 66P0QnxE0QnsWr7nk9OEcuKP1R3VOVP7ciILZpI7ysqr+s/MZE7GRGyv4IOOFBG4
# sJBJn8OmlM+D4BFiWqtYo7hhlyR2rG1D947qHfOI/ipL7h/0HW2sIEyIAEAutLlI
# 2tHEbsGt9DOFYQuJLF5rygNjUkox7r1VwxCnUaiSrxJgxZxdRmPTzOqfNcatCSB0
# hTUbRRKzllHhye2bgOV25fdWhesgnCiRVPtoulaFg1Vh/4hAXLENJgejxVfsFXkP
# lvSOdqGKZ1qNf05i9H7fEKDWFLqXMIIE7DCCA9SgAwIBAgITMwAAAMps1TISNcTh
# VQABAAAAyjANBgkqhkiG9w0BAQUFADB5MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSMwIQYDVQQDExpNaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QTAeFw0xNDA0MjIxNzM5MDBaFw0xNTA3MjIxNzM5MDBaMIGDMQswCQYDVQQGEwJV
# UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
# ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMQ0wCwYDVQQLEwRNT1BSMR4wHAYDVQQD
# ExVNaWNyb3NvZnQgQ29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAw
# ggEKAoIBAQCWcV3tBkb6hMudW7dGx7DhtBE5A62xFXNgnOuntm4aPD//ZeM08aal
# IV5WmWxY5JKhClzC09xSLwxlmiBhQFMxnGyPIX26+f4TUFJglTpbuVildGFBqZTg
# rSZOTKGXcEknXnxnyk8ecYRGvB1LtuIPxcYnyQfmegqlFwAZTHBFOC2BtFCqxWfR
# +nm8xcyhcpv0JTSY+FTfEjk4Ei+ka6Wafsdi0dzP7T00+LnfNTC67HkyqeGprFVN
# TH9MVsMTC3bxB/nMR6z7iNVSpR4o+j0tz8+EmIZxZRHPhckJRIbhb+ex/KxARKWp
# iyM/gkmd1ZZZUBNZGHP/QwytK9R/MEBnAgMBAAGjggFgMIIBXDATBgNVHSUEDDAK
# BggrBgEFBQcDAzAdBgNVHQ4EFgQUH17iXVCNVoa+SjzPBOinh7XLv4MwUQYDVR0R
# BEowSKRGMEQxDTALBgNVBAsTBE1PUFIxMzAxBgNVBAUTKjMxNTk1K2I0MjE4ZjEz
# LTZmY2EtNDkwZi05YzQ3LTNmYzU1N2RmYzQ0MDAfBgNVHSMEGDAWgBTLEejK0rQW
# WAHJNy4zFha5TJoKHzBWBgNVHR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1pY3Jv
# c29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNDb2RTaWdQQ0FfMDgtMzEtMjAx
# MC5jcmwwWgYIKwYBBQUHAQEETjBMMEoGCCsGAQUFBzAChj5odHRwOi8vd3d3Lm1p
# Y3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY0NvZFNpZ1BDQV8wOC0zMS0yMDEwLmNy
# dDANBgkqhkiG9w0BAQUFAAOCAQEAd1zr15E9zb17g9mFqbBDnXN8F8kP7Tbbx7Us
# G177VAU6g3FAgQmit3EmXtZ9tmw7yapfXQMYKh0nfgfpxWUftc8Nt1THKDhaiOd7
# wRm2VjK64szLk9uvbg9dRPXUsO8b1U7Brw7vIJvy4f4nXejF/2H2GdIoCiKd381w
# gp4YctgjzHosQ+7/6sDg5h2qnpczAFJvB7jTiGzepAY1p8JThmURdwmPNVm52Iao
# AP74MX0s9IwFncDB1XdybOlNWSaD8cKyiFeTNQB8UCu8Wfz+HCk4gtPeUpdFKRhO
# lludul8bo/EnUOoHlehtNA04V9w3KDWVOjic1O1qhV0OIhFeezCCBbwwggOkoAMC
# AQICCmEzJhoAAAAAADEwDQYJKoZIhvcNAQEFBQAwXzETMBEGCgmSJomT8ixkARkW
# A2NvbTEZMBcGCgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9z
# b2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5MB4XDTEwMDgzMTIyMTkzMloX
# DTIwMDgzMTIyMjkzMloweTELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0
# b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3Jh
# dGlvbjEjMCEGA1UEAxMaTWljcm9zb2Z0IENvZGUgU2lnbmluZyBQQ0EwggEiMA0G
# CSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCycllcGTBkvx2aYCAgQpl2U2w+G9Zv
# zMvx6mv+lxYQ4N86dIMaty+gMuz/3sJCTiPVcgDbNVcKicquIEn08GisTUuNpb15
# S3GbRwfa/SXfnXWIz6pzRH/XgdvzvfI2pMlcRdyvrT3gKGiXGqelcnNW8ReU5P01
# lHKg1nZfHndFg4U4FtBzWwW6Z1KNpbJpL9oZC/6SdCnidi9U3RQwWfjSjWL9y8lf
# RjFQuScT5EAwz3IpECgixzdOPaAyPZDNoTgGhVxOVoIoKgUyt0vXT2Pn0i1i8UU9
# 56wIAPZGoZ7RW4wmU+h6qkryRs83PDietHdcpReejcsRj1Y8wawJXwPTAgMBAAGj
# ggFeMIIBWjAPBgNVHRMBAf8EBTADAQH/MB0GA1UdDgQWBBTLEejK0rQWWAHJNy4z
# Fha5TJoKHzALBgNVHQ8EBAMCAYYwEgYJKwYBBAGCNxUBBAUCAwEAATAjBgkrBgEE
# AYI3FQIEFgQU/dExTtMmipXhmGA7qDFvpjy82C0wGQYJKwYBBAGCNxQCBAweCgBT
# AHUAYgBDAEEwHwYDVR0jBBgwFoAUDqyCYEBWJ5flJRP8KuEKU5VZ5KQwUAYDVR0f
# BEkwRzBFoEOgQYY/aHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJv
# ZHVjdHMvbWljcm9zb2Z0cm9vdGNlcnQuY3JsMFQGCCsGAQUFBwEBBEgwRjBEBggr
# BgEFBQcwAoY4aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNy
# b3NvZnRSb290Q2VydC5jcnQwDQYJKoZIhvcNAQEFBQADggIBAFk5Pn8mRq/rb0Cx
# MrVq6w4vbqhJ9+tfde1MOy3XQ60L/svpLTGjI8x8UJiAIV2sPS9MuqKoVpzjcLu4
# tPh5tUly9z7qQX/K4QwXaculnCAt+gtQxFbNLeNK0rxw56gNogOlVuC4iktX8pVC
# nPHz7+7jhh80PLhWmvBTI4UqpIIck+KUBx3y4k74jKHK6BOlkU7IG9KPcpUqcW2b
# Gvgc8FPWZ8wi/1wdzaKMvSeyeWNWRKJRzfnpo1hW3ZsCRUQvX/TartSCMm78pJUT
# 5Otp56miLL7IKxAOZY6Z2/Wi+hImCWU4lPF6H0q70eFW6NB4lhhcyTUWX92THUmO
# Lb6tNEQc7hAVGgBd3TVbIc6YxwnuhQ6MT20OE049fClInHLR82zKwexwo1eSV32U
# jaAbSANa98+jZwp0pTbtLS8XyOZyNxL0b7E8Z4L5UrKNMxZlHg6K3RDeZPRvzkbU
# 0xfpecQEtNP7LN8fip6sCvsTJ0Ct5PnhqX9GuwdgR2VgQE6wQuxO7bN2edgKNAlt
# HIAxH+IOVN3lofvlRxCtZJj/UBYufL8FIXrilUEnacOTj5XJjdibIa4NXJzwoq6G
# aIMMai27dmsAHZat8hZ79haDJLmIz2qoRzEvmtzjcT3XAH5iR9HOiMm4GPoOco3B
# oz2vAkBq/2mbluIQqBC0N1AI1sM9MIIGBzCCA++gAwIBAgIKYRZoNAAAAAAAHDAN
# BgkqhkiG9w0BAQUFADBfMRMwEQYKCZImiZPyLGQBGRYDY29tMRkwFwYKCZImiZPy
# LGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQDEyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZp
# Y2F0ZSBBdXRob3JpdHkwHhcNMDcwNDAzMTI1MzA5WhcNMjEwNDAzMTMwMzA5WjB3
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEwHwYDVQQDExhN
# aWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAw
# ggEKAoIBAQCfoWyx39tIkip8ay4Z4b3i48WZUSNQrc7dGE4kD+7Rp9FMrXQwIBHr
# B9VUlRVJlBtCkq6YXDAm2gBr6Hu97IkHD/cOBJjwicwfyzMkh53y9GccLPx754gd
# 6udOo6HBI1PKjfpFzwnQXq/QsEIEovmmbJNn1yjcRlOwhtDlKEYuJ6yGT1VSDOQD
# LPtqkJAwbofzWTCd+n7Wl7PoIZd++NIT8wi3U21StEWQn0gASkdmEScpZqiX5NMG
# gUqi+YSnEUcUCYKfhO1VeP4Bmh1QCIUAEDBG7bfeI0a7xC1Un68eeEExd8yb3zuD
# k6FhArUdDbH895uyAc4iS1T/+QXDwiALAgMBAAGjggGrMIIBpzAPBgNVHRMBAf8E
# BTADAQH/MB0GA1UdDgQWBBQjNPjZUkZwCu1A+3b7syuwwzWzDzALBgNVHQ8EBAMC
# AYYwEAYJKwYBBAGCNxUBBAMCAQAwgZgGA1UdIwSBkDCBjYAUDqyCYEBWJ5flJRP8
# KuEKU5VZ5KShY6RhMF8xEzARBgoJkiaJk/IsZAEZFgNjb20xGTAXBgoJkiaJk/Is
# ZAEZFgltaWNyb3NvZnQxLTArBgNVBAMTJE1pY3Jvc29mdCBSb290IENlcnRpZmlj
# YXRlIEF1dGhvcml0eYIQea0WoUqgpa1Mc1j0BxMuZTBQBgNVHR8ESTBHMEWgQ6BB
# hj9odHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9taWNy
# b3NvZnRyb290Y2VydC5jcmwwVAYIKwYBBQUHAQEESDBGMEQGCCsGAQUFBzAChjho
# dHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY3Jvc29mdFJvb3RD
# ZXJ0LmNydDATBgNVHSUEDDAKBggrBgEFBQcDCDANBgkqhkiG9w0BAQUFAAOCAgEA
# EJeKw1wDRDbd6bStd9vOeVFNAbEudHFbbQwTq86+e4+4LtQSooxtYrhXAstOIBNQ
# md16QOJXu69YmhzhHQGGrLt48ovQ7DsB7uK+jwoFyI1I4vBTFd1Pq5Lk541q1YDB
# 5pTyBi+FA+mRKiQicPv2/OR4mS4N9wficLwYTp2OawpylbihOZxnLcVRDupiXD8W
# mIsgP+IHGjL5zDFKdjE9K3ILyOpwPf+FChPfwgphjvDXuBfrTot/xTUrXqO/67x9
# C0J71FNyIe4wyrt4ZVxbARcKFA7S2hSY9Ty5ZlizLS/n+YWGzFFW6J1wlGysOUzU
# 9nm/qhh6YinvopspNAZ3GmLJPR5tH4LwC8csu89Ds+X57H2146SodDW4TsVxIxIm
# dgs8UoxxWkZDFLyzs7BNZ8ifQv+AeSGAnhUwZuhCEl4ayJ4iIdBD6Svpu/RIzCzU
# 2DKATCYqSCRfWupW76bemZ3KOm+9gSd0BhHudiG/m4LBJ1S2sWo9iaF2YbRuoROm
# v6pH8BJv/YoybLL+31HIjCPJZr2dHYcSZAI9La9Zj7jkIeW1sMpjtHhUBdRBLlCs
# lLCleKuzoJZ1GtmShxN1Ii8yqAhuoFuMJb+g74TKIdbrHk/Jmu5J4PcBZW+JC33I
# acjmbuqnl84xKf8OxVtc2E0bodj6L54/LlUWa8kTo/0xggSvMIIEqwIBATCBkDB5
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSMwIQYDVQQDExpN
# aWNyb3NvZnQgQ29kZSBTaWduaW5nIFBDQQITMwAAAMps1TISNcThVQABAAAAyjAJ
# BgUrDgMCGgUAoIHIMBkGCSqGSIb3DQEJAzEMBgorBgEEAYI3AgEEMBwGCisGAQQB
# gjcCAQsxDjAMBgorBgEEAYI3AgEVMCMGCSqGSIb3DQEJBDEWBBSEcPMgkl33Dk40
# NYVcTWtC6ifnOjBoBgorBgEEAYI3AgEMMVowWKA2gDQATQBpAGMAcgBvAHMAbwBm
# AHQAIABIAFAAQwAgAFAAYQBjAGsAIAAyADAAMQAyACAAUgAyoR6AHGh0dHA6Ly93
# d3cubWljcm9zb2Z0LmNvbS9IUEMwDQYJKoZIhvcNAQEBBQAEggEAexZ+OzuQDjyU
# 0j1dP/nDhPFp4L4elOFQ/4nAJX+gd9LiunKYlyW27jvmZdZyPtfcSsrtMgeGGV6x
# poJdPdz78F9m/PUXt3442X1bnGfyKyjuw0SoFbPBG3qUEb4ZpVjbY7m0mL8T4iir
# OIRXdDQI33pBv4J17y7ns7cAjmrJcLkw8XAhECGQNn1nF7LrlbtXFJDFvmINH1+X
# M2H3rTekvBPDCz9ozU2BMLLySgsnf59cyo/ZHHSyRR9C5h/ii2apa7CXRJszhKBY
# 0QgSwD4zuYc/Fv2ODC+jg3h23itgqKv6DrG9RLw0vrHa33bsuSNfmDMX2BbSW3qe
# 1wcyjsKbw6GCAigwggIkBgkqhkiG9w0BCQYxggIVMIICEQIBATCBjjB3MQswCQYD
# VQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEe
# MBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEwHwYDVQQDExhNaWNyb3Nv
# ZnQgVGltZS1TdGFtcCBQQ0ECEzMAAABcCatwAxH8XBUAAAAAAFwwCQYFKw4DAhoF
# AKBdMBgGCSqGSIb3DQEJAzELBgkqhkiG9w0BBwEwHAYJKoZIhvcNAQkFMQ8XDTE1
# MDIxMjE4MTQ1NlowIwYJKoZIhvcNAQkEMRYEFLmNHXRkQ3FSvDyK4FRBXFB690ER
# MA0GCSqGSIb3DQEBBQUABIIBAB5C4ihraJ8nLmnRqZjvSuSqb93MvnBJiqa9Q6n9
# gfeOQoNUNOQviAGUbwctoBEFfVPT8zpuAgO459dVQ+pyCL5ARRIxj7+Ml24I2Oq0
# JNqrKX7+d99cA3iengWqeLZEHLf+dIk/fNdfVFtc52OK3tZ7EHGokwPzKBL1D9d3
# QPRuslvvYH6WoDxu0mqGBSUC/mbpUJafXQ99EWlh2R4ruwkqe4b8ai0WJA8tXzQV
# U1wEMJ5XBM+Lw4OK7yR/AjJ0EERQoe1+88fuQw6KY7adEME5udsKE3U8+97c3/CS
# mDxR4xyX3Ro1iZSb35r2nbyjlU1gyANC032flE8FKrvgYWI=
# SIG # End signature block
