<#
.SYNOPSIS
  Create a VNET
.DESCRIPTION
  Create a VNET
.EXAMPLE
  The following example will create an affinity group wide virtual network "contosovnet" in affinity group "contoso-afgroup" with address space 192.168.0.0/20, 
  and a subnet named "Subnet-1"  with address space 192.168.0.0/23
    CreateNewVNet -VNetName "contosovnet" -AffinityGroupName "contoso-afgroup"

  The following example will create an affinity group wide virtual network "contosovnet" in affinity group "contoso-afgroup" with address space 10.0.0.0/20, 
  and a subnet named "Subnet22"  with address space 10.0.0.0/23
    CreateNewVNet -VNetName "contosovnet" -AffinityGroupName "contoso-afgroup" -AddressSpace "10.0.0.0/20" -SubnetName "Subnet22" -SubnetAddressSpace "10.0.0.0/23"

  The following example will create a region wide virtual network "contosoRegional" in the region "East Asia" with address space 10.0.0.0/20, 
  and a subnet named "Subnet-1"  with address space 10.0.0.0/23
    CreateNewVNet -VNetName "contosoRegional" -Location "East Asia" -AddressSpace "10.0.0.0/20" -SubnetAddressSpace "10.0.0.0/23"
#>
function CreateNewVNet
{
    Param
    (
        # The name of VNet
        [Parameter(Mandatory=$true)]
        [String] $VNetName,
		
        # The affinity group name for a affinity group wide virtual network
        [Parameter(Mandatory=$true, ParameterSetName="AffinityGroupVNet")]
        [String] $AffinityGroupName,

        # The location for a region wide virtual network
        [Parameter(Mandatory=$true, ParameterSetName="RegionalVNet")]
        [String] $Location,

        # The address space for the virtual network
        [Parameter(Mandatory=$false)]
        [String] $AddressSpace = "192.168.0.0/20",

        # Subnet name
        [Parameter(Mandatory=$false)]
        [String] $SubnetName = "Subnet-1",

        # The address space for the subnet
        [Parameter(Mandatory=$false)]
        [String] $SubnetAddressSpace = "192.168.0.0/23"
    )


    if($PsCmdlet.ParameterSetName -eq "AffinityGroupVNet")
    {
        TraceInfo "Creating virtual network $VNetName with Subnet $SubnetName in AffinityGroup $AffinityGroupName"
    }
    else
    {
        TraceInfo "Creating virtual network $VNetName with Subnet $SubnetName in Location $Location"
    }
            
    $maxRetries = 10
    $retry = 0
    while($true)
    {
        $vnetMutex = New-Object System.Threading.Mutex $false,'HPCIaaSUpdateVNetMutex'
        $mutexAcquired = $false
        try
        {
            try
            {
                $mutexAcquired = $vnetMutex.WaitOne()
            }
            catch [System.Threading.AbandonedMutexException]
            {
                $mutexAcquired = $true
            }

            $emptyVNetCfg = '<NetworkConfiguration xmlns="http://schemas.microsoft.com/ServiceHosting/2011/07/NetworkConfiguration">
      <VirtualNetworkConfiguration>
        <Dns />
      </VirtualNetworkConfiguration>
    </NetworkConfiguration>'

            $uniqueID = [Guid]::NewGuid().ToString()
            $vnetCfgFile = "$env:temp\VNetCfg_$uniqueID.xml"

            InvokeAzureCmdWithRetry -Command "Get-AzureVNetConfig -ExportToFile $vnetCfgFile" | Out-Null
            if(Test-Path -Path $vnetCfgFile)
            {
                $xmlDoc = [xml](Get-Content $vnetCfgFile)
            }
            else
            {
                $xmlDoc = [xml] $emptyVNetCfg
            }
            $namespaceUri = $xmlDoc.NetworkConfiguration.NamespaceURI
            $vnets = $xmlDoc.NetworkConfiguration.VirtualNetworkConfiguration.GetElementsByTagName("VirtualNetworkSites") | Select-Object -First(1)
            if($vnets -eq $null)
            {
               $vnets = $xmlDoc.CreateElement("VirtualNetworkSites", $namespaceUri)
               $xmlDoc.NetworkConfiguration.VirtualNetworkConfiguration.AppendChild($vnets) | Out-Null
            }

            $vnetSite = $vnets.GetElementsByTagName("VirtualNetworkSite") | Where-Object {$_.name -eq $VNetName}
            if($vnetSite -ne $null)
            {
                if($vnetSite.HasAttribute("Location"))
                {
                    if(($PsCmdlet.ParameterSetName -ne "RegionalVNet") -or ($vnetSite.Location -ne $Location))
                    {
                        throw "The virtual network $VNetName already exists but not in the region $Location."
                    }
                }
                elseif($vnetSite.HasAttribute("AffinityGroup"))
                {
                    if(($PsCmdlet.ParameterSetName -ne "AffinityGroupVNet") -or ($vnetSite.AffinityGroup -ne $AffinityGroupName))
                    {
                        throw "The virtual network $VNetName already exists but not in the affinity group $AffinityGroupName."
                    }
                }

                if(($vnetSite.GetElementsByTagName("Subnet") | Where-Object {$_.name -eq $SubnetName}) -eq $null)
                {
                    throw "The virtual network $VNetName already exists but not subnet $SubnetName in it."
                }
                else
                {
                    TraceInfo "The virtual network $VNetName already exists."
                    return
                }
            }

            # Create the new VNet
            $vnetSite = $xmlDoc.CreateElement("VirtualNetworkSite", $namespaceUri)
            $vnetSite.SetAttribute("name", $VNetName)
            if($PsCmdlet.ParameterSetName -eq "AffinityGroupVNet")
            {
                $vnetSite.SetAttribute("AffinityGroup", $AffinityGroupName)
            }
            else
            {
                $vnetSite.SetAttribute("Location", $Location)
            }

            $vnets.AppendChild($vnetSite) | Out-Null
            $addrspace = $xmlDoc.CreateElement("AddressSpace", $namespaceUri)
            $vnetSite.AppendChild($addrspace) | Out-Null
            $addrPrefix = $xmlDoc.CreateElement("AddressPrefix", $namespaceUri)
            $addrPrefix.InnerText = $AddressSpace
            $addrspace.AppendChild($addrPrefix) | Out-Null
            $subnets = $xmlDoc.CreateElement("Subnets", $namespaceUri)
            $vnetSite.AppendChild($subnets) | Out-Null
            $subnet1 = $xmlDoc.CreateElement("Subnet", $namespaceUri)
            $subnet1.SetAttribute("name", $SubnetName)
            $subnets.AppendChild($subnet1) | Out-Null
            $addrPrefix1 = $xmlDoc.CreateElement("AddressPrefix", $namespaceUri);
            $addrPrefix1.InnerText = $SubnetAddressSpace
            $subnet1.AppendChild($addrPrefix1) | Out-Null
            $xmlDoc.Save($vnetCfgFile)

            try
            {
                Set-AzureVNetConfig -ConfigurationPath $vnetCfgFile -ErrorAction Stop
                TraceInfo "Successfully created the virtual network $VNetName with subnet $SubnetName"
                return
            }
            catch
            {
                $errMsg = "Failed to Create virtual network $VNetName : " + ($_ | Out-String)
                if($retry -lt $maxRetries)
                {
                    TraceWarning $errMsg
                }
                else
                {
                    throw $errMsg
                }
            }
        }
        finally
        {
            if($mutexAcquired)
            {
                $vnetMutex.ReleaseMutex()
            }
        }
                    
        $interval = Get-Random -minimum 20 -maximum 40
        TraceInfo "Retry to create virtual network $VNetName after $interval seconds"
        Start-Sleep -Seconds $interval
        $retry++
    }
}

<#
.SYNOPSIS
   Modifies the virtual network configuration xml file to include a DNS service reference.
.DESCRIPTION
   This a small utility that programmatically modifies the vnet configuration file to add a DNS server
   then adds the DNS server's reference to the specified virtual network site.
.EXAMPLE
    ConfigVNetDnsServer -DnsServerName "contoso" -IpAddress "10.0.0.4" -VNetName "dcvnet"
.INPUTS
   None
.OUTPUTS
   None
#>
function ConfigVNetDnsServer
{
   param
    (
        [Parameter(Mandatory = $true)]
        [String] $DnsServerName,

        [Parameter(Mandatory = $true)]
        [String] $IpAddress,

        [Parameter(Mandatory = $true)]
        [String] $VNetName
    )
    
    TraceInfo "Configuring a DNS server for virtual network $VNetName : Name=$DnsServerName, IP=$IpAddress"
    $maxRetries = 10
    $retry = 0
    while($true)
    {
        $vnetMutex = New-Object System.Threading.Mutex $false,'HPCIaaSUpdateVNetMutex'
        $mutexAcquired = $false
        try
        {
            try
            {
                $mutexAcquired = $vnetMutex.WaitOne()
            }
            catch [System.Threading.AbandonedMutexException]
            {
                $mutexAcquired = $true
            }

            $vNet = Get-AzureVNetSiteWithRetry -VNetName $VNetName
            if ($vNet -eq $null)
            {
                throw "The virtual network $VNetName does not exist. Cannot add an DNS server for it."
            }
            $uniqueID = [Guid]::NewGuid().ToString()
            $vnetFilePath = "$env:temp\VNetCfg_$uniqueID.xml"
            InvokeAzureCmdWithRetry -Command "Get-AzureVNetConfig -ExportToFile $vnetFilePath" | Out-Null
            if (-not (Test-Path $vNetFilePath))
            {
                throw "Cannot retrieve the virtual network configuration file."
            }

            $xml = [Xml] (Get-Content $vnetFilePath)
            $nsUri = $xml.NetworkConfiguration.NamespaceURI

            $dnsElement = $xml.NetworkConfiguration.VirtualNetworkConfiguration.GetElementsByTagName("Dns") | Select-Object -First(1)
            if ($dnsElement -eq $null)
            {
                $dnsElement = $xml.CreateElement("Dns", $nsUri)
                $xml.NetworkConfiguration.VirtualNetworkConfiguration.AppendChild($dnsElement) | Out-Null
            }

            $dnsServersElement = $dnsElement.GetElementsByTagName("DnsServers") | Select-Object -First(1)
            if ($dnsServersElement -eq $null)
            {
                $dnsServersElement = $xml.CreateElement("DnsServers", $nsUri)
                $dnsElement.AppendChild($dnsServersElement) | Out-Null
            }

            $dnsSrvIPUpdated = $true
            $dnsServerElement = $dnsServersElement.GetElementsByTagName("DnsServer") | Where-Object {$_.name -eq $DnsServerName}
            if ($dnsServerElement -ne $null)
            {
                if($dnsServerElement.IpAddress -eq $IpAddress)
                {
                    $dnsSrvIPUpdated = $false
                }
                else
                {
                    $dnsServerElement.IpAddress = $IpAddress
                }
            }
            else
            {
                $dnsServerElement = $xml.CreateElement("DnsServer", $nsUri)
                $dnsServerElement.SetAttribute("name", $DnsServerName) | Out-Null
                $dnsServerElement.SetAttribute("IPAddress", $IpAddress) | Out-Null
                $dnsServersElement.AppendChild($dnsServerElement) | Out-Null
            }

            # Now set the DnsReference for the network site
            $vnetSiteElement = $xml.NetworkConfiguration.VirtualNetworkConfiguration.VirtualNetworkSites.GetElementsByTagName("VirtualNetworkSite") | Where-Object {$_.name -eq $VNetName}

            if ($vnetSiteElement -eq $null)
            {
                throw "Cannot find the virtual network $VNetName"
            }

            $dnsServersRefElement = $vnetSiteElement.GetElementsByTagName("DnsServersRef") | Select-Object -First(1)
            if ($dnsServersRefElement -eq $null)
            {
                $dnsServersRefElement = $xml.CreateElement("DnsServersRef", $nsUri)
                $vnetSiteElement.AppendChild($dnsServersRefElement) | Out-Null
            }
    
            $dnsServerRefElement = $dnsServersRefElement.GetElementsByTagName("DnsServerRef") | Where-Object {$_.name -eq $DnsServerName}
            if($dnsServerRefElement -eq $null)
            {
                $dnsServerRefElement = $xml.CreateElement("DnsServerRef", $nsUri)        
                $dnsServerRefElement.SetAttribute("name", $DnsServerName) | Out-Null
                $dnsServersRefElement.AppendChild($dnsServerRefElement) | Out-Null
            }
            else
            {
                if(-not $dnsSrvIPUpdated)
                {
                    TraceInfo "No update needed for virtual network $VNetName"
                    return
                }
            }

            $xml.Save($vnetFilePath)
            try
            {
                Set-AzureVNetConfig -ConfigurationPath $vnetFilePath -ErrorAction Stop
                TraceInfo "Successfully configured the DNS server for virtual network $VNetName"
                return
            }
            catch
            {
                $errMsg = "Failed to configure a DNS server for virtual network $VNetName :" + ($_ | Out-String)
                if($retry -lt $maxRetries)
                {
                    TraceWarning $errMsg
                }
                else
                {
                    throw $errMsg
                }
            }
        }
        finally
        {
            if($mutexAcquired)
            {
                $vnetMutex.ReleaseMutex()
            }
        }
        
        $interval = Get-Random -minimum 20 -maximum 40
        TraceInfo "Retry to configure the DNS server for virtual network $VNetName after $interval seconds"
        Start-Sleep -Seconds $interval
        $retry++
    }
}

function GetAvailableVNetIP
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $VNetName, 

        [Parameter(Mandatory=$true)]
        [String] $SubnetName       
    )

    $vnet = Get-AzureVNetSiteWithRetry -VNetName $VNetName
    if($vnet -eq $null)
    {
        throw "The virtual network $VNetName doesn't exist"
    }
    $subnet = $vnet.Subnets | Where-Object {$_.Name -eq $SubnetName}
    if($subnet -eq $null)
    {
        throw "The subnet $SubnetName doesn't exist in virtual network $VNetName"
    }
    
    $testIP = [String]$subnet.AddressPrefix
    $slashIndex = $testIP.IndexOf('/')
    $testIP = $testIP.Substring(0, $slashIndex)
    $testRet = InvokeAzureCmdWithRetry -Command "Test-AzureStaticVNetIP -VNetName $VNetName -IPAddress $testIP"
    if($testRet -eq $null)
    {
        throw "Failed to get available IP address in virtual network $VNetName"
    }
    if($testRet.IsAvailable)
    {
        return $testIP
    }
    else
    {
        return $testRet.AvailableAddresses[0]
    }
}


function CreateVMForDC
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [String] $VMName, 
        
        [Parameter(Mandatory=$true)]
        [String] $VMSize, 

        [Parameter(Mandatory=$true)]
        [String] $ImageName, 
        
        [Parameter(Mandatory=$true)]
        [String] $VNetName, 

        [Parameter(Mandatory=$true)]
        [String] $SubnetName,         

        [Parameter(Mandatory=$true)]
        [String] $DeploymentLabel,  

        [Parameter(Mandatory=$true)]
        [PSCredential] $AdminCredential
    )

    try
    {

        TraceInfo "Creating VM $VMName in cloud service $ServiceName for domain controller. Size:$VMSize Image:$ImageName VNet:$VNetName. Subnet:$SubnetName"

        if (VMExists -Name $VMName -ServiceName $ServiceName)
        {
            TraceWarning "The VM $VMName already exists in cloud service $ServiceName"
            return
        }

        $staticIP = GetAvailableVNetIP -VNetName $VNetName -SubnetName $SubnetName
        $dcVm = New-AzureVMConfig -Name $VMName -InstanceSize $VMSize -HostCaching ReadOnly -ImageName $ImageName | `
            Add-AzureProvisioningConfig -Windows -AdminUsername $AdminCredential.UserName -Password $AdminCredential.GetNetworkCredential().Password  | `
            Set-AzureVMBGInfoExtension | `
            Set-AzureSubnet -SubnetNames $SubnetName | `
            Set-AzureStaticVNetIP -IPAddress $staticIP

        New-AzureVMWithRetry -VM $dcVm -ServiceName $ServiceName -VNetName $VNetName -DeploymentLabel $DeploymentLabel

        WaitForVMReady -VMName $VMName -ServiceName $ServiceName
    }
    catch
    {
        TraceError "Failed to create virtual machine for domain controller"
        throw
    }
    
    ConfigVNetDnsServer -DnsServerName $VMName -IpAddress $staticIP -VNetName $VNetName
}

function PromoteDC
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [String] $VMName, 
        
        [Parameter(Mandatory=$true)]
        [String] $DomainFQDN, 
        
        [Parameter(Mandatory=$true)]
        [PSCredential] $AdminCredential
    )
    TraceInfo "Promoting VM $VMName as a domain controller, the domain FQDN is $DomainFQDN."
    $scriptBlock = { 
        param ([String]$Fqdn, [String]$Psw)
        
        # 0 for Standalone Workstation, 1 for Member Workstation, 2 for Standalone Server, 3 for Member Server, 4 for Backup Domain Controller, 5 for Primary Domain Controller
        $domainRole = (Get-WmiObject Win32_ComputerSystem).DomainRole
        if($domainRole -eq 5)
        {
            WriteLog "$env:COMPUTERNAME was already a domain controller."
            return
        }
        
        WriteLog "Installing windows features AD-Domain-Services and GPMC."
        Install-WindowsFeature -Name AD-Domain-Services,GPMC -IncludeManagementTools
        
        WriteLog "Start to install domain forest $Fqdn"
        # During the domain forest installation, the PSSession may be broken occasionally, which makes the current powershell process gets killed,
        # We create a runaway process to install the domain forest by "Invoke-WmiMethod" cmdlet, so that the installation process can survive in this case.
        # Because the password can contain special characters, which may happen to be escape character (e.g. '\') or meaningful character (e.g. '"'),
        # it will result in cmd line parsing error, so we will convert the password into ascii codes, then convert back in the powershell command line.
        $pswAscii = $Psw.ToCharArray() | %{[Convert]::ToUInt32($_)} | %{$_.ToString()}
        $pshCmd = 'Import-Module ADDSDeployment; $pswAscii = ''{0}''.Split('' '');' -f ($pswAscii -join ' ')
        $pshCmd += '$psw = -join ($pswAscii | %{[Convert]::ToChar([Uint32]$_)});$secpsw = ConvertTo-SecureString -String $psw -AsPlainText -Force;'
        $pshCmd += 'Install-ADDSForest -DomainName {0} -InstallDNS -SafeModeAdministratorPassword $secpsw -Force' -f $Fqdn

        $pobj = Invoke-WmiMethod -Path win32_process -Name Create -ArgumentList "Powershell.exe -Command ""$pshCmd"""
        $p = $null
        if($pobj.ReturnValue -eq 0)
        {
            $p = Get-Process -Id $pobj.ProcessId -ErrorAction SilentlyContinue
        }

        if($p -ne $null)
        {
            WriteLog "Installing domain forest $Fqdn"
            $p.WaitForExit()
        }
        else
        {
            throw "Failed to promote $env:COMPUTERNAME to a domain controller"
        }
    }

    # Error code 995 is expected here: The I/O operation has been aborted because of either a thread exit or an application request. 
    # Error code 5 is sometimes thrown when the remote call execution succeeded: Processing data for a remote command failed with the following error message: Access is denied. 
    # No particular action is taken for Error code 5, just retry as we do for other error codes, the script block already handles re-entry.
    RemotePshCall -VMName $VMName `
                  -ServiceName $ServiceName `
                  -Credential $AdminCredential `
                  -ScriptBlock  $scriptBlock `
                  -ArgumentList @($DomainFQDN, $AdminCredential.GetNetworkCredential().Password) `
                  -IgnoreErrorCodes 995 | Out-Null

    # Sleep 60 seconds to wait DC restart
    Start-Sleep -Seconds 60
    WaitForVMReady -VMName $VMName -ServiceName $ServiceName

    # Try to remove the DNS forwarders from the domain controller which causes DNS resolving issue
    try
    {
        RemotePshCall -VMName $VMName -ServiceName $ServiceName -Credential $AdminCredential -RetryTimes 10 -ScriptBlock {
            $forwarders = (Get-DnsServerForwarder).IPAddress
            if($forwarders -ne $null)
            {
                WriteLog "Removing DNS forwarders from the domain controller: $forwarders"
                Remove-DnsServerForwarder -IPAddress $forwarders -Force
            }
        } | Out-Null 
    }
    catch
    {
        TraceWarning "Failed to remove DNS forwarders in the domain controller: $_"
    }
}

function VMJoinToDomain
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [String] $VMName,
        
        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred, 
        
        [Parameter(Mandatory=$true)]
        [PSCredential] $DomainUserCred,

        [Parameter(Mandatory=$true)]
        [String] $DomainFQDN
    )
 
    $scriptblock = {
        param($cred, $domainFQDN)
        
        $maxRetries = 10
        $retry = 0
        while($true)
        {
            # -Force is added from PowerShell 3.0 on
            if($PSVersionTable["PSVersion"] -lt "3.0")
            {
                Add-Computer -DomainName $domainFQDN -Credential $cred -ErrorAction SilentlyContinue
            }
            else
            {
                Add-Computer -DomainName $domainFQDN -Credential $cred -Force -ErrorAction SilentlyContinue
            }
            
            $domainRole = (Get-WmiObject Win32_ComputerSystem).DomainRole
            if($domainRole -ge 3)
            {
                WriteLog "$env:COMPUTERNAME successfully joined domain $DomainFQDN and will restart."
                break
            }
            
            if($retry -lt $maxRetries)
            {
                WriteLog ("$env:COMPUTERNAME failed to join domain $DomainFQDN : " + $Error[0])
                Start-Sleep -Seconds 10
                $retry++
            }
            else
            {
                throw "$env:COMPUTERNAME failed to join domain $DomainFQDN after $maxRetries retries."
            }
        }

        try
        {
            Write-VolumeCache C -ErrorAction SilentlyContinue
        }
        catch
        {
            # Ignore the error
        }
    }
    
    $tryRestart = $false
    while($true)
    {
        try
        {
            RemotePshCall -ServiceName $ServiceName -VMName $VMName -Credential $LocalAdminCred -ScriptBlock $scriptblock -ArgumentList @($DomainUserCred, $DomainFQDN) | Out-Null
            InvokeAzureCmdWithRetry -Command "Restart-AzureVM -ServiceName $ServiceName -Name $VMName"
            break
        }
        catch [System.Management.Automation.RemoteException]
        {
            if(($_.Exception.Message.Contains("Failed to join domain $DomainFQDN")) -and (-not $tryRestart))
            {
                $tryRestart = $true
                TraceWarning "$VMName failed to join domain $DomainFQDN, will restart and try again."
                InvokeAzureCmdWithRetry -Command "Restart-AzureVM -ServiceName $ServiceName -Name $VMName"
                Start-Sleep -Seconds 20
                WaitForVMReady -VMName $VMName -ServiceName $ServiceName                
            }
            else
            {
                throw
            }
        }
    }
    
    Start-Sleep -Seconds 20
    WaitForVMReady -VMName $VMName -ServiceName $ServiceName
}

function AddADUserToLocalAdmin
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred,

        # Full user name, for example "consto\john"
        [Parameter(Mandatory=$true)]
        [String] $ADUserName
    )
    RemotePshCall -ServiceName $ServiceName -VMName $VMName -Credential $LocalAdminCred -ScriptBlock {
        param([String]$aduser)

        # 0 for Standalone Workstation, 1 for Member Workstation, 2 for Standalone Server, 3 for Member Server, 4 for Backup Domain Controller, 5 for Primary Domain Controller
        $domainRole = (Get-WmiObject Win32_ComputerSystem).DomainRole
        if($domainRole -lt 3)
        {
            throw "$env:COMPUTERNAME is not domain joined"
        }
        
        $maxRetries = 3
        $retry = 0
        while($true)
        {
            try
            {
                WriteLog "Adding user $aduser to local administrators group on $env:COMPUTERNAME"
                $adminGroup = [ADSI]("WinNT://$env:COMPUTERNAME/administrators, group")
                $usr = $aduser.Replace("\","/")
                $adminGroup.Add("WinNT://$usr, user")
                WriteLog "The user $aduser successfully added to local administrators group on $env:COMPUTERNAME"
                break
            }
            catch [System.Management.Automation.MethodInvocationException]
            {
                if($retry -lt $maxRetries)
                {
                    WriteLog ("Failed to add user $aduser to local administrators group, retrying after 5 seconds: "+ $_)
                    Start-Sleep -Seconds 5
                    $retry++
                }
                else
                {
                    throw ("Failed to add user $aduser to local administrators group after $maxRetries retries: " + $_)
                }
            }
        }

    } -ArgumentList $ADUserName | Out-Null
}

function ConfigRemoteDBServer
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [PSCredential] $DBAdminCred,

        # Full user name, for example "consto\john"
        [Parameter(Mandatory=$true)]
        [String] $HpcSetupUserName,

        # Full headnode machine account name, for example "consto\myheadnode$"
        [Parameter(Mandatory=$true)]
        [String] $HNMachineAccount
    )

    TraceInfo "Configuring remote DB server $VMName : SetupUser=$HpcSetupUserName HeadNode=$HNMachineAccount"
    try
    {
        $dbcreateSql = [IO.File]::ReadAllBytes("$env:HPCPSScriptRoot\HPCIaaSCreateDatabase.sql")
        $dbpermissionSql = [IO.File]::ReadAllBytes("$env:HPCPSScriptRoot\AddDbUserForHpcService.sql")
    }
    catch
    {
        throw ("Failed to load SQL scripts to configure remote DB server:" + $_)
    }

    RemotePshCall -ServiceName $ServiceName -VMName $VMName -Credential $DBAdminCred -ScriptBlock {
        param([string]$aduser, [String]$hnMachine, [Byte[]]$dbCreateScript, [Byte[]]$dbPermissionScript)

        # 0 for Standalone Workstation, 1 for Member Workstation, 2 for Standalone Server, 3 for Member Server, 4 for Backup Domain Controller, 5 for Primary Domain Controller
        $domainRole = (Get-WmiObject Win32_ComputerSystem).DomainRole
        if($domainRole -lt 3)
        {
            throw "The DB Server $env:COMPUTERNAME is not domain joined"
        }

        # Try to update IP address to DNS server
        Register-DnsClient -ErrorAction SilentlyContinue

        # Get the data disk ready
        $maxRetries = 10
        $retry = 0
        while($true)
        {
            try
            {
                $disk = Get-Disk -Number 2 -ErrorAction Stop
                if($disk.PartitionStyle -eq "RAW")
                {
                    WriteLog "Initializing the data disk."
                    Initialize-Disk -Number 2 -PartitionStyle MBR -ErrorAction Stop | Out-Null
                }

                $partition = Get-Partition -DiskNumber 2 -ErrorAction SilentlyContinue
                if($partition -eq $null)
                {
                    WriteLog "Creating new partition with drive letter F."
                    New-Partition -DiskNumber 2 -UseMaximumSize -IsActive -DriveLetter F -ErrorAction Stop | Out-Null
                }

                WriteLog "Formating volume F to NTFS file system."
                Format-Volume -DriveLetter F -FileSystem NTFS -NewFileSystemLabel "HPC Data" -Force:$true -confirm:$false -ErrorAction Stop | Out-Null

                WriteLog "Creating HPCData directory."
                New-Item -ItemType directory -Path F:\ -Name HPCData -Force -ErrorAction Stop | Out-Null
                break
            }
            catch
            {
                if($retry -lt $maxRetries)
                {
                    WriteLog "Failed to get data disk ready, will retry after 10 seconds: $_"
                    Start-Sleep -Seconds 10
                    $retry++
                }
                else
                {
                    throw
                }
            }
        }

        #Configure firewall to allow incoming SQL connections
        WriteLog "Configuring firewall for SQL Server"
        New-NetFirewallRule -DisplayName "Allow SQL Server TCP In" -Direction Inbound -Action Allow -LocalPort 1433 -Protocol TCP -Profile Domain | Out-Null

        # Add domain user as 'sysadmin' of SQL server
        Set-ExecutionPolicy -ExecutionPolicy RemoteSigned -Force
        WriteLog "Connecting to SQL server instance"
        $sqlConnection = New-Object -TypeName System.Data.SqlClient.SqlConnection
        $sqlConnection.ConnectionString = "Data Source=localhost;Initial Catalog=master;Integrated Security=True;Connection Timeout=90"

        $maxRetries = 50
        $retry = 0
        while($true)
        {
            try 
            {
               $sqlConnection.Open()
               WriteLog "Connected to SQL server instance"
               break
            }
            catch
            {
                if($retry -lt $maxRetries)
                {
                    WriteLog "Failed to connect SQL server instance, retry after 20 seconds: $_"
                    Start-Sleep -Seconds 20
                    $retry++
                }
                else
                {
                    throw "Failed to connect SQL Server instance after $maxRetries retries: $_"
                }
            }
            finally
            {
                $sqlConnection.Close()
            }
        }


        WriteLog "Adding user $aduser as a sysadmin of SQL server."
        $maxRetries = 3
        $retry = 0
        while($true)
        {
            $Command = $null
            try
            {
                $sqlConnection.Open()
                $Command = $sqlConnection.CreateCommand()
                $Command.CommandTimeout = 120
                $Command.CommandText = "if not exists (select * from sys.server_principals where name = '$aduser') create login [$aduser] from windows"
                $Command.ExecuteNonQuery() | Out-Null
                $Command.CommandText = "if is_srvrolemember('sysadmin', '$aduser')=0 exec sp_addsrvrolemember '$aduser', 'sysadmin'"
                $Command.ExecuteNonQuery() | Out-Null

                WriteLog "Successfully added user $aduser as a sysadmin of SQL server"
                break
            }
            catch
            {
                WriteLog "Failed to add user $aduser as a sysadmin of SQL server, retrying: $_"
                if($retry -lt $maxRetries)
                {
                    $retry++
                    Start-Sleep -Seconds 2
                }
                else
                {
                    throw
                }
            }
            finally
            {
                if($Command -ne $null)
                {
                    $Command.Dispose()
                    $Command = $null
                }

                $sqlConnection.Close()
            }
        }

        [IO.File]::WriteAllBytes("$env:temp\HPCIaaSCreateDatabase.sql", $dbCreateScript)
        [IO.File]::WriteAllBytes("$env:temp\AddDbUserForHpcService.sql", $dbPermissionScript)
        try
        {
            WriteLog "Creating HPC databases"
            Import-Module "sqlps" -DisableNameChecking -Force
            Invoke-Sqlcmd -ServerInstance localhost -InputFile "$env:temp\HPCIaaSCreateDatabase.sql" -QueryTimeout 0 -Variable "HPCDataPath='F:\HPCData'"
            WriteLog "HPC Databases created"
                        
            WriteLog "Configuring the permissions for HPC databases"
            Invoke-Sqlcmd -ServerInstance localhost -Database "HPCManagement" -InputFile "$env:temp\AddDbUserForHpcService.sql" -Variable "TargetAccount=$hnMachine"
            Invoke-Sqlcmd -ServerInstance localhost -Database "HPCScheduler" -InputFile "$env:temp\AddDbUserForHpcService.sql" -Variable "TargetAccount=$hnMachine"
            Invoke-Sqlcmd -ServerInstance localhost -Database "HPCMonitoring" -InputFile "$env:temp\AddDbUserForHpcService.sql" -Variable "TargetAccount=$hnMachine"
            Invoke-Sqlcmd -ServerInstance localhost -Database "HPCReporting" -InputFile "$env:temp\AddDbUserForHpcService.sql" -Variable "TargetAccount=$hnMachine"
            Invoke-Sqlcmd -ServerInstance localhost -Database "HPCDiagnostics" -InputFile "$env:temp\AddDbUserForHpcService.sql" -Variable "TargetAccount=$hnMachine"
            WriteLog "The permissions for HPC databases configured"
        }
        catch
        {
            WriteLog ("Failed to Create HPC databases: " + $_)
            throw
        }

    } -ArgumentList @($HpcSetupUserName,$HNMachineAccount,$dbcreateSql,$dbpermissionSql) | Out-Null
}

function PrepareHeadNode
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [PSCredential] $SetupUserCred,

        [Parameter(Mandatory=$true, ParameterSetName="localdb")]
        [Switch] $LocalDB, 

        [Parameter(Mandatory=$true, ParameterSetName="remotedb")]
        [Switch] $RemoteDB, 

        [Parameter(Mandatory=$true, ParameterSetName="remotedb")]
        [String] $DBServerInstance,

        [Parameter(Mandatory=$true)]
        [String] $SubscriptionId,

        [Parameter(Mandatory=$true)]
        [String] $VNetName,

        [Parameter(Mandatory=$true)]
        [String] $SubnetName,
        
        [Parameter(Mandatory=$true)]
        [String] $Location,  

        [Parameter(Mandatory=$false)]
        [String] $AffinityGroupName = "",

        [Parameter(Mandatory=$false)]
        [String] $PreConfigScirptFile = "",
        
        [Parameter(Mandatory=$false, ParameterSetName="remotedb")]
        [String] $ManagementDB = "HpcManagement",

        [Parameter(Mandatory=$false, ParameterSetName="remotedb")]
        [String] $SchedulerDB = "HpcScheduler",

        [Parameter(Mandatory=$false, ParameterSetName="remotedb")]
        [String] $MonitoringDB = "HpcMonitoring",

        [Parameter(Mandatory=$false, ParameterSetName="remotedb")]
        [String] $ReportingDB = "HpcReporting",

        [Parameter(Mandatory=$false, ParameterSetName="remotedb")]
        [String] $DiagnosticsDB = "HpcDiagnostics"
    )


    $deployment = InvokeAzureCmdWithRetry -Command "Get-AzureDeployment -ServiceName $ServiceName -Slot Production"
    if($deployment -eq $null)
    {
        throw "No deployment found in cloud service $ServiceName"
    }

    if($LocalDB.IsPresent)
    {
        $dbArgs = "-DBServerInstance .\COMPUTECLUSTER"
    }
    else
    {
        $dbArgs = "-RemoteDB -DBServerInstance $DBServerInstance -ManagementDB $ManagementDB -SchedulerDB $SchedulerDB -MonitoringDB $MonitoringDB -ReportingDB $ReportingDB -DiagnosticsDB $DiagnosticsDB"
    }

    if((-not [String]::IsNullOrEmpty($PreConfigScirptFile)) -and (Test-Path -Path $PreConfigScirptFile -PathType Leaf))
    {
        $preConfig = [IO.File]::ReadAllBytes($PreConfigScirptFile)
    }
    else
    {
        $preConfig = $null
    }

    RemotePshCall -ServiceName $ServiceName -VMName $VMName -Credential $SetupUserCred -ScriptBlock {
        Param ([String]$depid, [String]$subsid, [String]$vnet, [String]$subnet, [String]$agname, [String]$location, [String]$user, [String]$psw, [String]$dbArgs, [Byte[]]$PreConfigContent)

        # Try to clear local dns cache and update local IP to dns server
        Clear-DnsClientCache -ErrorAction SilentlyContinue
        Register-DnsClient -ErrorAction SilentlyContinue

        # Try to execute HPCHNPreConfig.ps1, but will continue whatever the execution result is
        if(($PreConfigContent -ne $null) -and ($PreConfigContent.Count -gt 0))
        {
            $preConfigScriptFile = "$env:temp\HPCHNPreConfig.ps1"
            try
            {
                WriteLog "Creating $preConfigScriptFile"
                [IO.File]::WriteAllBytes($preConfigScriptFile, $PreConfigContent)
                WriteLog "$preConfigScriptFile created"
                WriteLog "Executing $preConfigScriptFile"
                Start-Process -FilePath "PowerShell.exe" -ArgumentList "-ExecutionPolicy Unrestricted -File `"$preConfigScriptFile`"" -NoNewWindow -Wait
                WriteLog "$preConfigScriptFile executed"
            }
            catch
            {
                WriteLog "Failed to run $preConfigScriptFile : $_"
            }
        }

        New-Item -Path HKLM:\SOFTWARE\Microsoft\HPC -Name IaaSInfo -Force | Out-Null
        Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC\IaaSInfo -Name SubscriptionId -Value $subsid
        Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC\IaaSInfo -Name DeploymentId -Value $depid
        Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC\IaaSInfo -Name VNet -Value $vnet
        Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC\IaaSInfo -Name Subnet -Value $subnet
        Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC\IaaSInfo -Name AffinityGroup -Value $agname
        Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC\IaaSInfo -Name Location -Value $location
        WriteLog "The information needed for in-box management scripts succcessfully configured."

        Import-Module ScheduledTasks
        $HNPreparePsFile = [System.IO.Path]::Combine($env:CCP_HOME, "Bin\HPCHNPrepare.ps1")
        $action = New-ScheduledTaskAction -Execute "PowerShell.exe" -Argument "-ExecutionPolicy Unrestricted -Command `"& '$HNPreparePsFile' $dbArgs`""
        Register-ScheduledTask -TaskName "HPCPrepare" -Action $action -User $user -Password $psw -RunLevel Highest | Out-Null
        if(-not $?)
        {
            throw "Failed to schedule HPC head node preparation task"
        }
        WriteLog "HPC head node preparation task scheduled"

        #Manually trigger the task
        Start-ScheduledTask -TaskName "HPCPrepare"
        if(-not $?)
        {
            throw "Failed to start HPC head node preparation task"
        }
        WriteLog "Running HPC head node preparation task"

        Start-Sleep -Milliseconds 500
        $taskSucceeded = $false
        do
        {
            $taskState = (Get-ScheduledTask -TaskName "HPCPrepare").State
            if($taskState -eq "Ready")
            {
                $taskInfo = Get-ScheduledTaskInfo -TaskName "HPCPrepare"
                if($taskInfo.LastRunTime -eq $null)
                {
                    Start-ScheduledTask -TaskName "HPCPrepare"
                }
                else
                {
                    if($taskInfo.LastTaskResult -eq 0)
                    {
                        $taskSucceeded = $true
                        break
                    }
                    else
                    {
                        WriteLog ("The HPC head node preparation task failed:" + $taskInfo.LastTaskResult)
                        break
                    }
                }
            }
            elseif($taskState -ne "Queued" -and $taskState -ne "Running")
            {
                WriteLog "The HPC head node preparation task entered into unexpected state: $taskState"
                break
            }

            Start-Sleep -Seconds 2
        } while ($true)
        
        if($taskSucceeded)
        {
            WriteLog "Checking the HPC services status in HPC head node ..."
            $HNServiceList = @("HpcSdm", "HpcManagement", "HpcReporting", "HpcMonitoringClient", "HpcNodeManager", "msmpi", "HpcBroker", `
                "HpcDiagnostics", "HpcScheduler", "HpcMonitoringServer", "HpcSession", "HpcSoaDiagMon")
            foreach($svcname in $HNServiceList)
            {
                $service = Get-Service -Name $svcname -ErrorAction SilentlyContinue
                if($service -eq $null)
                {
                    WriteLog "Service $svcname not found"
                    $taskSucceeded = $false
                }
                elseif($service.Status -eq "Running")
                {
                    WriteLog "Service $svcname is running"
                }
                else
                {
                    WriteLog "Service $svcname is in $($service.Status) status"
                    $taskSucceeded = $false
                }
            }
        }

        Unregister-ScheduledTask -TaskName "HPCPrepare" -Confirm:$false

        if($taskSucceeded)
        {
            WriteLog "Succeeded to prepare HPC head node"
        }
        else
        {
            WriteLog "Failed to prepare HPC head node"
            if(Test-Path -Path "$env:windir\Temp\HPCHeadNodePrepare.log" -PathType Leaf)
            {
                WriteLog "The HPC head node preparation logs as below:"
                Get-Content -Path "$env:windir\Temp\HPCHeadNodePrepare.log" | Write-Verbose -Verbose
            }

            throw "Failed to prepare HPC head node"
        }
    } -ArgumentList @($deployment.DeploymentId, $SubscriptionId, $VNetName, $SubnetName, $AffinityGroupName, $Location, `
        $SetupUserCred.UserName, $SetupUserCred.GetNetworkCredential().Password, $dbArgs, $preConfig) | Out-Null
}


function ConfigHeadNode
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [PSCredential] $SetupUserCred,

        [Parameter(Mandatory=$true)]
        [String] $NodeNamePattern,

        [Parameter(Mandatory=$false)]
        [Switch] $EnableRESTAPI,

        [Parameter(Mandatory=$false)]
        [Switch] $EnableWebPortal,

        [Parameter(Mandatory=$false)]
        [String] $PostConfigScirptFile = ""
    )

    TraceInfo "Configuring the HeadNode $VMName"
    $connUri = InvokeAzureCmdWithRetry -Command "Get-AzureWinRMUri -ServiceName $ServiceName -Name $VMName" -RetryTimes 5
    if($connUri -eq $null)
    {
        throw "Failed to retrieve the WinRM https listener URI of the virtual machine $VMName in cloud service $ServiceName."
    }

    $postConfig = $null
    if((-not [String]::IsNullOrEmpty($PostConfigScirptFile)) -and (Test-Path -Path $PostConfigScirptFile -PathType Leaf))
    {
        $postConfig = [IO.File]::ReadAllBytes($PostConfigScirptFile)
    }

    $ret = RemotePshCall -ServiceName $ServiceName -VMName $VMName -Credential $SetupUserCred -ScriptBlock {
        param($creds, $nodenaming, $dnsHostName, $restAPI, $webPortal, [Byte[]]$postConfigContent)

        Add-PSSnapin Microsoft.HPC
        # setting network topology to 5 (enterprise)
        WriteLog "Setting network topologogy..."
        $nics = @(Get-WmiObject win32_networkadapterconfiguration -filter "IPEnabled='true' AND DHCPEnabled='true'")
        if ($nics.Count -ne 1)
        {
            throw "Cannot find a suitable network adapter for enterprise topology"
        }

        try
        {
            Set-HpcNetwork -Topology "Enterprise" -Enterprise $nics[0].Description -EnterpriseFirewall $true -ErrorAction SilentlyContinue | Out-Null
            $topo = Get-HpcNetworkTopology -ErrorAction SilentlyContinue
            WriteLog "Network topology is set to $topo"

            # Set installation credentials
            Set-HpcClusterProperty -InstallCredential $creds | Out-Null
            $hpccred = Get-HpcClusterProperty -InstallCredential
            WriteLog ("Installation Credentials set to " + $hpccred.Value)

		    # set node naming series
            Set-HpcClusterProperty -NodeNamingSeries $nodenaming | Out-Null
            WriteLog "Node naming series set to $nodenaming."
        
            # Create a default compute node template
            New-HpcNodeTemplate -Name "Default ComputeNode Template" -Description "This is the default compute node template" | Out-Null
            WriteLog "'Default ComputeNode Template' created."

            $cerContent = $null
            if($restAPI -or $webPortal)
            {
                WriteLog "Generating a self-signed certificate for the HPC web service ..."
                $hpcBinPath = [System.IO.Path]::Combine($env:CCP_HOME, "Bin")
                $thumbprint = . $hpcBinPath\New-HpcCert.ps1 -MachineName $dnsHostName -SelfSigned
                WriteLog "A self-signed certificate $thumbprint was created and installed"

                if($webPortal)
                {
                    WriteLog "Enabling HPC Pack web portal ..."
                    . $hpcBinPath\Set-HPCWebComponents.ps1 -Service Portal -enable -Certificate $thumbprint | Out-Null
                    WriteLog "HPC Pack web portal enabled."
                }

                if($restAPI)
                {
                    WriteLog "Starting HPC web service ..."
                    Set-Service -Name "HpcWebService" -StartupType Automatic | Out-Null
                    Start-Service -Name "HpcWebService" | Out-Null
                    WriteLog "HPC web service started."

                    WriteLog "Enabling HPC Pack REST API ..."
                    . $hpcBinPath\Set-HPCWebComponents.ps1 -Service REST -enable -Certificate $thumbprint | Out-Null
                    WriteLog "HPC Pack REST API enabled."

                    WriteLog "Restarting HPCScheduler service ..."
                    Restart-Service -Name "HpcScheduler" -Force | Out-Null
                    WriteLog "HPCScheduler service restarted."
                }

                $cert = Get-ChildItem -Path Cert:\LocalMachine\My\$thumbprint
                $cerFile = "$env:TEMP\hpcwebcomponent_{0}.cer" -f (Get-Random)
                Export-Certificate -Cert $cert -FilePath $cerFile | Out-Null
                $cerContent = [IO.File]::ReadAllBytes($cerFile)
                WriteLog "The certificate file with public key was exported: $thumbprint"
                Remove-Item $cerFile -Force -ErrorAction SilentlyContinue
            }

            # Run HPCHNPreConfig.ps1 in a separate process if it exists, will not wait for its completion.
            if(($postConfigContent -ne $null) -and ($postConfigContent.Count -gt 0))
            {
                $postConfigScriptFile = "$env:temp\HPCHNPostConfig.ps1"
                try
                {
                    WriteLog "Creating $postConfigScriptFile"
                    [IO.File]::WriteAllBytes($postConfigScriptFile, $postConfigContent)
                    WriteLog "$postConfigScriptFile created"
                    $pobj = Invoke-WmiMethod -Path win32_process -Name Create -ArgumentList "PowerShell.exe -ExecutionPolicy Unrestricted -File `"$postConfigScriptFile`""
                    if($pobj.ReturnValue -eq 0)
                    {
                        WriteLog "Start to run $postConfigScriptFile."
                    }
                    else
                    {
                        WriteLog "Failed to run $postConfigScriptFile."
                    }
                }
                catch
                {
                    WriteLog "Failed to run $postConfigScriptFile : $_"
                }
            }

            return @{"Certificate"=$cerContent}
        }
        catch
        {
            throw ("Failed to configure head node $env:COMPUTERNAME : " + ($_ | Out-String))
        }
    } -ArgumentList @($setupUserCred, $NodeNamePattern, $connUri.Host, $EnableRESTAPI.IsPresent, $EnableWebPortal.IsPresent, $postConfig)

    $cerFileName = ""
    if($EnableRESTAPI.IsPresent -or $EnableWebPortal.IsPresent)
    {
        try
        {
            if($ret["Certificate"] -ne $null)
            {
                $cerFileName = "HPCWebComponent_{0}_{1}.cer" -f $VMName, (Get-Date).ToString("yyyyMMddHHmmss")
                try
                {
                    $certdir = [Environment]::GetFolderPath("MyDocuments")
                }
                catch
                {
                }

                if(-not (Test-Path -Path $certdir -PathType Container))
                {
                    $certdir = $env:TEMP
                }

                $cerFileName = [IO.Path]::Combine($certdir, $cerFileName)
                [IO.File]::WriteAllBytes($cerFileName, $ret["Certificate"])
                    
                # Validate the certificate file
                $cert = GetCertificate $cerFileName
                TraceInfo "The certificate file for the HPC Pack web service was saved at: $cerFileName"
            }

            TraceInfo "Adding endpoint 'HTTPS'(Protocol=tcp, PublicPort=443, LocalPort=443) for virtual machine $VMName ..."
            $vm = InvokeAzureCmdWithRetry -Command "Get-AzureVM -Name $VMName -ServiceName $ServiceName" | Add-AzureEndpoint -Name 'HTTPS' -Protocol 'tcp' -PublicPort 443 -LocalPort 443
            Update-AzureVMWithRetry -Name $VMName -ServiceName $ServiceName -VM $vm.VM | Out-Null
            TraceInfo "Endpoint 'HTTPS' was added for virtual machine $VMName on cloud service $ServiceName" 
        }
        catch
        {
            throw ("Failed to configure head node $VMName : " + ($_ | Out-String))
        }
    }

    TraceInfo "Head node $VMName configured."
    return $cerFileName
}

function RemoveCorpAzureDisk
{
    param
    (
        #The name of Azure disk
        [Parameter(Mandatory=$true)]
        [String] $DiskName,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 10
    )

    $retry = 0
    while($true)
    {
        try
        {
            Remove-AzureDisk -DiskName $DiskName -DeleteVHD -ErrorAction Stop
            break
        }
        catch [Microsoft.WindowsAzure.CloudException]
        {
            if($_.Exception.ErrorCode -eq "ResourceNotFound")
            {
                TraceInfo "Azure Disk $DiskName removed"
                break
            }
        }
        catch
        {
            TraceWarning "Failed to remove Azure Disk $DiskName : $_"
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to remove Azure Disk $DiskName after 20 seconds"
            Start-Sleep -Seconds 20
            $retry++
        }
        else
        {
            throw "Failed to remove the Azure disk $DiskName, you can remove it manually."            
        }
    }
}

function CreateHPCNodeBaseImage
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String] $Name, 

        [Parameter(Mandatory=$true)]
        [String] $VMSize, 
        
        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 
        
        [Parameter(Mandatory=$true)]
        [String] $ClusterName, 
        
        [Parameter(Mandatory=$true)]
        [String] $HeadNodeServiceName, 
                
        [Parameter(Mandatory=$true)]
        [ValidateSet("ComputeNode","BrokerNode")]
        [String] $NodeType,
        
        [Parameter(Mandatory=$true)]
        [String] $VNetName, 

        [Parameter(Mandatory=$true)]
        [String] $SubnetName, 
                
        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred, 
        
        [Parameter(Mandatory=$true)]
        [PSCredential] $DomainUserCred, 
        
        [Parameter(Mandatory=$true)]
        [String] $DomainFQDN, 

        [Parameter(Mandatory=$true)]
        [String] $SourceImageName
    )

    # The source image is not a valid HPC CN/BN image for this cluster, create a base image
    # Use a non-HPC deployment label for base node

    CreateDomainJoinedVM -VMName $Name `
                         -ServiceName $ServiceName `
                         -VMSize $VMSize `
                         -ImageName $SourceImageName `
                         -VNetName $VNetName `
                         -SubnetName $SubnetName `
                         -LocalAdminCred $LocalAdminCred `
                         -DomainUserCred $DomainUserCred `
                         -DomainFQDN $DomainFQDN `
                         -DeploymentLabel "HPCBaseImage-$NodeType" `
                         -DisableGuestAgent

    # Wait for the base node to be ready
    WaitForVMReady -VMName $Name -ServiceName $ServiceName

    # Wait for the Head Node VM to be ready
    WaitForVMReady -VMName $ClusterName -ServiceName $HeadNodeServiceName

    #Install Compute/Broker Node
    $rmPshBlock = {
        param($clustername, $nodetype)

        if(Test-Path -Path "C:\HPCPatches")
        {
            Remove-Item -Path "C:\HPCPatches" -Recurse -Force -ErrorAction SilentlyContinue | Out-Null
        }

        New-Item -ItemType directory -Path "C:\HPCPatches" -Force | Out-Null
    
        # 0 for Standalone Workstation, 1 for Member Workstation, 2 for Standalone Server, 3 for Member Server, 4 for Backup Domain Controller, 5 for Primary Domain Controller
        $domainRole = (Get-WmiObject Win32_ComputerSystem).DomainRole
        if($domainRole -lt 3)
        {
            throw "$nodetype $env:COMPUTERNAME is not domain joined"
        }
    
        # Test the connection to head node
        WriteLog "Testing the connection to $ClusterName ..."
        $maxRetries = 50
        $retry = 0
        while($true)
        {
            # Flush the DNS cache in case the cached head node ip is wrong.
            # Do not use Clear-DnsClientCache because it is not supported in Windows Server 2008 R2
            Start-Process -FilePath ipconfig -ArgumentList "/flushdns" -Wait -NoNewWindow | Out-Null
            if(Test-Connection -ComputerName $ClusterName -Quiet)
            {
                WriteLog "Head node $ClusterName is now reachable."
                break
            }
            else
            {
                if($retry -lt 20)
                {
                    Start-Sleep -Seconds 20
                    $retry++
                }
                else
                {
                    throw "Head node $ClusterName is unreachable"
                }
            }
        }

        # Because ScheduledTasks PowerShell module not available in Windows Server 2008 R2,
        # We use ComObject Schedule.Service to schedule task
        try
        {
            $schdService = new-object -ComObject "Schedule.Service"
            $schdService.Connect("localhost") | Out-Null
            $rootFolder = $schdService.GetFolder("\")
            $taskDefinition = $schdService.NewTask(0)
            $testPathAction = $taskDefinition.Actions.Create(0)
            $testPathAction.Path = "PowerShell.exe"
            $testPathPshCmd = '$retry=0; while($true){if(Test-Path \\{ClusterName}\REMINST\setup.exe){Copy-Item \\{ClusterName}\REMINST\Patches\KB*.exe C:\HPCPatches -ErrorAction SilentlyContinue; return} elseif($retry -lt 30){$retry++; start-sleep -seconds 20} else{throw ''not available''}}'
            $testPathPshCmd = $testPathPshCmd.Replace("{ClusterName}", $ClusterName)
            $testPathAction.Arguments = "-Command `"$testPathPshCmd`""
            $Action = $taskDefinition.Actions.Create(0)
            $Action.Path = "\\$clustername\REMINST\setup.exe"
            $Action.Arguments = "-unattend -{0}:{1}" -f $nodetype.ToLower(), $clustername
            $hpcSetupTask = $Rootfolder.RegisterTaskDefinition("hpcsetup", $taskDefinition, 2, "system", $null, 5)
            WriteLog  "HPC $nodetype installation task scheduled"
        }
        catch
        {
            throw "Failed to schedule HPC $nodetype installation task" 
        }
    

        try
        {
            $hpcSetupTask.Run($null) | Out-Null
            WriteLog  "HPC $nodetype installation task started"
            WriteLog "Waiting for HPC $nodetype installation ..."
            Start-Sleep -Seconds 1
            $retry = 0
            while($true)
            {
                $taskState = $hpcSetupTask.State
                # 2:Queued, 3:Ready, 4:Running
                if($taskState -eq 3)
                {
                    #If the task never run, the lastRunTime is December 30, 1899 12:00:00 AM
                    if($hpcSetupTask.LastRunTime -lt "1900-01-01")
                    {
                        $hpcSetupTask.Run($null) | Out-Null
                    }
                    else
                    {
                        if($hpcSetupTask.LastTaskResult -eq 0)
                        {
                            WriteLog  "HPC $nodetype installation completed"
                            break
                        }
                        else
                        {
                            # The setup task failed last time, re-run it
                            if($retry -lt 3)
                            {
                                $hpcSetupTask.Run($null) | Out-Null
                                $retry++
                            }
                            else
                            {
                                throw ("HPC $nodetype installation Failed: " + $hpcSetupTask.LastTaskResult)
                            }
                        }
                    }
                }
                elseif($taskState -eq 4)
                {
                    # if the setup hangs, kill the process to make the task fail, and wait for retry
                    $p = Get-Process -Name "HpcSetupChainer" -ErrorAction SilentlyContinue
                    if($p -ne $null)
                    {
                        $elapsedMins = ((Get-Date) - $p.StartTime).TotalMinutes
                        if($elapsedMins -gt 20)
                        {
                            WriteLog "HPC $nodetype installation hung, kill the HPC setup process."
                            $p.Kill()  | Out-Null
                        }
                    }
                }
                elseif($taskState -ne 2)
                {
                    throw "The HPC $nodetype installation task entered into unexpected state: $taskState"
                }

                Start-Sleep -Seconds 2
            }
        }
        catch
        {
            WriteLog "HPC $nodetype installation Failed"
            throw
        }
        finally
        {
            $Rootfolder.DeleteTask("hpcsetup",0)  | Out-Null
        }

        # Apply the patches if any
        $patchfiles = @(Get-ChildItem "C:\HPCPatches" -Filter "KB*.exe" | select -ExpandProperty FullName)
        if($patchfiles.Count -gt 0)
        {
            $patchTable = @{}
            foreach($pfile in $patchfiles)
            {
                $versionStr = [System.Diagnostics.FileVersionInfo]::GetVersionInfo($pfile).ProductVersion
                $version = New-Object System.Version $versionStr
                $patchTable[$version] = $pfile
            }

            $versions = @($patchTable.Keys | Sort-Object)
            foreach($ver in $versions)
            {
                $pfile = $patchTable[$ver]
                WriteLog "Applying QFE hotfix $pfile"
                $p = Start-Process -FilePath $pfile -ArgumentList "/unattend" -PassThru -Wait
                if(($p.ExitCode -ne 0) -and ($p.ExitCode -ne 3010))
                {
                    throw "Failed to apply QFE hotfix $pfile : $($p.ExitCode)"
                }

                Start-Sleep -Seconds 10
            }
        }
    }

    try
    {
        $setupScriptBytes = [IO.File]::ReadAllBytes("$env:HPCPSScriptRoot\SetupComplete2.cmd")
    }
    catch
    {
        throw ("Failed to load SetupComplete2.cmd :" + $_)
    }

    $joinDomainRetried = $false
    while($true)
    {
        try
        {
            RemotePshCall -ServiceName $ServiceName -VMName $Name -Credential $LocalAdminCred -ScriptBlock $rmPshBlock -ArgumentList @($ClusterName, $NodeType) | Out-Null
            break
        }
        catch [System.Management.Automation.RemoteException]
        {
            if($_.Exception.Message.Contains("is not domain joined"))
            {
                if($joinDomainRetried)
                {
                    throw
                }
                else
                {
                    TraceWarning "$Name is not domain joined, trying to join domain $DomainFQDN"
                    VMJoinToDomain -ServiceName $ServiceName -VMName $Name -LocalAdminCred $LocalAdminCred -DomainUserCred $DomainUserCred -DomainFQDN $DomainFQDN
                    $joinDomainRetried = $true
                }
            }
            else
            {
                throw
            }
        }
    }
    
    $connUri = InvokeAzureCmdWithRetry -Command "Get-AzureWinRMUri -ServiceName $ServiceName -Name $Name" -RetryTimes 5
    if($connUri -eq $null)
    {
        throw "Failed to retrieve the WinRM https listener URI of the virtual machine $Name in cloud service $ServiceName."
    }

    $rmPshBlock = {
        param($nodetype, $dnsHostName, $setupscript)

        WriteLog "Generating $env:windir\OEM\SetupComplete2.cmd"
        [IO.File]::WriteAllBytes("$env:windir\OEM\SetupComplete2.cmd", $setupscript)

        Remove-Item -Path "C:\HPCPatches" -Recurse -Force -ErrorAction SilentlyContinue

        WriteLog "Removing HPC setup logs"
        Remove-Item -Path "$env:windir\Temp\HPCSetupLogs" -Recurse -Force -ErrorAction SilentlyContinue

        # Stop the HPC services so that we can remove the log files
        $hpcServices = @("HpcManagement", "HpcNodeManager", "msmpi", "HpcMonitoringClient", "HpcSoaDiagMon")
        if($nodetype -eq "BrokerNode")
        {
            $hpcServices += "HpcBroker"
        }

        $hpcServices | Stop-Service -Force -ErrorAction SilentlyContinue

        WriteLog "Removing HPC operation log files"
        $datadir = [System.Environment]::GetEnvironmentVariable("CCP_DATA", [System.EnvironmentVariableTarget]::Machine)
        $logFilesPath = [System.IO.Path]::Combine($datadir, "LogFiles")
        if(Test-Path $logFilesPath -PathType Container)
        {
            Get-ChildItem -Path $logFilesPath | select -ExpandProperty FullName | Remove-Item -Force -Recurse -ErrorAction SilentlyContinue
        }

        Remove-Item "$env:SYSTEMDRIVE\AzureData\CustomData.bin" -Force -ErrorAction SilentlyContinue

        WriteLog "Clear registry value ClusterName"
        Set-ItemProperty -Path HKLM:\SOFTWARE\Microsoft\HPC -Name ClusterName -Value ""
        if(Test-Path HKLM:\SOFTWARE\Wow6432Node\Microsoft\HPC)
        {
            Set-ItemProperty -Path HKLM:\SOFTWARE\Wow6432Node\Microsoft\HPC -Name ClusterName -Value ""
        }

        WriteLog "Removing cloud service certificate"
        $myStore = New-Object System.Security.Cryptography.x509Certificates.x509Store("My","LocalMachine")
        try
        {
            $myStore.Open("ReadWrite")
            $svcCerts = @($myStore.Certificates | ?{$_.Subject -eq "CN=$dnsHostName"})
            foreach($cert in $svcCerts)
            {
                $myStore.Remove($cert)
            }
        }
        catch
        {
            WriteLog "Failed to remove cloud service certificate"
        }
        finally
        {
            # Close doesn't throw even when Open not successfully called
            $myStore.Close()
        }

        # Sysprep the VM
        WriteLog "Syspreping HPC $nodetype $env:COMPUTERNAME"
        $pobj = Invoke-WmiMethod -Path win32_process -Name Create -ArgumentList "$env:windir\System32\Sysprep\sysprep.exe /oobe /quiet /generalize /shutdown"
        $p = $null
        if($pobj.ReturnValue -eq 0)
        {
            $p = Get-Process -Id $pobj.ProcessId -ErrorAction SilentlyContinue
        }

        if($p -ne $null)
        {
            $p.WaitForExit()
        }
        else
        {
            throw "Failed to sysprep $env:COMPUTERNAME"
        }
    }

    $loopRetry = 0
    # Set the remote powershell call retry times to 30 for the first loop because the VM may not be ready to accept remote powershell call
    while($true)
    {
        try
        {
            # Error code 995 is expected here: The I/O operation has been aborted because of either a thread exit or an application request. 
            # Error code 5 is sometimes thrown when the remote call execution succeeded: Processing data for a remote command failed with the following error message: Access is denied. 
            RemotePshCall -ServiceName $ServiceName `
                          -VMName $Name `
                          -Credential $LocalAdminCred `
                          -ScriptBlock $rmPshBlock `
                          -ArgumentList @($NodeType, $connUri.Host, $setupScriptBytes) `
                          -IgnoreErrorCodes 995 `
                          -NoRetryErrorCodes 5 `
                          -RetryTimes 3 | Out-Null

            break
        }
        catch [System.Management.Automation.Remoting.PSRemotingTransportException]
        {
            if($loopRetry -lt 3)
            {
                try
                {
                    WaitForVMReady -VMName $Name -ServiceName $ServiceName -TargetStatus "StoppedVM" -MaxWaitSeconds 300
                    break
                }
                catch
                {
                    TraceWarning "$Name didn't enter into StoppedVM state"
                }
                
                $loopRetry++
            }
            else
            {
                throw
            }
        }
    }
    
    $baseVM = InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $ServiceName -Name $Name"
    $baseVMDiskName = $baseVM.VM.OSVirtualHardDisk.DiskName
    WaitForVMReady -VMName $Name -ServiceName $ServiceName -TargetStatus "StoppedVM"
    $TargetImageName = NewBaseHPCNodeImageName -NodeType $NodeType
    TraceInfo "Creating base image $TargetImageName from $NodeType $Name"
    Save-AzureVMImageWithRetry -Name $Name -ServiceName $ServiceName -ImageName $TargetImageName -ImageLabel "HPC Pack $NodeType Image"
    return @{"ImageName"=$TargetImageName;"CorpDiskName"=$baseVMDiskName}
}

function StartHPCNodesForService
{

    param 
    (
        [Parameter(Mandatory=$true)]
        [String[]] $VMNames, 

        [Parameter(Mandatory=$true)]
        [String] $VMSize, 
        
        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 
        
        [Parameter(Mandatory=$true)]
        [String] $StorageAccountName, 

        [Parameter(Mandatory=$true)]
        [String] $VNetName, 

        [Parameter(Mandatory=$true)]
        [String] $SubnetName, 

        [Parameter(Mandatory=$true)]
        [String] $DeploymentLabel, 
                
        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred, 
        
        [Parameter(Mandatory=$true)]
        [PSCredential] $DomainUserCred, 
        
        [Parameter(Mandatory=$true)]
        [String] $DomainFQDN, 

        [Parameter(Mandatory=$true)]
        [String] $ImageName,

        [Parameter(Mandatory=$true)]
        [String] $CustomDataFile
    )

    TraceInfo ("Starting a job to deploy the following nodes in cloud service ${ServiceName}: " + ($VMNames -join ","))
    $subscriptionName = (Get-AzureSubscription -Current).SubscriptionName
    $job = Start-Job -Name $ServiceName -ScriptBlock {
        param(
            $subscriptionName, $StorageAccountName, $VMNames, $VMSize, $ServiceName, $VNetName, $SubnetName, $DeploymentLabel, $LocalAdminCred, 
            $DomainUserCred, $DomainFQDN, $ImageName, $CustomDataFile
        )

        try
        {
            Import-Module -Name Azure
            . "$env:HPCPSScriptRoot\HPCIaaSClusterUtils.ps1"
            . "$env:HPCPSScriptRoot\IaaSAzureCmdLib.ps1"
            . "$env:HPCPSScriptRoot\HPCIaaSClusterConfigLib.ps1"
            . "$env:HPCPSScriptRoot\HPCIaaSClusterLib.ps1"

            Register-EngineEvent -SourceIdentifier "DeployHPCNodeInService" -Forward

            Select-AzureSubscription -SubscriptionName $subscriptionName
            SetSubscriptionStorageAccount -SubscriptionName $subscriptionName -StorageAccountName $StorageAccountName


            $infiniband = IsVMSizeWithRDMA -VMSize $VMSize
            $netbios = $DomainFQDN.Split(".")[0].ToUpper()
            $domainUserName = $DomainUserCred.UserName
            if($DomainUserCred.UserName.Contains("\"))
            {
                $domainUserName = $domainUserName.Split("\")[1]
            }

            $vmConfigs = @{}
            foreach($vmName in $VMNames)
            {
                $vm = New-AzureVMConfig -Name $vmName -InstanceSize $VMSize -ImageName $ImageName |
                      Add-AzureProvisioningConfig -WindowsDomain -AdminUsername $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                          -Domain $netbios -EnableWinRMHttp -DomainUserName $domainUserName -DomainPassword $DomainUserCred.GetNetworkCredential().Password `
                          -JoinDomain $DomainFQDN -CustomDataFile $CustomDataFile -DisableGuestAgent:(-not $infiniband) |
                      Set-AzureSubnet -SubnetNames $SubnetName

                if($infiniband)
                {
                    $vm = $vm | Set-AzureVMExtension -ExtensionName "HpcVmDrivers" -Publisher "Microsoft.HpcCompute" -Version "1.*"
                }

                $vmConfigs[$vmName] = $vm
            }

            $maxRetries = 3
            $retry = 0
            while(($vmConfigs.Count -gt 0) -and ($retry -lt $maxRetries))
            {
                $leftVMNames = @($vmConfigs.Keys | sort)
                foreach($vmName in $leftVMNames)
                {
                    try
                    {
                        New-Event -SourceIdentifier "DeployHPCNodeInService" -Sender "InfoLevel" -MessageData "Creating HPC node $vmName" | Out-Null
                        New-AzureVMWithRetry -VM $vmConfigs[$vmName] -ServiceName $ServiceName -VNetName $VNetName -DeploymentLabel $DeploymentLabel -RetryTimes 4
                        $vmConfigs.Remove($vmName)
                    }
                    catch
                    {
                        New-Event -SourceIdentifier "DeployHPCNodeInService" -Sender "WarningLevel" -MessageData "Failed to create VM $vmName : $_" | Out-Null
                    }
                }

                $retry++
            }
    
            $startingVMs = @($VMNames | ?{$vmConfigs[$_] -eq $null})
            if($startingVMs.Count -gt 0)
            {
                New-Event -SourceIdentifier "DeployHPCNodeInService" -Sender "InfoLevel" -MessageData "$($startingVMs.Count) nodes created in cloud service $ServiceName, waiting for them to become ready." | Out-Null
                try
                {
                    WaitForVMReady -VMName $startingVMs -ServiceName $ServiceName
                    New-Event -SourceIdentifier "DeployHPCNodeInService" -Sender "InfoLevel" -MessageData "The virtual machines in cloud service $ServiceName are now ready for use: $($startingVMs -join ',')"  | Out-Null
                }
                catch
                {
                    New-Event -SourceIdentifier "DeployHPCNodeInService" -Sender "WarningLevel" -MessageData "$_" | Out-Null
                }
            }
        }
        catch
        {
            # If we don't add try-catch here, the script will continue running even when some exceptions happens
            throw
        }
    } -ArgumentList @($subscriptionName, $StorageAccountName, $VMNames, $VMSize, $ServiceName, $VNetName, $SubnetName, `
                $DeploymentLabel, $LocalAdminCred, $DomainUserCred, $DomainFQDN, $ImageName, $CustomDataFile)
    return $job
}

function CreateHPCNodeGroup
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [HashTable] $ServiceVMTable, 

        [Parameter(Mandatory=$true)]
        [HashTable] $ServiceStorageTable, 
        
        [Parameter(Mandatory=$true)]
        [String] $VMSize, 

        [Parameter(Mandatory=$true)]
        [String] $ClusterName, 
        
        [Parameter(Mandatory=$true)]
        [String] $HeadNodeServiceName, 
                
        [Parameter(Mandatory=$true)]
        [ValidateSet("ComputeNode","BrokerNode")]
        [String] $NodeType,
        
        [Parameter(Mandatory=$true)]
        [String] $VNetName, 

        [Parameter(Mandatory=$true)]
        [String] $SubnetName, 

        [Parameter(Mandatory=$true)]
        [String] $DeploymentLabel, 
                
        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred, 
        
        [Parameter(Mandatory=$true)]
        [PSCredential] $DomainUserCred, 
        
        [Parameter(Mandatory=$true)]
        [String] $DomainFQDN, 

        [Parameter(Mandatory=$true)]
        [String] $SourceImageName,

        [Parameter(Mandatory=$false)]
        [Switch] $SrcImageHPCInstalled
    )
    
    $baseVMDiskName = ""
    if($SrcImageHPCInstalled.IsPresent)
    {
        # The source image is already a valid HPC CN/BN image for this cluster, directly use it to create HPC node
        $TargetImageName = $SourceImageName
    }
    else
    {
        $baseServiceName = $ServiceVMTable.Keys | sort | select -First 1
        $baseVMName = $ServiceVMTable[$baseServiceName][0]
        $sub = Get-AzureSubscription -Current
        SetSubscriptionStorageAccount -SubscriptionName $sub.SubscriptionName -StorageAccountName $ServiceStorageTable[$baseServiceName]
        $ret = CreateHPCNodeBaseImage -Name $baseVMName -VMSize $VMSize -ServiceName $baseServiceName -ClusterName $ClusterName -HeadNodeServiceName $HeadNodeServiceName `
                    -NodeType $NodeType -VNetName $VNetName -SubnetName $SubnetName -LocalAdminCred $LocalAdminCred -DomainUserCred $DomainUserCred `
                    -DomainFQDN $DomainFQDN -SourceImageName $SourceImageName
        $TargetImageName = $ret["ImageName"]
        $baseVMDiskName = $ret["CorpDiskName"]
    }

    DeployHPCNodesWithImage -ServiceVMTable $ServiceVMTable -ServiceStorageTable $ServiceStorageTable -ClusterName $ClusterName -VMSize $VMSize -VNetName $VNetName `
            -SubnetName $SubnetName -DeploymentLabel $DeploymentLabel -LocalAdminCred $LocalAdminCred -DomainUserCred $DomainUserCred -DomainFQDN $DomainFQDN `
            -SourceImageName $TargetImageName

    # Remove the corpse disk of the base node
    if(-not [String]::IsNullOrEmpty($baseVMDiskName))
    {
        try
        {
            RemoveCorpAzureDisk -DiskName $baseVMDiskName
        }
        catch
        {
            # Just give a warning message
            TraceWarning "Failed to remove the Azure disk $baseVMDiskName, you can remove it manually."
        }
    }

    Write-Output $TargetImageName
}


function GetTemplateParameters
{
    param
    (
        [Parameter(Mandatory=$true)]
        [System.Xml.XmlElement] $Template
    )

    $params = @{}
    $params["TemplateName"] = $Template.TemplateName.Trim()
    $params["SubscriptionId"] = $Template.SubscriptionId.Trim()
    $params["CertificateThumbprint"] = $Template.CertificateId.Trim()
    $params["ServiceName"] = $Template.ServiceName.Trim()
    $params["StorageServiceName"] = $Template.StorageAccount.Trim()
    if($Template["NodeCount"] -ne $null)
    {
        $nodeCount = [Int]$Template.NodeCount.Trim()
        if($nodeCount -gt 0)
        {
            $params["NodeCount"] = $nodeCount
            $params["RoleSize"] = $Template.RoleSize.Trim()
        }
    }

    if($Template["Proxy"] -eq $null)
    {
        $params["UsesStaticProxyCount"] = "True"
        $params["ProxyMultiplicity"] = "2"
    }
    else
    {
        if($Template.Proxy.UsesStaticProxyCount.Trim() -eq "true")
        {
            $params["UsesStaticProxyCount"] = "True"
            if($Template.Proxy["ProxyMultiplicity"] -ne $null)
            {
                $params["ProxyMultiplicity"] = $Template.Proxy.ProxyMultiplicity.Trim()
            }
        }
        else
        {
            $params["UsesStaticProxyCount"] = "False"
            if($Template.Proxy["ProxyRatio"] -ne $null)
            {
                $params["ProxyRatio"] = $Template.Proxy.ProxyRatio.Trim()
            }

            if($Template.Proxy["ProxyRatioBase"] -ne $null)
            {
                $params["ProxyRatioBase"] = $Template.Proxy.ProxyRatioBase.Trim()
            }
        }
    }

    if($Template["OSVersion"] -ne $null)
    {
        $osversion = $Template.OSVersion.Trim()
        if($osversion -eq "WindowsServer2008R2")
        {
            $params["WorkerOSFamily"] = "Win2008R2"
        }
        elseif($osversion -eq "WindowsServer2012")
        {
            $params["WorkerOSFamily"] = "Win8"
        }
    }

    if($Template["VhdDriveUrl"] -ne $null)
    {
        $params["VhdDriveUrl"] = $Template.VhdDriveUrl.Trim()
    }

    if($Template["StartupScript"] -ne $null)
    {
        $params["StartupScript"] = $Template.StartupScript.Trim()
    }

    return $params
}

function GetAzureBustConfig
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $AzureBustConfigXml,

        [Parameter(Mandatory=$true)]
        [HashTable] $CertPasswords
    )

    $xmlBurst = ([xml]$AzureBustConfigXml).AzureBurst
    $certList = @()
    $certIdThumbprintMap = @{}
    foreach($cert in $xmlBurst.Certificate)
    {
        $certMap = @{}
        $certId = [Int]$cert.Id.Trim()
        $pfxFile = $cert.PfxFile.Trim()
        $certMap["CertName"] = (Get-item $pfxFile).BaseName
        $certMap["PfxContent"] = [IO.File]::ReadAllBytes($pfxFile)
        if($CertPasswords[$certId] -eq $null)
        {
            $certIdThumbprintMap[$certId] = GetThumbprintFromCertificateFile -FilePath $pfxFile
        }
        else
        {
            $certMap["Password"] = $CertPasswords[$certId]
            $certIdThumbprintMap[$certId] = GetThumbprintFromCertificateFile -FilePath $pfxFile -Password $CertPasswords[$certId]         
        }

        $certList += $certMap
    }

    $templateList = @()
    foreach($temp in $xmlBurst.AzureNodeTemplate)
    {
        $tempParams = GetTemplateParameters $temp
        $certId = [Int]$tempParams["CertificateThumbprint"]
        $tempParams["CertificateThumbprint"] = $certIdThumbprintMap[$certId]
        $templateList += $tempParams
    }

    return @{"CertList"=$certList;"TempList"=$templateList}
}

function CreateAzureNodeTemplates
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [PSCredential] $SetupUserCred,

        [Parameter(Mandatory=$true)]
        [Object[]] $CertList,

        [Parameter(Mandatory=$true)]
        [Object[]] $NodeTemplates,

        [Parameter(Mandatory=$false)]
        [String] $PostConfigScirptFile = ""
    )

    TraceInfo "Creating Azure node templates on the HPC head node $VMName"

    $DefaultAzureNodeTemplateXml = @'
<?xml version="1.0" encoding="utf-8"?>
<Template xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" Description="" Version="6">
  <Item Name="DeployAzureNode">
    <Parameter Name="CertificateThumbprint" Value="" />
    <Parameter Name="SubscriptionId" Value="" />
    <Parameter Name="StorageServiceName" Value="" />
    <Parameter Name="ServiceName" Value="" />
    <Parameter Name="UsesStaticProxyCount" Value="True" />
    <Parameter Name="ProxyMultiplicity" Value="2" />
    <Parameter Name="ProxyRatio" Value="200" />
    <Parameter Name="ProxyRatioBase" Value="300" />
    <Parameter Name="WorkerOSFamily" Value="Win8R2" />
    <Parameter Name="VhdDriveUrl" Value="" />
    <Parameter Name="UsesVMImage" Value="False" />
    <Parameter Name="AzureAdminUserName" Value="" />
    <Parameter Name="AzureAdminPassword" Value="" />
    <Parameter Name="StartupScript" Value="" />
    <ParameterList Name="SubNets" />
    <Parameter Name="Description" Value="Deploys an Windows Azure node with a given subscription, service and specified role type." />
  </Item>
  <Item Name="DeploymentSchedule">
    <Parameter Name="ManualAvailability" Value="True" />
    <Parameter Name="SerializedDeploymentSchedule" Value="000000000000000000000000000000000000000000000000000000000000000000000000000000000000" />
    <Parameter Name="GracePeriod" Value="0" />
    <Parameter Name="Description" Value="Determines when a given node is started and stopped." />
  </Item>
</Template>
'@

    $baseNodeTemp = [xml]$DefaultAzureNodeTemplateXml

    # Set the user name and password used for RDP to the Azure Nodes
    $usrName = $SetupUserCred.UserName.Split('\')[-1]
    $usrPsw = $SetupUserCred.GetNetworkCredential().Password
    $node = $baseNodeTemp.SelectSingleNode("//Parameter[@Name='AzureAdminUserName']")
    if($node -ne $null)
    {
        $node.Value = $usrName
    }
    $node = $baseNodeTemp.SelectSingleNode("//Parameter[@Name='AzureAdminPassword']")
    if($node -ne $null)
    {
        $node.Value = $usrPsw
    }

    $postConfig = $null
    if((-not [String]::IsNullOrEmpty($PostConfigScirptFile)) -and (Test-Path -Path $PostConfigScirptFile -PathType Leaf))
    {
        $postConfig = [IO.File]::ReadAllBytes($PostConfigScirptFile)
    }

    RemotePshCall -ServiceName $ServiceName -VMName $VMName -Credential $SetupUserCred -ScriptBlock  {
        param([String]$baseTemplate, [Int]$certNum, [Object[]]$CertList, [Int]$templateNum, [Object[]]$NodeTemplates, [Byte[]]$postConfigContent)

        Add-PSSnapin Microsoft.HPC
        $rootDir = "$env:TEMP\HPCAzureBurst" + (Get-Random)
        try
        {
            New-Item -ItemType directory -Path $rootDir -Force | Out-Null

            foreach($cert in $CertList)
            {
                $CertName = $Cert["CertName"]
                $pfxFileName = [IO.Path]::Combine($rootDir, "$CertName.pfx")
                WriteLog "Importing the certificate from '$CertName.pfx'"
                [IO.File]::WriteAllBytes($pfxFileName, $Cert["PfxContent"])
                try
                {
                    if(($Cert["Password"] -ne $null) -and ($Cert["Password"] -ne ""))
                    {
                        $password = $Cert["Password"]
                        $secPassword = ConvertTo-SecureString -String $password -AsPlainText -Force
                        $certificate = Import-PfxCertificate -FilePath $pfxFileName -CertStoreLocation cert:\localMachine\my -Password $secPassword
                    }
                    else
                    {
                        $certificate = Import-PfxCertificate -FilePath $pfxFileName -CertStoreLocation cert:\localMachine\my
                    }
                
                    if($certificate.Subject -eq $certificate.Issuer)
                    {
                        # Self-signed certificate, export a cer file and import to cert:\localMachine\root
                        WriteLog "Exporting $CertName.cer from $CertName.pfx"
                        $cerFileName = [IO.Path]::Combine($rootDir, "$CertName.cer")
                        Export-Certificate -Cert $certificate -FilePath $cerFileName -Force

                        WriteLog "Importing the certificate from $CertName.cer"
                        Import-Certificate -FilePath $cerFileName -CertStoreLocation cert:\localMachine\root
                    }
                }
                catch
                {
                    throw ("Failed to import certificate $CertName.pfx : " + ($_ | Out-String)) 
                }
            }

            foreach($tempParams in $NodeTemplates)
            {
                try
                {
                    $templateName = $tempParams["TemplateName"]
                    WriteLog "Creating Azure node template $templateName."
                    $tempXml = [xml]$baseTemplate
                    foreach($param in $tempParams.Keys)
                    {
                        if(($param -eq "TemplateName") -or ($param -eq "NodeCount") -or ($param -eq "RoleSize"))
                        {
                            continue
                        }

                        $node = $tempXml.SelectSingleNode("//Parameter[@Name='$param']")
                        if($node -ne $null)
                        {
                            $node.Value = $tempParams[$param]
                        }
                    }

                    $templateFile = [IO.Path]::Combine($rootDir,  "$templateName.xml")
                    $tempXml.Save($templateFile)
                    Import-HpcNodeTemplate $templateFile -Force
                }
                catch
                {
                    throw ("Failed to create Azure node template $templateName :" + ($_ | Out-String)) 
                }

                try
                {
                    if($tempParams["NodeCount"] -ne $null)
                    {
                        WriteLog "Adding Azure nodes for Azure node template $templateName."
                        Get-HpcNodeTemplate -Name $templateName | Add-HpcNodeSet -Quantity $tempParams["NodeCount"] -Size $tempParams["RoleSize"]
                    }
                }
                catch
                {
                    throw ("Failed to create Azure nodes for Azure node template $templateName :" + ($_ | Out-String)) 
                }
            }

            if(($postConfigContent -ne $null) -and ($postConfigContent.Count -gt 0))
            {
                $postConfigScriptFile = "$env:temp\HPCHNPostConfig.ps1"
                try
                {
                    WriteLog "Creating $postConfigScriptFile"
                    [IO.File]::WriteAllBytes($postConfigScriptFile, $postConfigContent)
                    WriteLog "$postConfigScriptFile created"
                    $pobj = Invoke-WmiMethod -Path win32_process -Name Create -ArgumentList "PowerShell.exe -ExecutionPolicy Unrestricted -File `"$postConfigScriptFile`""
                    if($pobj.ReturnValue -eq 0)
                    {
                        WriteLog "Start to run $postConfigScriptFile."
                    }
                    else
                    {
                        WriteLog "Failed to run $postConfigScriptFile."
                    }
                }
                catch
                {
                    WriteLog "Failed to run $postConfigScriptFile : $_"
                }
            }
        }
        finally
        {
            Remove-Item -Path $rootDir -Recurse -Force -ErrorAction SilentlyContinue
        }

    } -ArgumentList @($baseNodeTemp.OuterXml, $CertList.Count, $CertList, $NodeTemplates.Count, $NodeTemplates, $postConfig) | Out-Null

    TraceInfo "Azure node templates created"
}


function SetSubscriptionStorageAccount
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $SubscriptionName, 

        [Parameter(Mandatory=$true)]
        [String] $StorageAccountName
    )

    $curSubs = Get-AzureSubscription -Current -ErrorAction SilentlyContinue
    if(($curSubs -ne $null) -and ($curSubs.SubscriptionName -eq $SubscriptionName))
    {
        $subs = $curSubs
    }
    else
    {
        $subs = GetAzureSubscription -SubscriptionName $SubscriptionName
        if($subs -eq $null)
        {
            throw "The Azure subscription $SubscriptionName was not configured."
        }
    }

    if($subs.CurrentStorageAccountName -ne $StorageAccountName)
    {
        # use cross-process mutex to prevent from calling Set-AzureSubscription concurrently
        $setSubMutex = New-Object System.Threading.Mutex $false,'HPCIaaSSetAzureSubscriptionMutex'
        $mutexAcquired = $false
        try
        {
            try
            {
                $mutexAcquired = $setSubMutex.WaitOne()
            }
            catch [System.Threading.AbandonedMutexException]
            {
                $mutexAcquired = $true
            }

            $setSubRetry = 0
            while($true)
            {
                try
                {
                    Set-AzureSubscription -SubscriptionName $SubscriptionName -CurrentStorageAccountName $StorageAccountName -ErrorAction Stop | Out-Null
                    TraceInfo "Set current storage account to $StorageAccountName for subscription $SubscriptionName"
                    break
                }
                catch
                {
                    $exMsg = "Failed to set current storage account: " + ($_ | Out-String)
                    if($setSubRetry -lt 3)
                    {
                        TraceWarning $exMsg
                        Start-Sleep -Seconds (Get-Random -Minimum 2 -Maximum 5)
                        $setSubRetry++
                    }
                    else
                    {
                        throw $exMsg
                    }
                }
            }
        }
        finally
        {
            if($mutexAcquired)
            {
                $setSubMutex.ReleaseMutex()
            }
        }
    }
}

function GetStorageBlobEndpoint
{
    param
    (
        [Parameter(Mandatory=$true)]
        $StorageAccount
    )

    $storageBlobEndPoint = $StorageAccount.Endpoints | ?{$_ -match ".blob."}
    $storageBlobSuffix = $storageBlobEndPoint.Substring($storageBlobEndPoint.IndexOf(".blob.")).TrimEnd("/")
    return ($StorageAccount.StorageAccountName + $storageBlobSuffix)
}

function ResolveStorageEndpointIP
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Endpoint
    )
    
    $retry = 0
    while($retry -lt 3)
    {
        try
        {
            $ips = [System.Net.Dns]::GetHostAddresses($Endpoint)
            return $ips[0].IPAddressToString
        }
        catch
        {
            $retry++
            Clear-DnsClientCache
        }
    }
    
    throw "The Endpoint cannot be resolved: $Endpoint"  
}

function DeployHPCNodesWithImage
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [HashTable] $ServiceVMTable, 

        [Parameter(Mandatory=$true)]
        [HashTable] $ServiceStorageTable, 

        [Parameter(Mandatory=$true)]
        [String] $ClusterName, 
        
        [Parameter(Mandatory=$true)]
        [String] $VMSize, 

        [Parameter(Mandatory=$true)]
        [String] $VNetName, 

        [Parameter(Mandatory=$true)]
        [String] $SubnetName, 

        [Parameter(Mandatory=$true)]
        [String] $DeploymentLabel, 
                
        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred, 
        
        [Parameter(Mandatory=$true)]
        [PSCredential] $DomainUserCred, 
        
        [Parameter(Mandatory=$true)]
        [String] $DomainFQDN, 

        [Parameter(Mandatory=$true)]
        [String] $SourceImageName
    )

    Register-EngineEvent -SourceIdentifier "DeployHPCNodeInService" -Action {
        try
        {
            if($Sender -eq "ErrorLevel")
            {
                TraceError $event.MessageData
            }
            elseif($Sender -eq "WarningLevel")
            {
                TraceWarning $event.MessageData
            }
            else
            {
                TraceInfo $event.MessageData
            }
        }
        catch
        {
            #Ignore the error
        }
    }

    $provsionJobs = @()

    $hnNameFile = "$env:TEMP\HPCHN_{0}_{1}.txt" -f $ClusterName, (Get-Random)
    "HPCClusterName=$ClusterName" | Out-File $hnNameFile

    $existingStorages = @(InvokeAzureCmdWithRetry -Command "Get-AzureStorageAccount")
    $DestStorageAccount = @($ServiceVMTable.Keys | %{$ServiceStorageTable[$_]} | select -Unique)
    $environment = (Get-AzureSubscription -Current).Environment
    $userImageList = @(InvokeAzureCmdWithRetry -Command "Get-AzureVMImage" | ?{$_.Category -eq "User"})
    $srcImage = $userImageList | ?{$_.ImageName -eq $SourceImageName}
    if($srcImage -eq $null)
    {
        throw "The VM image $SourceImageName does not exist!"
    }

    $srcImageTimeStamp = GetVMImageTimeStamp $srcImage
    $srcMediaLink = GetVMImageMediaLink $srcImage
    $srcStorageAccount = GetStorageAccountFromMediaLink $srcMediaLink
    $srcLocation = $srcImage.Location
    if([String]::IsNullOrEmpty($srcLocation))
    {
        # in case the storage account is in an affinity group
        $srcLocation = $existingStorages | ?{$_.StorageAccountName -eq $srcStorageAccount} | %{$_.GeoPrimaryLocation}
    }
    if($SourceImageName -match "_$srcStorageAccount$")
    {
        $imageNameRoot = $SourceImageName.Substring(0, $SourceImageName.Length - $srcStorageAccount.Length - 1)
    }
    else
    {
        $imageNameRoot = $SourceImageName
    }

    $srcUris = @{}
    $srcUris[$srcStorageAccount] = $srcMediaLink
    foreach($img in $userImageList)
    {
        if(-not $img.ImageName.StartsWith($imageNameRoot))
        {
            continue
        }

        $mediaLink = GetVMImageMediaLink $img
        $stg = GetStorageAccountFromMediaLink $mediaLink
        $imgLocation = $img.Location
        if([String]::IsNullOrEmpty($imgLocation))
        {
            # in case the storage account is in an affinity group
            $imgLocation = $existingStorages | ?{$_.StorageAccountName -eq $stg} | %{$_.GeoPrimaryLocation}
        }
        
        if($imgLocation -ne $srcLocation)
        {
            continue
        }

        if(($img.ImageName -eq $imageNameRoot) -or ($img.ImageName -eq "${imageNameRoot}_${stg}"))
        {
            $imgTimeStamp = GetVMImageTimeStamp $img
            if($imgTimeStamp -eq $srcImageTimeStamp)
            {
                $imgName = $img.ImageName
                $srcUris[$stg] = $mediaLink
                # It is possible that some source images are not in the destination storage accounts
                if($DestStorageAccount -contains $stg)
                {
                    TraceInfo "The image $imgName already exists in the storage account $stg"
                    foreach($svc in @($ServiceVMTable.Keys | ?{$ServiceStorageTable[$_] -eq $stg}))
                    {
                        $provsionJobs += StartHPCNodesForService -VMNames $ServiceVMTable[$svc] -VMSize $VMSize -ServiceName $svc -StorageAccountName $stg `
                                                    -VNetName $VNetName -SubnetName $SubnetName -DeploymentLabel $DeploymentLabel -LocalAdminCred $LocalAdminCred `
                                                    -DomainUserCred $DomainUserCred -DomainFQDN $DomainFQDN -ImageName $imgName -CustomDataFile $hnNameFile
                    }
                }
            }
            else
            {
                # It is possible that some source images are not in the destination storage accounts
                if($DestStorageAccount -contains $stg)
                {
                    TraceInfo "Removing the image $imgName from storage account $stg because its time stamp is different from that of the source image $SourceImageName"
                    Remove-AzureVMImageWithRetry -ImageName $imgName -DeleteVHD
                }
            }
        }
    }

    $containerName = "hpcimages"
    $srcStorageTenants = @{}
    $destImageNames = @{}
    $destStorageTenants = @{}
    $storageContext = @{}
    $storageAccounts = @($DestStorageAccount + $srcUris.Keys | select -Unique)
    foreach($stg in $storageAccounts)
    {
        $storage = $existingStorages | ?{$_.StorageAccountName -eq $stg}
        if($storage -eq $null)
        {
            throw "The storage account $stg doesn't exist"
        }

        $endPoint = GetStorageBlobEndpoint -StorageAccount $storage
        $ipStr = ResolveStorageEndpointIP -Endpoint $endPoint
        if($srcUris[$stg] -eq $null)
        {
            $destImageNames[$stg] = "${imageNameRoot}_${stg}"
            $destStorageTenants[$ipStr] += @($stg)
        }
        else
        {
            # If multiple source storage accounts in one tenant, we just copy from one of them
            if($srcStorageTenants[$ipStr] -eq $null)
            {
                $srcStorageTenants[$ipStr] = $stg
            }
            else
            {
                continue
            }
        }

        $storageKey = (InvokeAzureCmdWithRetry -Command "Get-AzureStorageKey -StorageAccountName $stg").Primary
        $storageContext[$stg] = New-AzureStorageContext -StorageAccountName $stg -StorageAccountKey $storageKey -Environment $environment
        ## Create the container on the destination ### 
        if($srcUris[$stg] -eq $null)
        {
            $cont = Get-AzureStorageContainerWithRetry -Name $containerName -Context $storageContext[$stg]
            if($cont -eq $null)
            {
                $null = New-AzureStorageContainerWithRetry -Name $containerName -Context $storageContext[$stg]
            }
        }
    }

    $destVHDName =  "$imageNameRoot.vhd"
    $copyMaps = @{}
    # Copy from storage accounts in the same tenants
    foreach($ipStr in $srcStorageTenants.Keys)
    {
        $srcStg = $srcStorageTenants[$ipStr]
        if($destStorageTenants[$ipStr] -ne $null)
        {
            $copyMaps[$srcStg] = $destStorageTenants[$ipStr]
            $destStorageTenants.Remove($ipStr)
        }
    }

    # Copy from storage accounts in different tenants
    if($destStorageTenants.Count -gt 0)
    {
        $srcStgs = $srcStorageTenants.Values
        $destStorageCount = $destStorageTenants.Count
        $index = 0
        foreach($ipStr in $destStorageTenants.Keys)
        {
            $srcStgIndex = $index % $srcStgs.Count
            $srcStg = $srcStgs[$srcStgIndex]
            $copyMaps[$srcStg] += @($destStorageTenants[$ipStr][0])
            $index++
        }
    }

    $blobCopies = @{}
    foreach($stg in @($DestStorageAccount | ?{$srcUris.Keys -notcontains $_}))
    {
        $blobCopies[$stg] = $null
    }

    $queueTime = @{}
    $startTime = @{}
    $doneTime = @{}
    try
    {
        foreach($srcStg in $copyMaps.Keys)
        {
            foreach($destStg in $copyMaps[$srcStg])
            {
                TraceInfo "Start to copy $destVHDName to the storage account $destStg"
                ### Start the asynchronous copy - specify the source authentication with -SrcContext ### 
                $blobCopies[$destStg] += Start-AzureStorageBlobCopyWithRetry -AbsoluteUri $srcUris[$srcStg] `
                                                                             -Context $storageContext[$srcStg] `
                                                                             -DestContainer $containerName `
                                                                             -DestBlob $destVHDName `
                                                                             -DestContext $storageContext[$destStg]

                $queueTime[$destStg] = Get-Date
            }
        }

        while($blobCopies.Count -gt $doneTime.Count)
        {
            foreach($stg in $blobCopies.Keys)
            {
                if($blobCopies[$stg] -eq $null)
                {
                    # Not Started yet
                    continue                
                }

                if($doneTime[$stg] -ne $null)
                {
                    # Already completed
                    continue
                }

                try
                {
                    $status = $blobCopies[$stg] | Get-AzureStorageBlobCopyState -ErrorAction Stop
                }
                catch
                {
                    TraceWarning "Failed to get blob copy state for $stg : $_"
                    continue
                }

                if($status.Status -eq "Pending")
                {
                    $copied = 0
                    if((Get-Member -InputObject $status -Name "BytesCopied") -ne $null)
                    {
                        $copied = $status.BytesCopied
                    }
                    if(($startTime[$stg] -eq $null) -and $copied -gt 0)
                    {
                        $startTime[$stg] = Get-Date
                        TraceInfo "Started copying image to the storage account $otherstg"
                    }
                }
                elseif($status.Status -eq "Success")
                {
                    # Completed
                    if($startTime[$stg] -eq $null)
                    {
                        $startTime[$stg] = Get-Date
                    }
                    $doneTime[$stg] = Get-Date
                    TraceInfo "The VHD of $SourceImageName was copied to the storage account $stg successfully"

                    # Copy to other storage accounts in the same tenant
                    foreach($otherstg in @($destStorageTenants.Values | ?{$_ -contains $stg} | ?{$_ -ne $stg}))
                    {
                        if($blobCopies[$otherstg] -eq $null)
                        {
                            TraceInfo "Start to copy $destVHDName to the storage account $otherstg"
                            $blobCopies[$otherstg] = Start-AzureStorageBlobCopyWithRetry -SrcBlob $destVHDName `
                                                                -SrcContainer $containerName `
                                                                -Context $storageContext[$stg] `
                                                                -DestContainer $containerName `
                                                                -DestBlob $destVHDName `
                                                                -DestContext $storageContext[$otherstg]

                            $queueTime[$otherstg] = Get-Date
                        }
                    }

                    $imageName = $destImageNames[$stg]
                    Add-AzureVMImageWithRetry -ImageName $imageName -MediaLocation $blobCopies[$stg].ICloudBlob.Uri.ToString() -OS Windows `
                            -Label $srcImage.Label -Description $srcImage.Description -PublishedDate $srcImageTimeStamp | Out-Null
                    foreach($svc in @($ServiceVMTable.Keys | ?{$ServiceStorageTable[$_] -eq $stg}))
                    {
                        $provsionJobs += StartHPCNodesForService -VMNames $ServiceVMTable[$svc] -VMSize $VMSize -ServiceName $svc `
                                                -StorageAccountName $stg -VNetName $VNetName -SubnetName $SubnetName -DeploymentLabel $DeploymentLabel `
                                                -LocalAdminCred $LocalAdminCred -DomainUserCred $DomainUserCred -DomainFQDN $DomainFQDN `
                                                -ImageName $imageName -CustomDataFile $hnNameFile
                    }
                }
                else
                {                    
                    throw "Failed to copy the VHD of $SourceImageName to the storage account $stg, the status is $($status.Status)"
                }
            }

            Start-Sleep -Seconds 20
        }
    }
    finally
    {
        foreach($stg in $blobCopies.Keys)
        {
            if($doneTime[$stg] -eq $null)
            {
                try
                {
                    $blobCopies[$stg] | Stop-AzureStorageBlobCopy -ErrorAction Continue
                }
                catch
                {
                    TraceWarning "Failed to stop Azure storage blob copy: $_"
                }
            }
        }
    }

    if($provsionJobs.Count -gt 0)
    {
        while($true)
        {
            $leftJobs = @()
            foreach($job in $provsionJobs)
            {
                $isLeftJob = $false
                if($job.State -eq "Failed")
                {
                    TraceError ("The node provision job failed in cloud service $($job.Name): " + $job.ChildJobs[0].JobStateInfo.Reason.Message)
                }
                elseif($job.State -eq "Blocked")
                {
                    TraceError "The node provision job was blocked in cloud service $($job.Name)" 
                }
                elseif($job.State -eq "Completed")
                {
                    TraceInfo "The node provision job completed in cloud service $($job.Name)" 
                }
                else
                {
                    $isLeftJob = $true
                    $leftJobs += $job
                }

                if(-not $isLeftJob)
                {
                    TraceInfo "The detailed provision logs for cloud service $($job.Name) as below: "
                    TraceInfo "------ Start of VM provisioning in cloud service $($job.Name) ------" 
                    $job.ChildJobs[0].Verbose | Write-Verbose -Verbose
                    TraceInfo "------ End of VM provisioning in cloud service $($job.Name) ------" 
                }
            }
            if($leftJobs.Count -eq 0)
            {
                break
            }
            $provsionJobs = $leftJobs

            Start-Sleep -Seconds 10
        }
    }

    Remove-Item -Path $hnNameFile -Force
}
# SIG # Begin signature block
# MIIavwYJKoZIhvcNAQcCoIIasDCCGqwCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUg09CvH3xt/V1PksSiT0Cv2g/
# S/ugghV6MIIEuzCCA6OgAwIBAgITMwAAAFtDPHJBxWzA7gAAAAAAWzANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTQwNTIzMTcxMzE2
# WhcNMTUwODIzMTcxMzE2WjCBqzELMAkGA1UEBhMCVVMxCzAJBgNVBAgTAldBMRAw
# DgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24x
# DTALBgNVBAsTBE1PUFIxJzAlBgNVBAsTHm5DaXBoZXIgRFNFIEVTTjozMUM1LTMw
# QkEtN0M5MTElMCMGA1UEAxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZTCC
# ASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAJO2gxkc4ZlPSL6curaAqyM+
# cxMZpbyNUkEWY49+BhzsvpPybh2Hv1guJrhjk9NmY/SqrbCrUWX//TK650z4tkWi
# M7sUcRG8kpYzmfF61YdzcGDC+7phlMOg5nGmLtwTGbFq2hpkVe0Ush1SGEaJ7zeu
# lRzx7RxaNj7W8O3EZU3vI0rjTMSQiWu01MqBr8x2Ubfgk6HU/n9P4aVT0jCY1/N3
# TEy+ijg5n2xyysvD32VFFsXQY0OnVRp45SEltH/EJ6gzunUPvJJyxhUdMzwvFSxn
# nr04pWcMkevA7zlstOXaidfEdn+xR/0FQABh70OHGnfHVlQCpqaNtV7rdHPfv0cC
# AwEAAaOCAQkwggEFMB0GA1UdDgQWBBTNDGkkqFEi29Hu02Wg+BiCdff1IzAfBgNV
# HSMEGDAWgBQjNPjZUkZwCu1A+3b7syuwwzWzDzBUBgNVHR8ETTBLMEmgR6BFhkNo
# dHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNyb3Nv
# ZnRUaW1lU3RhbXBQQ0EuY3JsMFgGCCsGAQUFBwEBBEwwSjBIBggrBgEFBQcwAoY8
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNyb3NvZnRUaW1l
# U3RhbXBQQ0EuY3J0MBMGA1UdJQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEBBQUA
# A4IBAQCFZLbdqiTl4PvpaqSQwDnvTfrzf5LEB3/xqY832xz49neml+bIRHEIRH0K
# cJfJx68MK2DEhExOTZgsfvzoFn1H2ygKORzb7OKHfCItu//Rl22AQK3v3SgzzkHZ
# 53xu+bjJKyDaeDclxQR6COuUvJomC/QqzUG9/fyImMQZHFcFElrogLzIiSei6vcV
# UY7azSrz5fs8ySz9spMa2vyM2Tc6t1R+HekTNMUfxQPaD1qVrCmFszlMz58CWcVS
# Kd28O1SByMdZS+2LpzDPTLWVnyZVp2tnhF24jf+5oV+NLk82eOsbX8zMctyL36eb
# +8n+1flHpAnDRDpch5aHPlRovDSvMIIE7DCCA9SgAwIBAgITMwAAAMps1TISNcTh
# VQABAAAAyjANBgkqhkiG9w0BAQUFADB5MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSMwIQYDVQQDExpNaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QTAeFw0xNDA0MjIxNzM5MDBaFw0xNTA3MjIxNzM5MDBaMIGDMQswCQYDVQQGEwJV
# UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
# ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMQ0wCwYDVQQLEwRNT1BSMR4wHAYDVQQD
# ExVNaWNyb3NvZnQgQ29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAw
# ggEKAoIBAQCWcV3tBkb6hMudW7dGx7DhtBE5A62xFXNgnOuntm4aPD//ZeM08aal
# IV5WmWxY5JKhClzC09xSLwxlmiBhQFMxnGyPIX26+f4TUFJglTpbuVildGFBqZTg
# rSZOTKGXcEknXnxnyk8ecYRGvB1LtuIPxcYnyQfmegqlFwAZTHBFOC2BtFCqxWfR
# +nm8xcyhcpv0JTSY+FTfEjk4Ei+ka6Wafsdi0dzP7T00+LnfNTC67HkyqeGprFVN
# TH9MVsMTC3bxB/nMR6z7iNVSpR4o+j0tz8+EmIZxZRHPhckJRIbhb+ex/KxARKWp
# iyM/gkmd1ZZZUBNZGHP/QwytK9R/MEBnAgMBAAGjggFgMIIBXDATBgNVHSUEDDAK
# BggrBgEFBQcDAzAdBgNVHQ4EFgQUH17iXVCNVoa+SjzPBOinh7XLv4MwUQYDVR0R
# BEowSKRGMEQxDTALBgNVBAsTBE1PUFIxMzAxBgNVBAUTKjMxNTk1K2I0MjE4ZjEz
# LTZmY2EtNDkwZi05YzQ3LTNmYzU1N2RmYzQ0MDAfBgNVHSMEGDAWgBTLEejK0rQW
# WAHJNy4zFha5TJoKHzBWBgNVHR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1pY3Jv
# c29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNDb2RTaWdQQ0FfMDgtMzEtMjAx
# MC5jcmwwWgYIKwYBBQUHAQEETjBMMEoGCCsGAQUFBzAChj5odHRwOi8vd3d3Lm1p
# Y3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY0NvZFNpZ1BDQV8wOC0zMS0yMDEwLmNy
# dDANBgkqhkiG9w0BAQUFAAOCAQEAd1zr15E9zb17g9mFqbBDnXN8F8kP7Tbbx7Us
# G177VAU6g3FAgQmit3EmXtZ9tmw7yapfXQMYKh0nfgfpxWUftc8Nt1THKDhaiOd7
# wRm2VjK64szLk9uvbg9dRPXUsO8b1U7Brw7vIJvy4f4nXejF/2H2GdIoCiKd381w
# gp4YctgjzHosQ+7/6sDg5h2qnpczAFJvB7jTiGzepAY1p8JThmURdwmPNVm52Iao
# AP74MX0s9IwFncDB1XdybOlNWSaD8cKyiFeTNQB8UCu8Wfz+HCk4gtPeUpdFKRhO
# lludul8bo/EnUOoHlehtNA04V9w3KDWVOjic1O1qhV0OIhFeezCCBbwwggOkoAMC
# AQICCmEzJhoAAAAAADEwDQYJKoZIhvcNAQEFBQAwXzETMBEGCgmSJomT8ixkARkW
# A2NvbTEZMBcGCgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9z
# b2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5MB4XDTEwMDgzMTIyMTkzMloX
# DTIwMDgzMTIyMjkzMloweTELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0
# b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3Jh
# dGlvbjEjMCEGA1UEAxMaTWljcm9zb2Z0IENvZGUgU2lnbmluZyBQQ0EwggEiMA0G
# CSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCycllcGTBkvx2aYCAgQpl2U2w+G9Zv
# zMvx6mv+lxYQ4N86dIMaty+gMuz/3sJCTiPVcgDbNVcKicquIEn08GisTUuNpb15
# S3GbRwfa/SXfnXWIz6pzRH/XgdvzvfI2pMlcRdyvrT3gKGiXGqelcnNW8ReU5P01
# lHKg1nZfHndFg4U4FtBzWwW6Z1KNpbJpL9oZC/6SdCnidi9U3RQwWfjSjWL9y8lf
# RjFQuScT5EAwz3IpECgixzdOPaAyPZDNoTgGhVxOVoIoKgUyt0vXT2Pn0i1i8UU9
# 56wIAPZGoZ7RW4wmU+h6qkryRs83PDietHdcpReejcsRj1Y8wawJXwPTAgMBAAGj
# ggFeMIIBWjAPBgNVHRMBAf8EBTADAQH/MB0GA1UdDgQWBBTLEejK0rQWWAHJNy4z
# Fha5TJoKHzALBgNVHQ8EBAMCAYYwEgYJKwYBBAGCNxUBBAUCAwEAATAjBgkrBgEE
# AYI3FQIEFgQU/dExTtMmipXhmGA7qDFvpjy82C0wGQYJKwYBBAGCNxQCBAweCgBT
# AHUAYgBDAEEwHwYDVR0jBBgwFoAUDqyCYEBWJ5flJRP8KuEKU5VZ5KQwUAYDVR0f
# BEkwRzBFoEOgQYY/aHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJv
# ZHVjdHMvbWljcm9zb2Z0cm9vdGNlcnQuY3JsMFQGCCsGAQUFBwEBBEgwRjBEBggr
# BgEFBQcwAoY4aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNy
# b3NvZnRSb290Q2VydC5jcnQwDQYJKoZIhvcNAQEFBQADggIBAFk5Pn8mRq/rb0Cx
# MrVq6w4vbqhJ9+tfde1MOy3XQ60L/svpLTGjI8x8UJiAIV2sPS9MuqKoVpzjcLu4
# tPh5tUly9z7qQX/K4QwXaculnCAt+gtQxFbNLeNK0rxw56gNogOlVuC4iktX8pVC
# nPHz7+7jhh80PLhWmvBTI4UqpIIck+KUBx3y4k74jKHK6BOlkU7IG9KPcpUqcW2b
# Gvgc8FPWZ8wi/1wdzaKMvSeyeWNWRKJRzfnpo1hW3ZsCRUQvX/TartSCMm78pJUT
# 5Otp56miLL7IKxAOZY6Z2/Wi+hImCWU4lPF6H0q70eFW6NB4lhhcyTUWX92THUmO
# Lb6tNEQc7hAVGgBd3TVbIc6YxwnuhQ6MT20OE049fClInHLR82zKwexwo1eSV32U
# jaAbSANa98+jZwp0pTbtLS8XyOZyNxL0b7E8Z4L5UrKNMxZlHg6K3RDeZPRvzkbU
# 0xfpecQEtNP7LN8fip6sCvsTJ0Ct5PnhqX9GuwdgR2VgQE6wQuxO7bN2edgKNAlt
# HIAxH+IOVN3lofvlRxCtZJj/UBYufL8FIXrilUEnacOTj5XJjdibIa4NXJzwoq6G
# aIMMai27dmsAHZat8hZ79haDJLmIz2qoRzEvmtzjcT3XAH5iR9HOiMm4GPoOco3B
# oz2vAkBq/2mbluIQqBC0N1AI1sM9MIIGBzCCA++gAwIBAgIKYRZoNAAAAAAAHDAN
# BgkqhkiG9w0BAQUFADBfMRMwEQYKCZImiZPyLGQBGRYDY29tMRkwFwYKCZImiZPy
# LGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQDEyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZp
# Y2F0ZSBBdXRob3JpdHkwHhcNMDcwNDAzMTI1MzA5WhcNMjEwNDAzMTMwMzA5WjB3
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEwHwYDVQQDExhN
# aWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAw
# ggEKAoIBAQCfoWyx39tIkip8ay4Z4b3i48WZUSNQrc7dGE4kD+7Rp9FMrXQwIBHr
# B9VUlRVJlBtCkq6YXDAm2gBr6Hu97IkHD/cOBJjwicwfyzMkh53y9GccLPx754gd
# 6udOo6HBI1PKjfpFzwnQXq/QsEIEovmmbJNn1yjcRlOwhtDlKEYuJ6yGT1VSDOQD
# LPtqkJAwbofzWTCd+n7Wl7PoIZd++NIT8wi3U21StEWQn0gASkdmEScpZqiX5NMG
# gUqi+YSnEUcUCYKfhO1VeP4Bmh1QCIUAEDBG7bfeI0a7xC1Un68eeEExd8yb3zuD
# k6FhArUdDbH895uyAc4iS1T/+QXDwiALAgMBAAGjggGrMIIBpzAPBgNVHRMBAf8E
# BTADAQH/MB0GA1UdDgQWBBQjNPjZUkZwCu1A+3b7syuwwzWzDzALBgNVHQ8EBAMC
# AYYwEAYJKwYBBAGCNxUBBAMCAQAwgZgGA1UdIwSBkDCBjYAUDqyCYEBWJ5flJRP8
# KuEKU5VZ5KShY6RhMF8xEzARBgoJkiaJk/IsZAEZFgNjb20xGTAXBgoJkiaJk/Is
# ZAEZFgltaWNyb3NvZnQxLTArBgNVBAMTJE1pY3Jvc29mdCBSb290IENlcnRpZmlj
# YXRlIEF1dGhvcml0eYIQea0WoUqgpa1Mc1j0BxMuZTBQBgNVHR8ESTBHMEWgQ6BB
# hj9odHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9taWNy
# b3NvZnRyb290Y2VydC5jcmwwVAYIKwYBBQUHAQEESDBGMEQGCCsGAQUFBzAChjho
# dHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY3Jvc29mdFJvb3RD
# ZXJ0LmNydDATBgNVHSUEDDAKBggrBgEFBQcDCDANBgkqhkiG9w0BAQUFAAOCAgEA
# EJeKw1wDRDbd6bStd9vOeVFNAbEudHFbbQwTq86+e4+4LtQSooxtYrhXAstOIBNQ
# md16QOJXu69YmhzhHQGGrLt48ovQ7DsB7uK+jwoFyI1I4vBTFd1Pq5Lk541q1YDB
# 5pTyBi+FA+mRKiQicPv2/OR4mS4N9wficLwYTp2OawpylbihOZxnLcVRDupiXD8W
# mIsgP+IHGjL5zDFKdjE9K3ILyOpwPf+FChPfwgphjvDXuBfrTot/xTUrXqO/67x9
# C0J71FNyIe4wyrt4ZVxbARcKFA7S2hSY9Ty5ZlizLS/n+YWGzFFW6J1wlGysOUzU
# 9nm/qhh6YinvopspNAZ3GmLJPR5tH4LwC8csu89Ds+X57H2146SodDW4TsVxIxIm
# dgs8UoxxWkZDFLyzs7BNZ8ifQv+AeSGAnhUwZuhCEl4ayJ4iIdBD6Svpu/RIzCzU
# 2DKATCYqSCRfWupW76bemZ3KOm+9gSd0BhHudiG/m4LBJ1S2sWo9iaF2YbRuoROm
# v6pH8BJv/YoybLL+31HIjCPJZr2dHYcSZAI9La9Zj7jkIeW1sMpjtHhUBdRBLlCs
# lLCleKuzoJZ1GtmShxN1Ii8yqAhuoFuMJb+g74TKIdbrHk/Jmu5J4PcBZW+JC33I
# acjmbuqnl84xKf8OxVtc2E0bodj6L54/LlUWa8kTo/0xggSvMIIEqwIBATCBkDB5
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSMwIQYDVQQDExpN
# aWNyb3NvZnQgQ29kZSBTaWduaW5nIFBDQQITMwAAAMps1TISNcThVQABAAAAyjAJ
# BgUrDgMCGgUAoIHIMBkGCSqGSIb3DQEJAzEMBgorBgEEAYI3AgEEMBwGCisGAQQB
# gjcCAQsxDjAMBgorBgEEAYI3AgEVMCMGCSqGSIb3DQEJBDEWBBRiBMs0yTns4Z9F
# IaiMsP/5+3NIujBoBgorBgEEAYI3AgEMMVowWKA2gDQATQBpAGMAcgBvAHMAbwBm
# AHQAIABIAFAAQwAgAFAAYQBjAGsAIAAyADAAMQAyACAAUgAyoR6AHGh0dHA6Ly93
# d3cubWljcm9zb2Z0LmNvbS9IUEMwDQYJKoZIhvcNAQEBBQAEggEAjzzvzNC3PggN
# PRlxk7FlafH8VD2Lq/PBrFky2cnYMcI1B0S9u8qaq8zAwvZC+F8jOPd+Q+dz3MXw
# ePqlBfK/ilp5kEwd2LIT0HtKW0DKpJ5y9YZrvo2Jrf8JPJCH75OdD9DR6P4B+uHe
# 0x5DcPvYdElBKc4M7eXwFlru5caDbPq5m5vEgV/VGV9inUN22dX4YCeU2j3FrFjR
# RKkdRYIqgPYhMjFkR6xA8IO1Y6z7h8GV6uKJnaEOy/lBUcGnQ8HWya8H4tM556SK
# 5NDjz4i29zyeT/6YdLvdJP/vLsv3c+8uHCJGH3KHMid027aX/QcVqjKVqIiEvfH0
# G1CxERgGS6GCAigwggIkBgkqhkiG9w0BCQYxggIVMIICEQIBATCBjjB3MQswCQYD
# VQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEe
# MBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEwHwYDVQQDExhNaWNyb3Nv
# ZnQgVGltZS1TdGFtcCBQQ0ECEzMAAABbQzxyQcVswO4AAAAAAFswCQYFKw4DAhoF
# AKBdMBgGCSqGSIb3DQEJAzELBgkqhkiG9w0BBwEwHAYJKoZIhvcNAQkFMQ8XDTE1
# MDIxMjE4MTQ1M1owIwYJKoZIhvcNAQkEMRYEFNONr3JBKn9IE/iptOr2JuDzX31u
# MA0GCSqGSIb3DQEBBQUABIIBABDLe5HrL2MxV1PjZ1m7MI+XLjrM1wfYjooHt7e+
# nNoX5/flFl7sNn9PwMNdWOBRdC9+EA6ZuBxvc2VFOfg27E3XZVHUK8JjLdVdu1P9
# knHVV2fRJiTlufYuJbPxScLvNV6BsQBg8ATLJDRKZNpnQ18RIyXfDhh+71fVV3v8
# HTvB36tvOefcGVpWbp1ckpVndMHVqx32niPmPKFMpm7PXVargXywi1a+L4OvGqKn
# IWh8We80C35h9ExlAGzJlJxjwJpaAtIMPnhVWx/DwfOv6epb0WKjUnAuqNn9duER
# lyfWCnfPg2uMy2ZqOcN6B4IIv8DxTxS8cGOdWnJ0bwz9aSw=
# SIG # End signature block
