﻿<#
.SYNOPSIS
  Invoke an Azure cmdlet with retries
.DESCRIPTION
  Invoke an Azure cmdlet with retries. The Azure cmdlet must be a "query" command (typically, Get-Azurexxx or Test-Azurexxx), must not modify any Azure resources. 
.EXAMPLE
  InvokeAzureCmdWithRetry -Command "Get-AzureVMImage"
  InvokeAzureCmdWithRetry -Command "Get-AzureDeployment -ServiceName testservice" -RetryTimes 10
#>
function InvokeAzureCmdWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Command,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    # if ErrorAction is not specified, specify it as Stop so that the error can be thrown by "Invoke-Expression" command.
    if(($Command -notmatch " -ErrorAction ") -and ($Command -notmatch " -EA "))
    {
        $newCommand = $Command + " -ErrorAction Stop"
    }
    else
    {
        $newCommand = $Command
    }
    $retry = 0
    while($true)
    {
        try
        {
            $ret = Invoke-Expression -Command $newCommand
            return $ret
        }
        catch
        {
            $caughtException = $_.Exception
            $errMsg = "Failed to execute the command '$Command': " + ($_ | Out-String)
            $shouldThrow = $true
            if($caughtException -is [Microsoft.WindowsAzure.CloudException])
            {
                if($_.Exception.ErrorCode -eq "ResourceNotFound")
                {
                    TraceInfo "The command '$Command' executed: the resource not found."
                    return
                }

                $shouldThrow = $false
            }
            elseif(($caughtException -is [System.Net.Http.HttpRequestException]) -or ($caughtException -is [System.Threading.Tasks.TaskCanceledException]))
            {
                $shouldThrow = $false
            }
        }

        if($shouldThrow -or ($retry -ge $RetryTimes))
        {
            TraceError $errMsg
            throw $caughtException
        }
        else
        {
            TraceWarning $errMsg
            TraceInfo "Retry to execute the command '$Command' after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
    }
}

function New-AzureServiceWithRetry
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true, ParameterSetName="InAffinityGroup")]
        [String] $AffinityGroupName,

        [Parameter(Mandatory=$true, ParameterSetName="InRegion")]
        [String] $Location,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    $retry = 0
    while($true)
    {
        try
        {
            if($PsCmdlet.ParameterSetName -eq "InAffinityGroup")
            {
                TraceInfo "Creating the cloud service $ServiceName in the affinity group $AffinityGroupName"
                New-AzureService -ServiceName $ServiceName -AffinityGroup $AffinityGroupName -ErrorAction Stop | Out-Null
                TraceInfo "The cloud service $ServiceName created in the affinity group $AffinityGroupName"
            }
            else
            {
                TraceInfo "Creating the cloud service $ServiceName in the region $Location"
                New-AzureService -ServiceName $ServiceName -Location $Location -ErrorAction Stop | Out-Null
                TraceInfo "The cloud service $ServiceName created in the region $Location"
            }
            return
        }
        catch
        {
            TraceWarning ("Failed to create cloud service $ServiceName :" + ($_ | Out-String))
        }
        
        $service = InvokeAzureCmdWithRetry -Command "Get-AzureService -ServiceName $ServiceName"
        if($service -ne $null)
        {
            TraceInfo "Cloud service $ServiceName Created"
            return
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to create cloud service $ServiceName after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
        else
        {
            throw "Failed to create cloud service $ServiceName  after $RetryTimes retries"
        }
  
    }
}

function New-AzureVMWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [Microsoft.WindowsAzure.Commands.ServiceManagement.Model.PersistentVM] $VM,

        [Parameter(Mandatory=$true)]
        [String] $VNetName,

        [Parameter(Mandatory=$true)]
        [String] $DeploymentLabel,

        [Parameter(Mandatory=$false)]
        [Switch] $WaitForBoot,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    $retry = 0
    $vmname = $VM.RoleName
    while($true)
    {
        try
        {
            TraceInfo "Creating the virtual machine $vmname in cloud service $ServiceName"
            if($WaitForBoot.IsPresent)
            {
                New-AzureVM -VMs $VM -ServiceName $ServiceName -DeploymentLabel $DeploymentLabel -VNetName $VNetName -WaitForBoot -ErrorAction Stop -WarningAction SilentlyContinue | Out-Null
            }
            else
            {
                New-AzureVM -VMs $VM -ServiceName $ServiceName -DeploymentLabel $DeploymentLabel -VNetName $VNetName -ErrorAction Stop -WarningAction SilentlyContinue | Out-Null
            }

            TraceInfo "The virtual machine $vmname Created in cloud service $ServiceName"           
            return
        }
        catch
        {
            TraceWarning ("Failed to create the virtual machine $vmname : " + ($_ | Out-String))
        }
        
        $newvm = InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $ServiceName -Name $vmname"
        if($newvm -ne $null)
        {
            TraceInfo "The virtual machine $vmname Created in cloud service $ServiceName"
            return
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to create the virtual machine $vmname after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
        else
        {
            throw "Failed to create the virtual machine $vmname after $RetryTimes retries"
        }
    }
}

function Update-AzureVMWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [String] $Name,

        [Parameter(Mandatory=$true)]
        [Microsoft.WindowsAzure.Commands.ServiceManagement.Model.PersistentVM] $VM,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    $retry = 0
    while($true)
    {
        try
        {
            TraceInfo "Updating the virtual machine $Name in cloud service $ServiceName ..."
            Update-AzureVM -Name $Name -ServiceName $ServiceName -VM $VM -ErrorAction Stop -WarningAction SilentlyContinue | Out-Null
            TraceInfo "The virtual machine $Name updated."           
            return
        }
        catch
        {
            TraceWarning ("Failed to update the virtual machine $Name : " + ($_ | Out-String))
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to update the virtual machine $Name after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
        else
        {
            throw "Failed to update the virtual machine $Name after $RetryTimes retries"
        }
    }
}

function New-AzureAffinityGroupWithRetry
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $Name,

        [Parameter(Mandatory=$true)]
        [String] $Location,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    $retry = 0
    while($true)
    {
        try
        {
            TraceInfo "Creating the affinity group $Name at region $Location"
            New-AzureAffinityGroup -Name $Name -Location $Location -Label $Name -ErrorAction Stop | Out-Null
            TraceInfo "The affinity group $Name created at region $Location"
            return
        }
        catch
        {
            TraceWarning ("Failed to create affinity group $Name : " + ($_ | Out-String))
        }
        
        $ag = InvokeAzureCmdWithRetry -Command "Get-AzureAffinityGroup -Name $Name"
        if($ag -ne $null)
        {
            TraceInfo "The affinity group $Name Created"
            return
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to create the affinity group $Name after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
        else
        {
            throw "Failed to create the affinity group $Name after $RetryTimes retries"
        }
    }
}

function New-AzureStorageAccountWithRetry
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $StorageAccountName,

        [Parameter(Mandatory=$true)]
        [String] $Location,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    $retry = 0
    while($true)
    {
        try
        {
            TraceInfo "Creating the storage account $StorageAccountName at region $Location"
            New-AzureStorageAccount -StorageAccountName $StorageAccountName -Location $Location -ErrorAction Stop | Out-Null
            TraceInfo "Storage account $StorageAccountName created at region $Location"
            return
        }
        catch
        {
            TraceWarning ("Failed to create the storage account $StorageAccountName : " + ($_ | Out-String))
        }
        
        $storage = InvokeAzureCmdWithRetry -Command "Get-AzureStorageAccount -StorageAccountName $StorageAccountName"
        if($storage -ne $null)
        {
            TraceInfo "Storage account $StorageAccountName Created"
            return
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to create the storage account $StorageAccountName after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++  
        }
        else
        {
            throw "Failed to create the storage account $StorageAccountName  after $RetryTimes retries"
        }
    }
}

function Save-AzureVMImageWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Name,

        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [String] $ImageName,
        
        [Parameter(Mandatory=$false)]
        [String] $ImageLabel = "HPC Image",
        
        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    $retry = 0
    while($true)
    {
        try
        {
            TraceInfo "Saving Azure VM Image $ImageName"
            Save-AzureVMImage -Name $Name -ServiceName $ServiceName -ImageName $ImageName -ImageLabel $ImageLabel -OSState Generalized -ErrorAction Stop | Out-Null
            TraceInfo "The VM image $ImageName was saved"
            return
        }
        catch
        {
            TraceWarning ("Failed to save the VM image $ImageName :" + ($_ | Out-String))
        }

        $image = InvokeAzureCmdWithRetry -Command "Get-AzureVMImage -ImageName $ImageName"
        if($image -ne $null)
        {
            TraceInfo "The VM image $ImageName already saved"
            return
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to save the VM image $ImageName after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
        else
        {
            throw "Failed to save the VM image $ImageName after $RetryTimes retries"
        }
    }
}

function Add-AzureVMImageWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ImageName,

        [Parameter(Mandatory=$true)]
        [String] $MediaLocation,

        [Parameter(Mandatory=$true)]
        [ValidateSet("Linux","Windows")]
        [String] $OS,
        
        [Parameter(Mandatory=$false)]
        [String] $Label,

        [Parameter(Mandatory=$false)]
        [String] $Description,

        [Parameter(Mandatory=$false)]
        [System.DateTime] $PublishedDate,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    if([String]::IsNullOrEmpty($Label))
    {
        $Label = $ImageName
    }
    if([String]::IsNullOrEmpty($Description))
    {
        $Description = $Label
    }
    
    if($PublishedDate -eq $null)
    {
        $PublishedDate = Get-Date
    }

    $retry = 0
    while($true)
    {
        try
        {
            TraceInfo "Adding Azure VM Image $ImageName"
            Add-AzureVMImage -ImageName $ImageName -MediaLocation $MediaLocation -OS $OS -Label $Label -Description $Description -PublishedDate $PublishedDate -ErrorAction Stop | Out-Null
            TraceInfo "The VM image $ImageName was successfully added"
            return
        }
        catch
        {
            TraceWarning ("Failed to add the VM image $ImageName :" + ($_ | Out-String))
        }

        $image = InvokeAzureCmdWithRetry -Command "Get-AzureVMImage -ImageName $ImageName"
        if($image -ne $null)
        {
            TraceInfo "The VM image $ImageName was already added"
            return
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to add the VM image $ImageName after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
        else
        {
            throw "Failed to add the VM image $ImageName after $RetryTimes retries"
        }
    }
}

function Remove-AzureVMImageWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ImageName,

        [Parameter(Mandatory=$false)]
        [Switch] $DeleteVHD,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 20
    )

    $retry = 0
    while($true)
    {
        try
        {
            TraceInfo "Removing Azure VM Image $ImageName"
            Remove-AzureVMImage -ImageName $ImageName -DeleteVHD:$DeleteVHD -ErrorAction Stop | Out-Null
            TraceInfo "The VM image $ImageName was successfully removed"
            return
        }
        catch
        {
            if($_.Exception -is [Microsoft.WindowsAzure.CloudException])
            {
                if($_.Exception.ErrorCode -eq "ResourceNotFound")
                {
                    TraceInfo "The VM image $ImageName was already removed"
                    return
                }
            }

            TraceWarning ("Failed to remove the VM image $ImageName :" + ($_ | Out-String))
        }

        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to remove the VM image $ImageName after 10 seconds"
            Start-Sleep -Seconds 10
            $retry++
        }
        else
        {
            throw "Failed to remove the VM image $ImageName after $RetryTimes retries"
        }
    }
}

function Get-AzureVNetSiteWithRetry
{
    Param
    (
        # The Name
        [Parameter(Mandatory=$true)]
        [String] $VNetName
    )
    
    $vnetSite = $null
    try
    {
        $vnetSite = InvokeAzureCmdWithRetry -Command "Get-AzureVNetSite -VNetName $VNetName"
    }
    catch
    {
    }

    return $vnetSite
}

   
function Remove-AzureVMWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$true)]
        [String] $VMName,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 3
    )

    $retry = 0
    TraceInfo "Trying to remove the virtual machine $VMName from the cloud service $ServiceName"
    while($true)
    {
        try
        {
            TraceInfo "Removing the virtual machine $VMName from the cloud service $ServiceName"
            Remove-AzureVM -Name $VMName -ServiceName $ServiceName -DeleteVHD -ErrorAction Stop -WarningAction SilentlyContinue
            TraceInfo "The virtual machine $VMName removed from the cloud service $ServiceName"
            return
        }
        catch
        {
            TraceWarning ("Failed to remove the virtual machine $VMName : " + ($_ | Out-String))
        }

        $vm = InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $ServiceName -Name $VMName -WarningAction SilentlyContinue"
        if($vm -eq $null)
        {
            TraceInfo "The virtual machine $VMName already removed from the cloud service $ServiceName"
            return
        }
        
        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to remove the virtual machine $VMName after 5 seconds"
            Start-Sleep -Seconds 5
            $retry++
        }
        else
        {
            throw "Failed to remove the virtual machine $VMName after $RetryTimes retries"
        }
    }
}

function Remove-AzureServiceWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $ServiceName,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 3
    )

    $retry = 0
    TraceInfo "Trying to remove the cloud service $ServiceName"
    while($true)
    {
        try
        {
            TraceInfo "Removing the cloud service $ServiceName"
            Remove-AzureService -ServiceName $ServiceName -DeleteAll -Force -ErrorAction Stop -WarningAction SilentlyContinue
            TraceInfo "The cloud service $ServiceName removed"
            return
        }
        catch
        {
            TraceWarning ("Failed to remove the cloud service $ServiceName : " + ($_ | Out-String))
        }

        $svc = InvokeAzureCmdWithRetry -Command "Get-AzureService -ServiceName $ServiceName"
        if($svc -eq $null)
        {
            TraceInfo "The cloud service $ServiceName already removed"
            return
        }
                
        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to remove the cloud service $ServiceName after 5 seconds"
            Start-Sleep -Seconds 5
            $retry++
        }
        else
        {
            throw "Failed to remove the cloud service $ServiceName after $RetryTimes retries"
        }
    }
}

function GetStorageAccountFromMediaLink
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $MediaLink
    )

    if($MediaLink -ne $null)
    {
        $startIndex = $MediaLink.IndexOf("//") + 2
        $endIndex = $MediaLink.IndexOf(".")
        if(($endIndex -gt ($startIndex + 2)) -and ($startIndex -gt 0))
        {
            return $MediaLink.Substring($startIndex, $endIndex - $startIndex)
        }
    }

    return ""
}


function DecodeLabelIfBase64Encoded
{
    param($orgLabel)

    try
    {
        $bytes = [System.Convert]::FromBase64String($orgLabel)
        return [System.Text.Encoding]::UTF8.GetString($bytes)
    }
    catch
    {
        return $orgLabel
    }
}


<#
.SYNOPSIS
  Check whether a VM exists in the given cloud service
.DESCRIPTION
  Check whether a VM exists in the given cloud service
.EXAMPLE
  VMExists -Name "vmname" -ServiceName "servicename"
#>
function VMExists 
{
    param
    (
        #The name of VM
        [Parameter(Mandatory=$true)]
        [String] $Name, 

        #The name of the host cloud service
        [Parameter(Mandatory=$true)]
        [String] $ServiceName
    )

    $vm = InvokeAzureCmdWithRetry -Command "Get-AzureVM -Name $Name -ServiceName $ServiceName"
    if ($vm -ne $null) 
    {
        return $true
    }
    
    return $false
}

<#
.SYNOPSIS
  Wait the specified VMs in a cloud service to become the target status
.DESCRIPTION
  Wait the specified VMs to become the target status in a given maximum waiting time, throw if the timer expires
.EXAMPLE
  The following example will wait the VM "mytestvm" in cloud service "mytestservice" to become 'ReadyRole' for a maximum time 900 seconds
    WaitForVMReady -VMName @("testvm1", "testvm2") -ServiceName "mytestservice" -MaxWaitSeconds 900
#>
function WaitForVMReady 
{
    param 
    (
        [Parameter(Mandatory=$true)]
        [String[]] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$false)]
        [String] $TargetStatus = "ReadyRole", 

        [Parameter(Mandatory=$false)]
        [ValidateRange(0, 3600)]
        [int] $MaxWaitSeconds = 1800
    )

    TraceInfo "Waiting for the following VM(s) to become $TargetStatus status: $VMName"
    $noReadyVMNames = $VMName
    $beginTime = Get-Date
    while ($noReadyVMNames.Count -gt 0)
    {
        $vms = @(InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $ServiceName -WarningAction SilentlyContinue" | ?{($noReadyVMNames -contains $_.InstanceName) -and ($_.InstanceStatus -eq $TargetStatus)})
        if($vms.Count -gt 0)
        {
            TraceInfo "The following VM(s) are now in $TargetStatus status: $($vms.InstanceName)"
            if($vms.Count -eq $noReadyVMNames.Count)
            {
                break
            }

            $noReadyVMNames = @($noReadyVMNames | ?{$vms.InstanceName -notcontains $_})
        }

        $timeSpan = (Get-Date) - $beginTime
        $elapsedSeconds = [int]$timeSpan.TotalSeconds
        if ($elapsedSeconds -ge $MaxWaitSeconds)
        {
            throw "Timed out waiting for the following virtual machine(s) to become $TargetStatus : $noReadyVMNames"
        }

        Start-Sleep -Seconds 10
    }
}

function GetVMSizesWithRDMA
{
    return @("A8", "A9")
}

function IsVMSizeWithRDMA
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $VMSize
    )
    
    $sizes = @(GetVMSizesWithRDMA)
    return ($sizes -contains $VMSize)
}

function CreateDomainJoinedVM
{
    param
    (
        [Parameter(Mandatory=$true)]
        [String] $VMName, 

        [Parameter(Mandatory=$true)]
        [String] $ServiceName, 

        [Parameter(Mandatory=$true)]
        [String] $VMSize, 

        [Parameter(Mandatory=$true)]
        [String] $ImageName, 
        
        [Parameter(Mandatory=$true)]
        [String] $VNetName, 
        
        [Parameter(Mandatory=$true)]
        [String] $SubnetName, 

        [Parameter(Mandatory=$true)]
        [PSCredential] $LocalAdminCred, 
        
        [Parameter(Mandatory=$true)]
        [PSCredential] $DomainUserCred,

        [Parameter(Mandatory=$true)]
        [String] $DomainFQDN, 

        [Parameter(Mandatory=$true)]
        [String] $DeploymentLabel, 

        [Parameter(Mandatory=$false)]
        [ValidateSet("ReadOnly","ReadWrite")]
        [String] $HostCaching = "",

        [Parameter(Mandatory=$false)]
        [String] $CustomDataFile = "",

        [Parameter(Mandatory=$false)]
        [Switch] $DisableGuestAgent,

        [Parameter(Mandatory=$false)]
        [Switch] $Wait
    )

    TraceInfo "Creating a domain joined VM $VMName in Service $ServiceName `n Size:  `t$VMSize`n Image: `t$ImageName`n VNet:  `t$VNetName`n Subnet:`t$SubnetName`n Domain:`t$DomainFQDN"

    $netbios = $DomainFQDN.Split(".")[0].ToUpper()
    $domainUserName = $DomainUserCred.UserName
    if($DomainUserCred.UserName.Contains("\"))
    {
        $domainUserName = $domainUserName.Split("\")[1]
    }

    try
    {
        if([String]::IsNullOrEmpty($HostCaching))
        {
            $vm = New-AzureVMConfig -Name $VMName -InstanceSize $VMSize -ImageName $ImageName
        }
        else
        {
            $vm = New-AzureVMConfig -Name $VMName -InstanceSize $VMSize -HostCaching $HostCaching -ImageName $ImageName
        }    

        if ([String]::IsNullOrEmpty($CustomDataFile))
        {
            $vm = $vm | `
                  Add-AzureProvisioningConfig -AdminUsername $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                      -Domain $netbios -DomainUserName $domainUserName -DomainPassword $DomainUserCred.GetNetworkCredential().Password `
                      -JoinDomain $DomainFQDN -WindowsDomain -DisableGuestAgent:$DisableGuestAgent
        }
        else
        {
            $vm = $vm | `
                  Add-AzureProvisioningConfig -AdminUsername $LocalAdminCred.UserName -Password $LocalAdminCred.GetNetworkCredential().Password `
                      -Domain $netbios -DomainUserName $domainUserName -DomainPassword $DomainUserCred.GetNetworkCredential().Password `
                      -JoinDomain $DomainFQDN -WindowsDomain -CustomDataFile $CustomDataFile -DisableGuestAgent:$DisableGuestAgent
        }

        $vm = $vm | Set-AzureSubnet -SubnetNames $SubnetName

        if(-not $DisableGuestAgent.IsPresent)
        {
            $vm = $vm | Set-AzureVMBGInfoExtension

            if(IsVMSizeWithRDMA -VMSize $VMSize)
            {
                $vm = $vm | Set-AzureVMExtension -ExtensionName "HpcVmDrivers" -Publisher "Microsoft.HpcCompute" -Version "1.*"
            }
        }
        else
        {
            if(IsVMSizeWithRDMA -VMSize $VMSize)
            {
                TraceWarning "The RDMA network for $VMName cannot work because Azure VM extension 'HpcVmDrivers' is not enabled."
            }
        }

        New-AzureVMWithRetry -VM $vm -ServiceName $ServiceName -VNetName $VNetName -DeploymentLabel $DeploymentLabel
    }
    catch
    {
        TraceError "Failed to create a VM $VMName joined to domain $DomainFQDN in cloud service $ServiceName"
        throw
    }

    if ($Wait.IsPresent)
    {
        WaitForVMReady -VMName $VMName -ServiceName $ServiceName
    }
}


function ConvertSecureStrToPlain 
{
    param
    ( 
        [Parameter(Mandatory=$true)]
        [System.Security.SecureString] $SecurePassword
    )
    
    $pswPointer = [Runtime.InteropServices.Marshal]::SecureStringToBSTR($SecurePassword)
    $plainPsw = [Runtime.InteropServices.Marshal]::PtrToStringAuto($pswPointer)
    [Runtime.InteropServices.Marshal]::ZeroFreeBSTR($pswPointer)
    return $plainPsw
}

function GetAzureSubscription
{
    param
    (
        [Parameter(Mandatory=$false)]
        [string] $SubscriptionName = ""
    )

    $retry = 0
    while($true)
    {
        try
        {
            if([String]::IsNullOrEmpty($SubscriptionName))
            {
                $sub = @(Get-AzureSubscription -ErrorAction Stop)
            }
            else
            {
                $sub = @(Get-AzureSubscription -SubscriptionName $subscriptionName -ErrorAction Stop)
            }
            
            return $sub
        }
        catch
        {
            if($retry -lt 10)
            {
                TraceWarning "Failed to get azure subscription, will retry after 5 seconds: $_"
                Start-Sleep -Seconds 5
                $retry++
            }
            else
            {
                throw
            }
        }
    }
}

function Get-AzureStorageContainerWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Name,

        [Parameter(Mandatory=$true)]
        $Context,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 10
    )

    $retry = 0
    while($true)
    {
        try
        {
            $container = Get-AzureStorageContainer -Name $Name -Context $Context -ErrorAction Stop
            return $container
        }
        catch
        {
            if($_.Exception -is [Microsoft.WindowsAzure.CloudException])
            {
                if($_.Exception.ErrorCode -eq "ResourceNotFound")
                {
                    return
                }
            }
            elseif($_.Exception -is [Microsoft.WindowsAzure.Commands.Storage.Common.ResourceNotFoundException])
            {
                return
            }

            if($retry -lt $RetryTimes)
            {
                TraceInfo ("Failed to get the storage container $Name :" + ($_ | Out-String))
                Start-Sleep -Seconds 10
                $retry++
            }
            else
            {
                TraceError ("Failed to get the storage container $Name :" + ($_ | Out-String))
                throw
            }
        }
    }
}

function New-AzureStorageContainerWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true)]
        [String] $Name,

        [Parameter(Mandatory=$true)]
        $Context,

        [Parameter(Mandatory=$false)]
        $Permission = "Off",

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 10
    )

    $stgName = $Context.StorageAccountName
    $retry = 0
    while($true)
    {
        try
        {
            TraceInfo "Creating the container $Name in the storage account $stgName"
            $container = New-AzureStorageContainer -Name $Name -Context $Context -Permission $Permission -ErrorAction Stop
            TraceInfo "The container $Name was created in the storage account $stgName"
            return $container
        }
        catch
        {
            TraceWarning ("Failed to create container $Name in the storage account $stgName : " + ($_ | Out-String))
        }

        $container = Get-AzureStorageContainerWithRetry -Name $Name -Context $Context
        if($container -ne $null)
        {
            TraceInfo "The container $Name was already created in the storage account $stgName"
            return $container
        }
                
        if($retry -lt $RetryTimes)
        {
            TraceInfo "Retry to create the container $Name in the storage account $stgName after 5 seconds"
            Start-Sleep -Seconds 5
            $retry++
        }
        else
        {
            throw "Failed to create the container $Name in the storage account $stgName after $RetryTimes retries"
        }
    }
}

function Start-AzureStorageBlobCopyWithRetry
{
    Param
    (
        [Parameter(Mandatory=$true, ParameterSetName="BlobUri")]
        [alias("SrcUri")]
        [String] $AbsoluteUri,

        [Parameter(Mandatory=$true, ParameterSetName="BlobName")]
        [String] $SrcBlob,

        [Parameter(Mandatory=$true, ParameterSetName="BlobName")]
        [String] $SrcContainer,

        [Parameter(Mandatory=$true)]
        [String] $DestBlob,

        [Parameter(Mandatory=$true)]
        [String] $DestContainer,

        [Parameter(Mandatory=$true)]
        $Context,

        [Parameter(Mandatory=$true)]
        $DestContext,

        [Parameter(Mandatory=$false)]
        [Int] $RetryTimes = 10
    )

    $srcStgName = $Context.StorageAccountName
    $destStgName = $DestContext.StorageAccountName
    $retry = 0
    while($true)
    {
        try
        {
            if($PsCmdlet.ParameterSetName -eq "BlobUri")
            {            
                TraceInfo "Queuing the blob copy from $AbsoluteUri to [$destStgName, $DestContainer, $DestBlob]"
                $blobCopy = Start-AzureStorageBlobCopy -AbsoluteUri $AbsoluteUri -DestContainer $DestContainer -DestBlob $DestBlob `
                        -Context $Context -DestContext $DestContext -Force -ErrorAction Stop
                TraceInfo "The blob copy from $AbsoluteUri to [$destStgName, $DestContainer, $DestBlob] was successfully queued"
                return $blobCopy
            }
            else
            {
                TraceInfo "Queuing the blob copy from [$srcStgName, $SrcContainer, $SrcBlob] to [$destStgName, $DestContainer, $DestBlob]"
                $blobCopy = Start-AzureStorageBlobCopy -SrcContainer $SrcContainer -SrcBlob $SrcBlob -DestContainer $DestContainer -DestBlob $DestBlob `
                        -Context $Context -DestContext $DestContext -Force -ErrorAction Stop
                TraceInfo "The blob copy from [$srcStgName, $SrcContainer, $SrcBlob] to [$destStgName, $DestContainer, $DestBlob] was successfully queued"
                return $blobCopy                
            }
        }
        catch
        {
            $noRetry = $false
            if($_.Exception -is [Microsoft.WindowsAzure.Storage.StorageException])
            {
                $noRetry = $true
            }

            
            if((-not $noRetry) -and ($retry -lt $RetryTimes))
            {
                TraceWarning ("Failed to queue the blob copy to [$destStgName, $DestContainer, $DestBlob]: " + ($_ | Out-String))
                Start-Sleep -Seconds 5
                $retry++
            }
            else
            {
                TraceError ("Failed to queue the blob copy to [$destStgName, $DestContainer, $DestBlob]: " + ($_ | Out-String))
                throw
            }
        }
    }
}

# SIG # Begin signature block
# MIIavwYJKoZIhvcNAQcCoIIasDCCGqwCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUQTRCm18/fUfGmDINeSC3G4dY
# J9ugghV6MIIEuzCCA6OgAwIBAgITMwAAAFtDPHJBxWzA7gAAAAAAWzANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTQwNTIzMTcxMzE2
# WhcNMTUwODIzMTcxMzE2WjCBqzELMAkGA1UEBhMCVVMxCzAJBgNVBAgTAldBMRAw
# DgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24x
# DTALBgNVBAsTBE1PUFIxJzAlBgNVBAsTHm5DaXBoZXIgRFNFIEVTTjozMUM1LTMw
# QkEtN0M5MTElMCMGA1UEAxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZTCC
# ASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAJO2gxkc4ZlPSL6curaAqyM+
# cxMZpbyNUkEWY49+BhzsvpPybh2Hv1guJrhjk9NmY/SqrbCrUWX//TK650z4tkWi
# M7sUcRG8kpYzmfF61YdzcGDC+7phlMOg5nGmLtwTGbFq2hpkVe0Ush1SGEaJ7zeu
# lRzx7RxaNj7W8O3EZU3vI0rjTMSQiWu01MqBr8x2Ubfgk6HU/n9P4aVT0jCY1/N3
# TEy+ijg5n2xyysvD32VFFsXQY0OnVRp45SEltH/EJ6gzunUPvJJyxhUdMzwvFSxn
# nr04pWcMkevA7zlstOXaidfEdn+xR/0FQABh70OHGnfHVlQCpqaNtV7rdHPfv0cC
# AwEAAaOCAQkwggEFMB0GA1UdDgQWBBTNDGkkqFEi29Hu02Wg+BiCdff1IzAfBgNV
# HSMEGDAWgBQjNPjZUkZwCu1A+3b7syuwwzWzDzBUBgNVHR8ETTBLMEmgR6BFhkNo
# dHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNyb3Nv
# ZnRUaW1lU3RhbXBQQ0EuY3JsMFgGCCsGAQUFBwEBBEwwSjBIBggrBgEFBQcwAoY8
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNyb3NvZnRUaW1l
# U3RhbXBQQ0EuY3J0MBMGA1UdJQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEBBQUA
# A4IBAQCFZLbdqiTl4PvpaqSQwDnvTfrzf5LEB3/xqY832xz49neml+bIRHEIRH0K
# cJfJx68MK2DEhExOTZgsfvzoFn1H2ygKORzb7OKHfCItu//Rl22AQK3v3SgzzkHZ
# 53xu+bjJKyDaeDclxQR6COuUvJomC/QqzUG9/fyImMQZHFcFElrogLzIiSei6vcV
# UY7azSrz5fs8ySz9spMa2vyM2Tc6t1R+HekTNMUfxQPaD1qVrCmFszlMz58CWcVS
# Kd28O1SByMdZS+2LpzDPTLWVnyZVp2tnhF24jf+5oV+NLk82eOsbX8zMctyL36eb
# +8n+1flHpAnDRDpch5aHPlRovDSvMIIE7DCCA9SgAwIBAgITMwAAAMps1TISNcTh
# VQABAAAAyjANBgkqhkiG9w0BAQUFADB5MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSMwIQYDVQQDExpNaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QTAeFw0xNDA0MjIxNzM5MDBaFw0xNTA3MjIxNzM5MDBaMIGDMQswCQYDVQQGEwJV
# UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
# ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMQ0wCwYDVQQLEwRNT1BSMR4wHAYDVQQD
# ExVNaWNyb3NvZnQgQ29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAw
# ggEKAoIBAQCWcV3tBkb6hMudW7dGx7DhtBE5A62xFXNgnOuntm4aPD//ZeM08aal
# IV5WmWxY5JKhClzC09xSLwxlmiBhQFMxnGyPIX26+f4TUFJglTpbuVildGFBqZTg
# rSZOTKGXcEknXnxnyk8ecYRGvB1LtuIPxcYnyQfmegqlFwAZTHBFOC2BtFCqxWfR
# +nm8xcyhcpv0JTSY+FTfEjk4Ei+ka6Wafsdi0dzP7T00+LnfNTC67HkyqeGprFVN
# TH9MVsMTC3bxB/nMR6z7iNVSpR4o+j0tz8+EmIZxZRHPhckJRIbhb+ex/KxARKWp
# iyM/gkmd1ZZZUBNZGHP/QwytK9R/MEBnAgMBAAGjggFgMIIBXDATBgNVHSUEDDAK
# BggrBgEFBQcDAzAdBgNVHQ4EFgQUH17iXVCNVoa+SjzPBOinh7XLv4MwUQYDVR0R
# BEowSKRGMEQxDTALBgNVBAsTBE1PUFIxMzAxBgNVBAUTKjMxNTk1K2I0MjE4ZjEz
# LTZmY2EtNDkwZi05YzQ3LTNmYzU1N2RmYzQ0MDAfBgNVHSMEGDAWgBTLEejK0rQW
# WAHJNy4zFha5TJoKHzBWBgNVHR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1pY3Jv
# c29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNDb2RTaWdQQ0FfMDgtMzEtMjAx
# MC5jcmwwWgYIKwYBBQUHAQEETjBMMEoGCCsGAQUFBzAChj5odHRwOi8vd3d3Lm1p
# Y3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY0NvZFNpZ1BDQV8wOC0zMS0yMDEwLmNy
# dDANBgkqhkiG9w0BAQUFAAOCAQEAd1zr15E9zb17g9mFqbBDnXN8F8kP7Tbbx7Us
# G177VAU6g3FAgQmit3EmXtZ9tmw7yapfXQMYKh0nfgfpxWUftc8Nt1THKDhaiOd7
# wRm2VjK64szLk9uvbg9dRPXUsO8b1U7Brw7vIJvy4f4nXejF/2H2GdIoCiKd381w
# gp4YctgjzHosQ+7/6sDg5h2qnpczAFJvB7jTiGzepAY1p8JThmURdwmPNVm52Iao
# AP74MX0s9IwFncDB1XdybOlNWSaD8cKyiFeTNQB8UCu8Wfz+HCk4gtPeUpdFKRhO
# lludul8bo/EnUOoHlehtNA04V9w3KDWVOjic1O1qhV0OIhFeezCCBbwwggOkoAMC
# AQICCmEzJhoAAAAAADEwDQYJKoZIhvcNAQEFBQAwXzETMBEGCgmSJomT8ixkARkW
# A2NvbTEZMBcGCgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9z
# b2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5MB4XDTEwMDgzMTIyMTkzMloX
# DTIwMDgzMTIyMjkzMloweTELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0
# b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3Jh
# dGlvbjEjMCEGA1UEAxMaTWljcm9zb2Z0IENvZGUgU2lnbmluZyBQQ0EwggEiMA0G
# CSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCycllcGTBkvx2aYCAgQpl2U2w+G9Zv
# zMvx6mv+lxYQ4N86dIMaty+gMuz/3sJCTiPVcgDbNVcKicquIEn08GisTUuNpb15
# S3GbRwfa/SXfnXWIz6pzRH/XgdvzvfI2pMlcRdyvrT3gKGiXGqelcnNW8ReU5P01
# lHKg1nZfHndFg4U4FtBzWwW6Z1KNpbJpL9oZC/6SdCnidi9U3RQwWfjSjWL9y8lf
# RjFQuScT5EAwz3IpECgixzdOPaAyPZDNoTgGhVxOVoIoKgUyt0vXT2Pn0i1i8UU9
# 56wIAPZGoZ7RW4wmU+h6qkryRs83PDietHdcpReejcsRj1Y8wawJXwPTAgMBAAGj
# ggFeMIIBWjAPBgNVHRMBAf8EBTADAQH/MB0GA1UdDgQWBBTLEejK0rQWWAHJNy4z
# Fha5TJoKHzALBgNVHQ8EBAMCAYYwEgYJKwYBBAGCNxUBBAUCAwEAATAjBgkrBgEE
# AYI3FQIEFgQU/dExTtMmipXhmGA7qDFvpjy82C0wGQYJKwYBBAGCNxQCBAweCgBT
# AHUAYgBDAEEwHwYDVR0jBBgwFoAUDqyCYEBWJ5flJRP8KuEKU5VZ5KQwUAYDVR0f
# BEkwRzBFoEOgQYY/aHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJv
# ZHVjdHMvbWljcm9zb2Z0cm9vdGNlcnQuY3JsMFQGCCsGAQUFBwEBBEgwRjBEBggr
# BgEFBQcwAoY4aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNy
# b3NvZnRSb290Q2VydC5jcnQwDQYJKoZIhvcNAQEFBQADggIBAFk5Pn8mRq/rb0Cx
# MrVq6w4vbqhJ9+tfde1MOy3XQ60L/svpLTGjI8x8UJiAIV2sPS9MuqKoVpzjcLu4
# tPh5tUly9z7qQX/K4QwXaculnCAt+gtQxFbNLeNK0rxw56gNogOlVuC4iktX8pVC
# nPHz7+7jhh80PLhWmvBTI4UqpIIck+KUBx3y4k74jKHK6BOlkU7IG9KPcpUqcW2b
# Gvgc8FPWZ8wi/1wdzaKMvSeyeWNWRKJRzfnpo1hW3ZsCRUQvX/TartSCMm78pJUT
# 5Otp56miLL7IKxAOZY6Z2/Wi+hImCWU4lPF6H0q70eFW6NB4lhhcyTUWX92THUmO
# Lb6tNEQc7hAVGgBd3TVbIc6YxwnuhQ6MT20OE049fClInHLR82zKwexwo1eSV32U
# jaAbSANa98+jZwp0pTbtLS8XyOZyNxL0b7E8Z4L5UrKNMxZlHg6K3RDeZPRvzkbU
# 0xfpecQEtNP7LN8fip6sCvsTJ0Ct5PnhqX9GuwdgR2VgQE6wQuxO7bN2edgKNAlt
# HIAxH+IOVN3lofvlRxCtZJj/UBYufL8FIXrilUEnacOTj5XJjdibIa4NXJzwoq6G
# aIMMai27dmsAHZat8hZ79haDJLmIz2qoRzEvmtzjcT3XAH5iR9HOiMm4GPoOco3B
# oz2vAkBq/2mbluIQqBC0N1AI1sM9MIIGBzCCA++gAwIBAgIKYRZoNAAAAAAAHDAN
# BgkqhkiG9w0BAQUFADBfMRMwEQYKCZImiZPyLGQBGRYDY29tMRkwFwYKCZImiZPy
# LGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQDEyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZp
# Y2F0ZSBBdXRob3JpdHkwHhcNMDcwNDAzMTI1MzA5WhcNMjEwNDAzMTMwMzA5WjB3
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEwHwYDVQQDExhN
# aWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAw
# ggEKAoIBAQCfoWyx39tIkip8ay4Z4b3i48WZUSNQrc7dGE4kD+7Rp9FMrXQwIBHr
# B9VUlRVJlBtCkq6YXDAm2gBr6Hu97IkHD/cOBJjwicwfyzMkh53y9GccLPx754gd
# 6udOo6HBI1PKjfpFzwnQXq/QsEIEovmmbJNn1yjcRlOwhtDlKEYuJ6yGT1VSDOQD
# LPtqkJAwbofzWTCd+n7Wl7PoIZd++NIT8wi3U21StEWQn0gASkdmEScpZqiX5NMG
# gUqi+YSnEUcUCYKfhO1VeP4Bmh1QCIUAEDBG7bfeI0a7xC1Un68eeEExd8yb3zuD
# k6FhArUdDbH895uyAc4iS1T/+QXDwiALAgMBAAGjggGrMIIBpzAPBgNVHRMBAf8E
# BTADAQH/MB0GA1UdDgQWBBQjNPjZUkZwCu1A+3b7syuwwzWzDzALBgNVHQ8EBAMC
# AYYwEAYJKwYBBAGCNxUBBAMCAQAwgZgGA1UdIwSBkDCBjYAUDqyCYEBWJ5flJRP8
# KuEKU5VZ5KShY6RhMF8xEzARBgoJkiaJk/IsZAEZFgNjb20xGTAXBgoJkiaJk/Is
# ZAEZFgltaWNyb3NvZnQxLTArBgNVBAMTJE1pY3Jvc29mdCBSb290IENlcnRpZmlj
# YXRlIEF1dGhvcml0eYIQea0WoUqgpa1Mc1j0BxMuZTBQBgNVHR8ESTBHMEWgQ6BB
# hj9odHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9taWNy
# b3NvZnRyb290Y2VydC5jcmwwVAYIKwYBBQUHAQEESDBGMEQGCCsGAQUFBzAChjho
# dHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY3Jvc29mdFJvb3RD
# ZXJ0LmNydDATBgNVHSUEDDAKBggrBgEFBQcDCDANBgkqhkiG9w0BAQUFAAOCAgEA
# EJeKw1wDRDbd6bStd9vOeVFNAbEudHFbbQwTq86+e4+4LtQSooxtYrhXAstOIBNQ
# md16QOJXu69YmhzhHQGGrLt48ovQ7DsB7uK+jwoFyI1I4vBTFd1Pq5Lk541q1YDB
# 5pTyBi+FA+mRKiQicPv2/OR4mS4N9wficLwYTp2OawpylbihOZxnLcVRDupiXD8W
# mIsgP+IHGjL5zDFKdjE9K3ILyOpwPf+FChPfwgphjvDXuBfrTot/xTUrXqO/67x9
# C0J71FNyIe4wyrt4ZVxbARcKFA7S2hSY9Ty5ZlizLS/n+YWGzFFW6J1wlGysOUzU
# 9nm/qhh6YinvopspNAZ3GmLJPR5tH4LwC8csu89Ds+X57H2146SodDW4TsVxIxIm
# dgs8UoxxWkZDFLyzs7BNZ8ifQv+AeSGAnhUwZuhCEl4ayJ4iIdBD6Svpu/RIzCzU
# 2DKATCYqSCRfWupW76bemZ3KOm+9gSd0BhHudiG/m4LBJ1S2sWo9iaF2YbRuoROm
# v6pH8BJv/YoybLL+31HIjCPJZr2dHYcSZAI9La9Zj7jkIeW1sMpjtHhUBdRBLlCs
# lLCleKuzoJZ1GtmShxN1Ii8yqAhuoFuMJb+g74TKIdbrHk/Jmu5J4PcBZW+JC33I
# acjmbuqnl84xKf8OxVtc2E0bodj6L54/LlUWa8kTo/0xggSvMIIEqwIBATCBkDB5
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSMwIQYDVQQDExpN
# aWNyb3NvZnQgQ29kZSBTaWduaW5nIFBDQQITMwAAAMps1TISNcThVQABAAAAyjAJ
# BgUrDgMCGgUAoIHIMBkGCSqGSIb3DQEJAzEMBgorBgEEAYI3AgEEMBwGCisGAQQB
# gjcCAQsxDjAMBgorBgEEAYI3AgEVMCMGCSqGSIb3DQEJBDEWBBRkfj/XgPoyFhlW
# Tveucc9gh1F7QjBoBgorBgEEAYI3AgEMMVowWKA2gDQATQBpAGMAcgBvAHMAbwBm
# AHQAIABIAFAAQwAgAFAAYQBjAGsAIAAyADAAMQAyACAAUgAyoR6AHGh0dHA6Ly93
# d3cubWljcm9zb2Z0LmNvbS9IUEMwDQYJKoZIhvcNAQEBBQAEggEATgx6b757t0gB
# EqN5uifjXJBQBq/5ULIWRBkhRgRKW55qHfnthffEKb+M8pGDUp/hezgD8Kx5kNk0
# /IZIVYJ98J0+UUAiB124SyrDAgEkh27/zYOPpUWcJMkCK6xc9m74m1hjpluyoghB
# v3kpuQwqqfXMJvLJEKhxvS8n/G89+wLvF2eiwCLyqsPEQp8qmAUDAvPUV8aUNUUo
# S2rDgpx7Pln2iBMZ2uhg0MV81euoYyUBaEnq/wQjs73ChfptEylMOnozQm69Zqfv
# Sjpz5xsVRDsohObsheZwOpHxARXXaUZ+bk58qqmZWnhtbeDOiQjph6UQKCdF5EbJ
# GunZp1Y5yaGCAigwggIkBgkqhkiG9w0BCQYxggIVMIICEQIBATCBjjB3MQswCQYD
# VQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEe
# MBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEwHwYDVQQDExhNaWNyb3Nv
# ZnQgVGltZS1TdGFtcCBQQ0ECEzMAAABbQzxyQcVswO4AAAAAAFswCQYFKw4DAhoF
# AKBdMBgGCSqGSIb3DQEJAzELBgkqhkiG9w0BBwEwHAYJKoZIhvcNAQkFMQ8XDTE1
# MDIxMjE4MTQ1NVowIwYJKoZIhvcNAQkEMRYEFDtRTHzRO/dCorbOf8hP0ckDFZnt
# MA0GCSqGSIb3DQEBBQUABIIBACRyp7LIzMRpe6u7+o1/amzc+DOcXkJJ0g/x4JRk
# cgUwsolWodSkbTeopvssCSYjzuTWNf66mwdX10Z1NTccmtXS48na3lY6ziVRQoVz
# wiAY7JJRdNLZiPZJBxnykRcRT3gkAZhmXPO0XRwK4D7P7lIYX0iJEIDWJvNbh0Ms
# WsPdbygsuB9zPicknOZXbjjp5ElrkKWlkrceNf3cF3PQW1PBg3ltIOQqilLjrHxU
# 0QPIxC2UeLIim4cHrxnUqpArnTe3VL0Q/P2Q5m98Gc6DvhAyPwi2aTQwr3iys+Rc
# +HttRVQCISIUsnF+QaWIEP26eZw3w5TObYU33cln5qCco2k=
# SIG # End signature block
