<#
.Synopsis
   Deploys a Microsoft HPC Pack cluster in Microsoft Azure
.DESCRIPTION
   This script creates a Microsoft HPC Pack cluster in Microsoft Azure
   Author :  Microsoft HPC Pack team
   Version:  4.3.1
.EXAMPLE
   .\New-HPCIaasCluster.ps1 -ConfigFile ".\HPCIaaSCluster.xml" -AdminUserName "johnlee" -AdminPassword "!!123AAab"

.INPUTS
   None
.OUTPUTS
   None
#>
[CmdletBinding(SupportsShouldProcess = $true, ConfirmImpact = "High")]
Param
(
    # The file path of the configuration file for deploying the HPC Pack cluster in Microsoft Azure, please refer to the the file "Manual.rtf" which can be found in the same folder of this script.
    [Parameter(Mandatory=$true)]
    [String] $ConfigFile,

    # The administrator user name. It is the local administrator user name for all VMs as well as the domain administrator user name if the domain forest is created by the script. If the domain forest already exists, specify the domain user name to install HPC Pack.
    [Parameter(Mandatory=$true)]
    [String] $AdminUserName,

    # The password for both the local administrator and domain user. If not specified in the command line, the script will prompt the user to input the password during execution.
    [Parameter(Mandatory=$false)]
    [String] $AdminPassword = "",

    # The HPC Pack image name used to deploy the HPC cluster nodes. It must be a standard HPC Image from the Microsoft Azure Marketplace. If not specified (recommended for most users), the script chooses the latest published HPC Image. Note: Specifying an invalid HPC Image will result in deployment failure. 
    [Parameter(Mandatory=$false)]
    [String] $HPCImageName = "",

    # The file path of the deployment log file. If not specified, an auto-generated log file will be placed under the user's temp directory.
    [Parameter(Mandatory=$false)]
    [String] $LogFile = "",

    # If specified, suppresses all confirmation prompts.
    [Parameter(Mandatory=$false)]
    [Switch] $Force,

    # If specified, does not remove the VMs that are not successfully deployed. The user has to remove these unsuccessful VMs manually before rerunning the script to continue the deployment, or the deployment will fail. 
    [Parameter(Mandatory=$false)]
    [Switch] $NoCleanOnFailure,

    # If specified, does not install certificates during the deployment. For every cloud service with VMs deployed by this script, a self-signed certificate is auto-generated by Azure, and all the VMs in the cloud service use this certificate as the default Windows Remote Management (WinRM) certificate. To deploy HPC features in these Azure VMs, the script by default temporarily installs these certificates in the "Local Computer\Trusted Root Certification Authorities" store of the client computer to suppress the "not trusted CA" security error during the script execution, and the certificates are removed at the end of script execution. If "PSSessionSkipCACheck" is specified, the certificates are not installed in the client computer, and the security warning is suppressed. This parameter is not recommended for production deployments.
    [Parameter(Mandatory=$false)]
    [Switch] $PSSessionSkipCACheck
)

##################################################################################################
## Preparation
##################################################################################################

# Check the Windows Azure Powershell version
$azureModule = Get-Module -ListAvailable -Name Azure 
if ($azureModule -eq $null)
{
    throw "Azure Powershell not found. Install the latest version from http://azure.microsoft.com/downloads/#cmd-line-tools"
}
if ($azureModule.Version -lt "0.8.7")
{
    $ver = $azureModule.Version
    throw "The Azure Powershell Version $ver is too old. Install the latest version from http://azure.microsoft.com/downloads/#cmd-line-tools"
}

# The script has been tested on Powershell 3.0
Set-StrictMode -Version 3

Import-Module -Name Azure

# Import the libraries
. "$PSScriptRoot\HPCIaaSClusterUtils.ps1"
. "$PSScriptRoot\IaaSAzureCmdLib.ps1"
. "$PSScriptRoot\HPCIaaSClusterConfigLib.ps1"
. "$PSScriptRoot\HPCIaaSClusterLib.ps1"
. "$PSScriptRoot\OutputFramework.ps1"

if(-not (Test-Administrator))
{
    throw "You must run this script with administrator privileges"
}

$azureSubscriptions = GetAzureSubscription
if($azureSubscriptions -eq $null)
{
    throw "You must configure the Azure subscription before running this script. Use the Get-AzurePublishSettingsFile and Import-AzurePublishSettingsFile cmdlet or use Add-AzureAccount cmdlet to configure the subscription, and then try again. For more details about how to connection to your subscription, see http://go.microsoft.com/fwlink/?LinkId=511412."
}

# Initialize Output Framework
Initialize-Progress

##################################################################################################
## Syntax Check
##################################################################################################
if([String]::IsNullOrEmpty($LogFile))
{
    $StartTime = Get-Date -Format "yyyy_MM_dd-hh_mm_ss"
    $LogFile = "$env:temp\HPCIaaSDeploymentLog-$StartTime.txt"
    Write-Verbose "The logs will be written to $LogFile" -Verbose
}

### Step 1. Config File Semantics Check
ConfigFileSemanticsCheck -ConfigFile $ConfigFile

### Step 2. Validate the Complexity of the user name and password
ValidateUserName -UserName $AdminUserName
if([String]::IsNullOrEmpty($AdminPassword))
{
    while($true)
    {
        $firstSecPsw = Read-Host -Prompt "Please input the password for User $AdminUserName" -AsSecureString
        $firstPlainPsw = ConvertSecureStrToPlain -SecurePassword $firstSecPsw
        if([String]::IsNullOrEmpty($firstPlainPsw))
        {
            continue
        }

        $secSecPsw = Read-Host -Prompt "Please input the password for User $AdminUserName again" -AsSecureString
        $firstPlainPsw = ConvertSecureStrToPlain -SecurePassword $firstSecPsw
        $secPlainPsw = ConvertSecureStrToPlain -SecurePassword $secSecPsw
        if($firstPlainPsw -eq $secPlainPsw)
        {
            $AdminPassword = $firstPlainPsw
            break
        }
        else
        {
            TraceError "The passwords you input do not match, please input again!"
        }
    }
}

ValidatePasswordComplexity -Password $AdminPassword

### Step 3. Syntax check passed, load the configuration
$config = ([xml] (Get-Content $ConfigFile)).IaaSClusterConfig
$configAzureBurst = $false
if($config["AzureBurst"] -ne $null)
{
    $certificatePasswords = ValidateCertificates $config["AzureBurst"]
    $configAzureBurst = $true
}

$configSubscriptionName = $config.Subscription.SubscriptionName.Trim()
$subs = @($azureSubscriptions | ?{$_.SubscriptionName -eq $configSubscriptionName})
if($subs.Count -eq 0)
{
    throw "The subscription $configSubscriptionName doesn't exist. It may be because the subscription name is incorrect or you haven't imported the correct Azure publishing profile. Specify a correct subscription name, or use the Get-AzurePublishSettingsFile and Import-AzurePublishSettingsFile cmdlet to download and import the Azure publishing profile, and then try again. For more details about the Azure publishing profile, see http://go.microsoft.com/fwlink/?LinkId=511412."
}
elseif($subs.Count -gt 1)
{
    throw "There are two or more subscriptions with same name $configSubscriptionName, the subscription name must be unique."
}

$subscriptionId = $subs[0].SubscriptionId        

Select-AzureSubscription -SubscriptionName $configSubscriptionName | Out-Null
$DefaultWinServerOSVersion = "WindowsServer2012R2"
$configDefaultStorageAccount = $config.Subscription.StorageAccount.Trim()
$storageAccountArray = @($configDefaultStorageAccount)
$configLanguage = ""
if($config["Language"] -ne $null)
{
    $configLanguage = $config.Language.Trim()
}
$configLocation = $config.Location.Trim()
$configDefaultAGName = ""
$agArray = @()
$serviceAGTable = @{}
if($config["AffinityGroup"] -ne $null)
{
    $configDefaultAGName = $config.AffinityGroup.Trim()
    $agArray += $configDefaultAGName
}

$configVNetName = $config.VNet.VNetName.Trim()
$configSubnetName = $config.VNet.SubNetName.Trim()

$configDomainFQDN = $config.Domain.DomainFQDN.Trim()
$domainNetBios = $configDomainFQDN.Split(".")[0].ToUpper()
$configDCOption = $config.Domain.DCOption.Trim()

if($configDCOption -eq "NewDC")
{
    $configDCName = $config.Domain.DomainController.VMName.Trim()
    $configDCSvcName = $config.Domain.DomainController.ServiceName.Trim()
    $configDCSize = $config.Domain.DomainController.VMSize.Trim()
    $configDCAGName = $configDefaultAGName
    if($config.Domain.DomainController["AffinityGroup"] -ne $null)
    {
        $configDCAGName = $config.Domain.DomainController.AffinityGroup.Trim()
        if($agArray -notcontains $configDCAGName)
        {
            $agArray += $configDCAGName
        }
    }

    $serviceAGTable[$configDCSvcName] = $configDCAGName
}
elseif($configDCOption -eq "ExistingDC")
{
    $dcExists = $true
}

$configDBOption = $config.Database.DBOption.Trim()
if($configDBOption -eq "NewRemoteDB")
{
    $configDBVersion = $config.Database.DBVersion.Trim()
    $configRemoteDBName = $config.Database.DBServer.VMName.Trim()
    $configRemoteDBSvcName = $config.Database.DBServer.ServiceName.Trim()
    $configRemoteDBSize = $config.Database.DBServer.VMSize.Trim()
    $configRemoteDBAGName = $configDefaultAGName
    if($config.Database.DBServer["AffinityGroup"] -ne $null)
    {
        $configRemoteDBAGName = $config.Database.DBServer.AffinityGroup.Trim()
        if($agArray -notcontains $configRemoteDBAGName)
        {
            $agArray += $configRemoteDBAGName
        }
    }

    $serviceAGTable[$configRemoteDBSvcName] = $configRemoteDBAGName

    $configRemoteDBDataDiskSize = 200
    if($config.Database.DBServer["DataDiskSizeInGB"] -ne $null)
    {
        $configRemoteDBDataDiskSize = [int]$config.Database.DBServer.DataDiskSizeInGB.Trim()
    }
}
elseif($configDBOption -eq "ExistingDB")
{
    $dbExists = $true
    $configMgmtDBName = "HpcManagement"
    $configSchdDBName = "HpcScheduler"
    $configMonDBName = "HpcMonitoring"
    $configRptDBName = "HpcReporting"
    $configDiagDBName = "HpcDiagnostics"
    $configDBServerInstance = $config.Database.DBConnection.ServerInstance.Trim()
    if($config.Database.DBConnection["ManagementDB"] -ne $null)
    {
        $configMgmtDBName = $config.Database.DBConnection.ManagementDB.Trim()
    }

    if($config.Database.DBConnection["SchedulerDB"] -ne $null)
    {
        $configSchdDBName = $config.Database.DBConnection.SchedulerDB.Trim()
    }

    if($config.Database.DBConnection["MonitoringDB"] -ne $null)
    {
        $configMonDBName = $config.Database.DBConnection.MonitoringDB.Trim()
    }

    if($config.Database.DBConnection["ReportingDB"] -ne $null)
    {
        $configRptDBName = $config.Database.DBConnection.ReportingDB.Trim()
    }

    if($config.Database.DBConnection["DiagnosticsDB"] -ne $null)
    {
        $configDiagDBName = $config.Database.DBConnection.DiagnosticsDB.Trim()
    }
}

$configHNName = $config.HeadNode.VMName.Trim()
$configHNSvcName = $config.HeadNode.ServiceName.Trim()
$configHNSize = $config.HeadNode.VMSize.Trim()
$configHNAGName = $configDefaultAGName
if($config.HeadNode["AffinityGroup"] -ne $null)
{
    $configHNAGName = $config.HeadNode.AffinityGroup.Trim()
    if($agArray -notcontains $configHNAGName)
    {
        $agArray += $configHNAGName
    }
}

$configHNEnableRESTAPI = $false
if(($config.HeadNode["EnableRESTAPI"] -ne $null) -and ($config.HeadNode.EnableRESTAPI.Trim() -ne "false"))
{
    $configHNEnableRESTAPI = $true
}

$configHNEnableWebPortal = $false
if(($config.HeadNode["EnableWebPortal"] -ne $null) -and ($config.HeadNode.EnableWebPortal.Trim() -ne "false"))
{
    $configHNEnableWebPortal = $true
}

$configHNPreConfigScript = ""
if($config.HeadNode["PreConfigScript"] -ne $null)
{
    $configHNPreConfigScript = $config.HeadNode.PreConfigScript.Trim()
}

$configHNPostConfigScript = ""
if($config.HeadNode["PostConfigScript"] -ne $null)
{
    $configHNPostConfigScript = $config.HeadNode.PostConfigScript.Trim()
}

$serviceAGTable[$configHNSvcName] = $configHNAGName

if($configDCOption -eq "HeadNodeAsDC")
{
    $configDCName = $configHNName
    $configDCSvcName = $configHNSvcName
    $configDCSize = $configHNSize
    $configDCAGName = $configHNAGName
}

$NodeNamePattern = "AZUREVM-%0000%"
$configCNCount = 0
if ($config["ComputeNodes"] -ne $null)
{
    $configCNNamePattern = $config.ComputeNodes.VMNamePattern.Trim()
    $NodeNamePattern = $configCNNamePattern
    $configCNSize = $config.ComputeNodes.VMSize.Trim()
    $configCNCount = [int]$config.ComputeNodes.NodeCount.Trim()
    
    if($config.ComputeNodes["ServiceName"] -ne $null)
    {
        $configCNWithServicePattern = $false
        $configCNSvcName = $config.ComputeNodes.ServiceName.Trim()
    }
    else
    {
        $configCNWithServicePattern = $true
        $configCNSvcNamePattern = $config.ComputeNodes.ServiceNamePattern.Trim()
        $configCNMaxNodeCountPerSvc = [int]$config.ComputeNodes.MaxNodeCountPerService.Trim()
        $configCNStorageNamePattern = $config.ComputeNodes.StorageAccountNamePattern.Trim()
    }

    if($configCNCount -ne 0)
    {
        $configCNAGName = $configDefaultAGName
        if($config.ComputeNodes["AffinityGroup"] -ne $null)
        {
            $configCNAGName = $config.ComputeNodes.AffinityGroup.Trim()
            if($agArray -notcontains $configCNAGName)
            {
                $agArray += $configCNAGName
            }
        }

        $configCNServiceStorageTable = @{}
        if($configCNWithServicePattern)
        {
            $configCNSetCount = [Math]::Ceiling($configCNCount/$configCNMaxNodeCountPerSvc)
            $configCNServiceNames = @(GetNamesByPattern -NamePattern $configCNSvcNamePattern -Count $configCNSetCount)
            foreach($svc in $configCNServiceNames)
            {
                $serviceAGTable[$svc] = $configCNAGName 
            }
            $configCNStorageAccounts = @(GetNamesByPattern -NamePattern $configCNStorageNamePattern -Count $configCNSetCount)
            $storageAccountArray += $configCNStorageAccounts
            for($i=0; $i -lt $configCNSetCount; $i++)
            {
                $configCNServiceStorageTable[$configCNServiceNames[$i]] = $configCNStorageAccounts[$i]
            }
        }
        else
        {
            $serviceAGTable[$configCNSvcName] = $configCNAGName
            $configCNServiceStorageTable[$configCNSvcName] = $configDefaultStorageAccount
        }
        $configCNImage = $null
        $configCNOSVersion = $null
        $configCNImageHPCInstalled = $false
        if($config.ComputeNodes["ImageName"] -ne $null)
        {
            $configCNImage = $config.ComputeNodes["ImageName"].InnerText.Trim()
            if($config.ComputeNodes["ImageName"].HasAttribute("HPCPackInstalled"))
            {
                if($config.ComputeNodes["ImageName"].GetAttribute("HPCPackInstalled").Trim() -eq "true")
                {
                    $configCNImageHPCInstalled = $true
                }
            }
        }
        elseif($config.ComputeNodes["OSVersion"] -ne $null)
        {
            $configCNOSVersion = $config.ComputeNodes.OSVersion.Trim()
        }
        else
        {
            $configCNOSVersion = $DefaultWinServerOSVersion
        }
    }
}

$configBNCount = 0
if ($config["BrokerNodes"] -ne $null)
{
    $configBNNamePattern = $config.BrokerNodes.VMNamePattern.Trim()
    $configBNSvcName = $config.BrokerNodes.ServiceName.Trim()
    $configBNSize = $config.BrokerNodes.VMSize.Trim()
    $configBNCount = [int]$config.BrokerNodes.NodeCount.Trim()
    if($configBNCount -ne 0)
    {
        $configBNAGName = $configDefaultAGName
        if($config.BrokerNodes["AffinityGroup"] -ne $null)
        {
            $configBNAGName = $config.BrokerNodes.AffinityGroup.Trim()
            if($agArray -notcontains $configBNAGName)
            {
                $agArray += $configBNAGName
            }
        }

        $configBNServiceStorageTable = @{$configBNSvcName=$configDefaultStorageAccount}
        $serviceAGTable[$configBNSvcName] = $configBNAGName
        $configBNImage = $null
        $configBNOSVersion = $null
        $configBNImageHPCInstalled = $false
        if($config.BrokerNodes["ImageName"] -ne $null)
        {
            $configBNImage = $config.BrokerNodes["ImageName"].InnerText.Trim()
            if($config.BrokerNodes["ImageName"].HasAttribute("HPCPackInstalled"))
            {
                if($config.BrokerNodes["ImageName"].GetAttribute("HPCPackInstalled").Trim() -eq "true")
                {
                    $configBNImageHPCInstalled = $true
                }
            }
        }
        elseif($config.BrokerNodes["OSVersion"] -ne $null)
        {
            $configBNOSVersion = $config.BrokerNodes.OSVersion.Trim()
        }
        else
        {
            $configBNOSVersion = $DefaultWinServerOSVersion
        }
    }
}

$deploymentLabel = "HpcDeploy-$configVNetName-$configHNName"
# Set process enviroment variables
[Environment]::SetEnvironmentVariable("HPCCurSubscriptionName", $configSubscriptionName, [System.EnvironmentVariableTarget]::Process)
[Environment]::SetEnvironmentVariable("HPCPSScriptRoot", $PSScriptRoot, [System.EnvironmentVariableTarget]::Process)
if($PSSessionSkipCACheck.IsPresent)
{
    [Environment]::SetEnvironmentVariable("HPCTempWinRMCertListFile", "", [System.EnvironmentVariableTarget]::Process)
}
else
{
    $uniqueId = [Guid]::NewGuid()
    $WinRMCertListFile = "$env:temp\HPCIaaSDeployment_WinRMCertThumbprints_$uniqueId.list"
    [Environment]::SetEnvironmentVariable("HPCTempWinRMCertListFile", $WinRMCertListFile, [System.EnvironmentVariableTarget]::Process)
}

$jobArray = @()
$JobInitScript =  {
    Set-StrictMode -Version 3
    Import-Module -Name Azure
    . "$env:HPCPSScriptRoot\HPCIaaSClusterUtils.ps1"
    . "$env:HPCPSScriptRoot\IaaSAzureCmdLib.ps1"
    . "$env:HPCPSScriptRoot\HPCIaaSClusterConfigLib.ps1"
    . "$env:HPCPSScriptRoot\HPCIaaSClusterLib.ps1"
    Select-AzureSubscription -SubscriptionName $env:HPCCurSubscriptionName | Out-Null
}

##################################################################################################
## Current Status Check
##################################################################################################
$CheckStepSequnce = @()
$CheckStepTable = @{}
$CheckStepSequnce += $CheckStepTable["CheckAzureCapabilities"]     = "Getting Azure VM sizes and images"
$CheckStepSequnce += $CheckStepTable["CheckStorageAccount"]        = "Checking storage account"
$CheckStepSequnce += $CheckStepTable["CheckLocationAffinityGroup"] = "Checking region and affinity groups"
$CheckStepSequnce += $CheckStepTable["CheckVNet"]                  = "Checking virtual network settings"
$CheckStepSequnce += $CheckStepTable["CheckService"]               = "Checking cloud services"
$CheckStepSequnce += $CheckStepTable["CheckDomain"]                = "Checking domain settings"
$CheckStepSequnce += $CheckStepTable["CheckDatabase"]              = "Checking database settings"
$CheckStepSequnce += $CheckStepTable["CheckHeadNode"]              = "Checking head node settings"

if($configCNCount -ne 0)
{
    $CheckStepSequnce += $CheckStepTable["CheckComputeNodes"]      = "Checking compute nodes settings"
}
if($configBNCount -ne 0)
{
    $CheckStepSequnce += $CheckStepTable["CheckBrokerNodes"]       = "Checking broker nodes settings"
}

Start-Progress "Validating the Configuration File" $CheckStepSequnce.Count
$CheckStepSequnce | %{Report-Progress $_ Queued}

try
{
    ### Step 1. Check the subscription setting
    # Check the subscription and storage account
    $jobArray += Start-Job -Name "CheckAzureCapabilities" -InitializationScript $JobInitScript -ScriptBlock {
        TraceInfo "Getting the available VM Sizes"
        $vmsizes = InvokeAzureCmdWithRetry -Command "Get-AzureRoleSize" | ?{$_.SupportedByVirtualMachines} | %{$_.InstanceSize}
        TraceInfo "Getting the available VM Images"
        $vmimages = InvokeAzureCmdWithRetry -Command "Get-AzureVMImage"
        if(-not [String]::IsNullOrEmpty($using:HPCImageName))
        {
            TraceInfo "Checking the HPC Image $using:HPCImageName"
            CheckAzureVMImage -ImageName $using:HPCImageName -Location $using:configLocation -FromImageList $vmimages
        }

        return @{"VMSizes"=$vmsizes; "VMImages"=$vmimages}
    }
    
    $getAzureCapJob = $jobArray[-1]
    Report-Progress -subProgress $CheckStepTable["CheckAzureCapabilities"] -job $getAzureCapJob

    ### Step 2. Check the storage account
    $jobArray += Start-Job -Name "CheckStorageAccount" -InitializationScript $JobInitScript -ScriptBlock {
        CheckStorageAccount -StorageAccountName $using:storageAccountArray -Location $using:configLocation
    }

    $checkStorageJob = $jobArray[-1]
    Report-Progress -subProgress $CheckStepTable["CheckStorageAccount"] -job $checkStorageJob

    ### Step 3. Check Location and affinity groups
    $jobArray += Start-Job -Name "CheckLocationAffinityGroup" -InitializationScript $JobInitScript -ScriptBlock  {
        CheckLocationAndAffinityGroup -AGNames $using:agArray -Location $using:configLocation
    }

    $checkLocationAGJob = $jobArray[-1]
    Report-Progress -subProgress $CheckStepTable["CheckLocationAffinityGroup"] -job $checkLocationAGJob

    ### Step 4. Check VNet
    $jobArray += Start-Job -Name "CheckVNet" -InitializationScript $JobInitScript -ScriptBlock  {
        CheckVNet -VNetName $using:configVNetName -SubnetName $using:configSubnetName -Location $using:configLocation
    }

    $checkVNetJob = $jobArray[-1]
    Report-Progress -subProgress $CheckStepTable["CheckVNet"] -job $checkVNetJob

    ### Step 5. Check Services
    $jobArray += Start-Job -Name "CheckService" -InitializationScript $JobInitScript -ScriptBlock {
        $serviceAGTable = $using:serviceAGTable 
        $svcStateTable = CheckService -ServiceName ($serviceAGTable.Keys | Sort-Object) -VNetName $using:configVNetName -Location $using:configLocation `
                         -ServiceAffinityGroupTable $serviceAGTable -DeploymentLabel $using:deploymentLabel
        return $svcStateTable
    }

    $checkServiceJob = $jobArray[-1]
    Report-Progress -subProgress $CheckStepTable["CheckService"] -job $checkServiceJob

    # Wait for above jobs completion
    Wait-Progress -Jobs $jobArray
    CheckJobStatus -Jobs $jobArray

    $availSizes = $getAzureCapJob.ChildJobs[0].Output[-1]["VMSizes"]
    $availVMImages = $getAzureCapJob.ChildJobs[0].Output[-1]["VMImages"]
    $storageAccountStateTable = $checkStorageJob.ChildJobs[0].Output[-1]
    $agStateTable = $checkLocationAGJob.ChildJobs[0].Output[-1]
    $serviceStateTable = $checkServiceJob.ChildJobs[0].Output[-1]
    $vnetExists = $checkVNetJob.ChildJobs[0].Output[-1]

    ### Step 6. Check Domain Setttings
    $domainNetBios = $configDomainFQDN.Split(".")[0].ToUpper()
    $checkDCJob = $null
    $dcExists = $false
    if($configDCOption -eq "ExistingDC")
    {
        if(-not $vnetExists)
        {
            Report-Progress -subProgress $CheckStepTable["CheckDomain"]  -status Failed -message "Wrong Domain Setting"
            throw "Domain controller and HPC cluster must be in the same virtual network. Please check that the virtual network $configVNetName exists. If it does not exist, please specify 'DCOption' as 'NewDC'."
        }

        $dcExists = $true
        Report-Progress -subProgress $CheckStepTable["CheckDomain"] -status Passed -message "Correct Domain Setting"
    }
    else
    {
        $dcSvcState = $serviceStateTable[$configDCSvcName]
        $jobArray += Start-Job -Name "CheckDomain" -InitializationScript $JobInitScript -ScriptBlock {
            ValidateVMSize -VMName $using:configDCName -VMSize $using:configDCSize -AvailableSizes $using:availSizes
            $exists = $false
            if($using:dcSvcState -eq "Deployed")
            {
                $exists = ValidateVMConfig -VMName $using:configDCName -ServiceName $using:configDCSvcName -SubnetName $using:configSubnetName
            }
            else
            {
                TraceInfo "The virtual machine $using:configDCName doesn't exist"
            }

            return $exists
        }

        $checkDCJob = $jobArray[-1]
        Report-Progress -subProgress $CheckStepTable["CheckDomain"] -job $checkDCJob
        # Wait for the job completion
        Wait-Progress -Jobs $checkDCJob
        CheckJobStatus -Jobs $checkDCJob
        $dcExists = $checkDCJob.ChildJobs[0].Output[-1]
    }

    ### Step 7. Check Database Settings
    $checkDBJob = $null
    if($configDBOption -eq "NewRemoteDB")
    {
        $dbSvcState = $serviceStateTable[$configRemoteDBSvcName]
        $jobArray += Start-Job -Name "CheckDatabase" -InitializationScript $JobInitScript -ScriptBlock {
            ValidateVMSize -VMName $using:configRemoteDBName -VMSize $using:configRemoteDBSize -AvailableSizes $using:availSizes
            $vmexists = $false
            if($using:dbSvcState -eq "Deployed")
            {
                $vmexists = ValidateVMConfig -VMName $using:configRemoteDBName -ServiceName $using:configRemoteDBSvcName -SubnetName $using:configSubnetName
            }
            else
            {
                TraceInfo "The virtual machine $using:configRemoteDBName doesn't exist"
            }

            if($vmexists -and (-not $using:dcExists))
            {
                throw "Remote database server $using:configRemoteDBName should not exist when the domain controller does not exist"
            }

            return $vmexists
        }

        $checkDBJob = $jobArray[-1]
        Report-Progress -subProgress $CheckStepTable["CheckDatabase"] -job $checkDBJob
    }
    else
    {
        Report-Progress -subProgress $CheckStepTable["CheckDatabase"] -status Passed -message "Correct Database Setting"
    }

    ### Step 8. Check HeadNode Settings
    $checkHNJob = $null
    if($configDCOption -ne "HeadNodeAsDC")
    {
        $hnExists = $false
        $hnSvcState = $serviceStateTable[$configHNSvcName]
        $jobArray += Start-Job -Name "CheckHeadNode" -InitializationScript $JobInitScript -ScriptBlock {
            $vmExists = $false
            ValidateVMSize -VMName $using:configHNName -VMSize $using:configHNSize -AvailableSizes $using:availSizes
            if($using:hnSvcState -eq "Deployed")
            {
                $vmExists = ValidateVMConfig -VMName $using:configHNName -ServiceName $using:configHNSvcName -SubnetName $using:configSubnetName
            }
            else
            {
                TraceInfo "The virtual machine $using:configHNName doesn't exist"
            }

            if($vmexists -and (-not $using:dcExists))
            {
                throw "The head node $using:configHNName should not exist when the domain controller does not exist"    
            }

            return $vmExists
        }

        $checkHNJob = $jobArray[-1]
        Report-Progress -subProgress $CheckStepTable["CheckHeadNode"] -job $checkHNJob
    }
    else
    {
        $hnExists = $dcExists
        Report-Progress -subProgress $CheckStepTable["CheckHeadNode"] -status Passed
    }

    ### Step 9. Check Compute Nodes Settings
    $checkCNJob = $null
    if($configCNCount -ne 0)
    {
        if($configCNWithServicePattern)
        {
            $jobArray += Start-Job -Name "CheckComputeNodes" -InitializationScript $JobInitScript -ScriptBlock {
                CheckNodeGroup -NamePattern $using:configCNNamePattern `
                               -NodeCount $using:configCNCount `
                               -VMSize $using:configCNSize `
                               -AvailableVMSizes $using:availSizes `
                               -AvailableVMImages $using:availVMImages `
                               -ServiceNamePattern $using:configCNSvcNamePattern `
                               -StorageAccountNamePattern $using:configCNStorageNamePattern `
                               -MaxNodeCountPerService $using:configCNMaxNodeCountPerSvc `
                               -ServiceStateTable $using:serviceStateTable `
                               -NodeType ComputeNode `
                               -ClusterName $using:configHNName `
                               -ImageName $using:configCNImage `
                               -Location $using:configLocation `
                               -ImageHPCInstalled:$usng:configCNImageHPCInstalled
            }
        }
        else
        {
            $jobArray += Start-Job -Name "CheckComputeNodes" -InitializationScript $JobInitScript -ScriptBlock {            
                CheckNodeGroup -NamePattern $using:configCNNamePattern `
                               -NodeCount $using:configCNCount `
                               -VMSize $using:configCNSize `
                               -AvailableVMSizes $using:availSizes `
                               -AvailableVMImages $using:availVMImages `
                               -ServiceName $using:configCNSvcName `
                               -ServiceStateTable $using:serviceStateTable `
                               -NodeType ComputeNode `
                               -ClusterName $using:configHNName `
                               -ImageName $using:configCNImage `
                               -Location $using:configLocation `
                               -ImageHPCInstalled:$usng:configCNImageHPCInstalled
            }
        }

        $checkCNJob = $jobArray[-1]
        Report-Progress -subProgress $CheckStepTable["CheckComputeNodes"] -job $checkCNJob
    }

    ### Step 10.  Check Broker Nodes Settings
    $checkBNJob = $null
    if($configBNCount -ne 0)
    {
        $jobArray += Start-Job -Name "CheckBrokerNodes" -InitializationScript $JobInitScript -ScriptBlock {
            CheckNodeGroup -NamePattern $using:configBNNamePattern `
                           -NodeCount $using:configBNCount `
                           -VMSize $using:configBNSize `
                           -AvailableVMSizes $using:availSizes `
                           -AvailableVMImages $using:availVMImages `
                           -ServiceName $using:configBNSvcName `
                           -ServiceStateTable $using:serviceStateTable `
                           -NodeType BrokerNode `
                           -ClusterName $using:configHNName `
                           -ImageName $using:configBNImage `
                           -Location $using:configLocation `
                           -ImageHPCInstalled:$usng:configBNImageHPCInstalled
        }

        $checkBNJob = $jobArray[-1]
        Report-Progress -subProgress $CheckStepTable["CheckBrokerNodes"] -job $checkBNJob
    }

    Wait-Progress
    CheckJobStatus -Jobs $jobArray

    if($checkDBJob -ne $null)
    {
        $dbExists = $checkDBJob.ChildJobs[0].Output[-1]
    }
    if($checkHNJob -ne $null)
    {
        $hnExists = $checkHNJob.ChildJobs[0].Output[-1]
    }

    $cnServiceVMTable = @{}
    if($checkCNJob -ne $null)
    {
        $cnServiceVMTable = $checkCNJob.ChildJobs[0].Output[0]
        if(-not [String]::IsNullOrEmpty($checkCNJob.ChildJobs[0].Output[1]))
        {
            # Use the CN image directly
            $configCNImage =  $checkCNJob.ChildJobs[0].Output[1]
            $configCNImageHPCInstalled = $true
        }
    }

    $bnServiceVMTable = @{}
    if($checkBNJob -ne $null)
    {
        $bnServiceVMTable = $checkBNJob.ChildJobs[0].Output[0]
        if(-not [String]::IsNullOrEmpty($checkBNJob.ChildJobs[0].Output[1]))
        {
            $configBNImage = $checkBNJob.ChildJobs[0].Output[1]
            $configBNImageHPCInstalled = $true
        }
    }
}
catch
{
    Write-Host "The HPC cluster deployment failed: " -ForegroundColor Red
    $_ | Out-String | Write-Host -ForegroundColor Red
    Uninitialize-Progress
    exit 1
}
finally
{
    $jobArray | Stop-Job
    WriteJobLogsToFile $jobArray $LogFile 
    $jobArray | Remove-Job
    $jobArray = @()
}

##################################################################################################
## Provision
##################################################################################################
$ProvisionState = "Idle"
$jobArray = @()
$ProvisionStepTable = @{}
$ProvisionStepSequence = @()
$SummaryFormat = "{0,-18}:{1}"
$VMSummaryFormat = "    {0, -18}{1, -25}{2, -15}{3}"
try
{
    $ProvisionToDoSummary = @("", "The following resources will be created for your HPC cluster in Azure under your subscription '$configSubscriptionName'($subscriptionId):")
    $VMsToProvision = @{}
    if($agStateTable.Count -gt 0 -and $agStateTable.Values -contains $false)
    {
        $ProvisionStepSequence += $ProvisionStepTable["CreateAffinityGroup"]  = "Creating affinity groups"
        $ProvisionToDoSummary += $SummaryFormat -f "Affinity group(s)", "$($agStateTable.Keys | ?{-not $agStateTable[$_]})"
    }

    if($storageAccountStateTable.Values -contains $false)
    {
        $ProvisionStepSequence += $ProvisionStepTable["CreateStorageAccount"] = "Creating storage account"
        $ProvisionToDoSummary += $SummaryFormat -f "Storage account(s)", "$($storageAccountStateTable.Keys | ?{$storageAccountStateTable[$_] -eq $false} | sort)"
    }

    if(-not $vnetExists)
    {
        $ProvisionStepSequence += $ProvisionStepTable["CreateVNet"] = "Creating virtual network"
        $ProvisionToDoSummary += $SummaryFormat -f "Virtual network", $configVNetName
        $ProvisionToDoSummary += $SummaryFormat -f "Subnet", $configSubnetName
    }

    if($serviceStateTable.Values -contains "NotCreated")
    {
        $ProvisionStepSequence += $ProvisionStepTable["CreateServices"] = "Creating cloud services"
        $ProvisionToDoSummary += $SummaryFormat -f "Cloud service(s)", "$($serviceStateTable.Keys | ?{$serviceStateTable[$_] -eq "NotCreated"} | sort)"
    }

    $ProvisionToDoSummary += $SummaryFormat -f "Virtual machine(s)", ""
    $ProvisionToDoSummary += $VMSummaryFormat -f "Name", "ServiceName", "Size", "Role"
    $ProvisionToDoSummary += $VMSummaryFormat -f "----", "-----------", "----", "----"

    if(-not $dcExists)
    {
        $ProvisionStepSequence += $ProvisionStepTable["CreateVMForDC"] = "Creating VM for domain controller"
        $ProvisionStepSequence += $ProvisionStepTable["PromoteDC"]     = "Promoting domain controller"
        if($configDCOption -eq "HeadNodeAsDC")
        {
            $ProvisionToDoSummary += $VMSummaryFormat -f $configDCName, $configDCSvcName,  $configDCSize, "Domain controller, Head node"
        }
        else
        {
            $ProvisionToDoSummary += $VMSummaryFormat -f $configDCName, $configDCSvcName,  $configDCSize, "Domain controller"
        }
    }

    if(($configDCOption -ne "HeadNodeAsDC") -and (-not $hnExists))
    {
        $ProvisionStepSequence += $ProvisionStepTable["CreateVMForHN"] = "Creating VM for head node"
        $ProvisionToDoSummary += $VMSummaryFormat -f $configHNName, $configHNSvcName, $configHNSize, "Head node"
    }

    if(($configDBOption -eq "NewRemoteDB") -and (-not $dbExists))
    {
        $ProvisionStepSequence += $ProvisionStepTable["CreateDBServer"] = "Creating remote database server"
        $ProvisionToDoSummary += $VMSummaryFormat -f $configRemoteDBName, $configRemoteDBSvcName, $configRemoteDBSize, "Database server"
    }

    if($cnServiceVMTable.Count -gt 0)
    {
        $ProvisionStepSequence += $ProvisionStepTable["CreateComputeNodes"] = "Creating compute nodes"
        foreach($svc in @($cnServiceVMTable.keys | sort))
        {
            foreach($cnname in $cnServiceVMTable[$svc])
            {
                $ProvisionToDoSummary += $VMSummaryFormat -f $cnname, $svc, $configCNSize, "Compute node"
            }
        }
    }

    if($bnServiceVMTable.Count -gt 0)
    {
        $ProvisionStepSequence += $ProvisionStepTable["CreateBrokerNodes"] = "Creating broker nodes"
        foreach($svc in @($bnServiceVMTable.keys | sort))
        {
            foreach($bnname in $bnServiceVMTable[$svc])
            {
                $ProvisionToDoSummary += $VMSummaryFormat -f $bnname, $svc, $configBNSize, "Broker node"
            }
        }
    }

    if(($configDBOption -eq "NewRemoteDB") -and (-not $dbExists))
    {
        $ProvisionStepSequence += $ProvisionStepTable["ConfigDBServer"] = "Configuring remote database server"
    }

    if(-not $hnExists)
    {
        $ProvisionStepSequence += $ProvisionStepTable["PrepareHeadNode"] = "Preparing head node"
        $ProvisionStepSequence += $ProvisionStepTable["ConfigHeadNode"] = "Configuring head node"
        if($configAzureBurst)
        {
            $ProvisionStepSequence += $ProvisionStepTable["CreatePaaSNodes"] = "Creating Azure burst nodes"            
        }
    }

    if($ProvisionStepTable.Count -eq 0)
    {
        Write-Host "The HPC cluster was already deployed."
        return
    }

    $ProvisionToDoSummary | Write-Host
    Write-Host
    if(($cnServiceVMTable.Count -gt 0) -and (-not [String]::IsNullOrEmpty($configCNImage)))
    {
        Write-Host "The VM image used to deploy HPC Pack compute nodes is $configCNImage."
    }
    if(($bnServiceVMTable.Count -gt 0) -and (-not [String]::IsNullOrEmpty($configBNImage)))
    {
        Write-Host "The VM image used to deploy HPC Pack broker nodes is $configBNImage."
    }

    if(-not $Force.IsPresent)
    {
        if(-not $PsCmdlet.ShouldProcess("Start to deploy the HPC cluster", "Continue to deploy the HPC cluster?", "Confirm"))
        {
            $ProvisionStepTable.Clear()
            return
        }
    }

    # The deployment starting time
    $StartDeployTime = Get-Date

    # Get the Image Names
    $DefaultWinServerOSImage = Get-LatestWindowsServerImage -Version $DefaultWinServerOSVersion -Language $configLanguage -FromImageList $availVMImages
    if([String]::IsNullOrEmpty($HPCImageName))
    {
        $hnImageName = Get-LatestHPCImage -Language $configLanguage -FromImageList $availVMImages
    }
    else
    {
        $hnImageName = $HPCImageName  
    }

    if($configDBOption -eq "NewRemoteDB")
    {
        $dbImageName = Get-LatestSQLServerImage -Version $configDBVersion -Language $configLanguage -FromImageList $availVMImages
    }

    if($configDCOption -eq "NewDC")
    {
        $dcImageName = $DefaultWinServerOSImage
    }
    elseif($configDCOption -eq "HeadNodeAsDC")
    {
        $dcImageName = $hnImageName
    }

    if(($cnServiceVMTable.Count -gt 0) -and [String]::IsNullOrEmpty($configCNImage))
    {
        $configCNImage = Get-LatestWindowsServerImage -Version $configCNOSVersion -Language $configLanguage -FromImageList $availVMImages
    }

    if(($bnServiceVMTable.Count -gt 0) -and [String]::IsNullOrEmpty($configBNImage))
    {
        $configBNImage = Get-LatestWindowsServerImage -Version $configBNOSVersion -Language $configLanguage -FromImageList $availVMImages
    }

    Write-Host
    Start-Progress "Deploying the HPC IaaS Cluster" $ProvisionStepSequence.Count
    $ProvisionStepSequence | %{Report-Progress $_ Queued}

    $ProvisionState = "Running"
    ### Step 1. Create Affinity Group
    $createAGJob = $null
    if($agStateTable.Count -gt 0 -and $agStateTable.Values -contains $false)
    {
        $jobArray += Start-Job -Name "CreateAffinityGroup" -InitializationScript $JobInitScript -ScriptBlock {
            $stateTable = $using:agStateTable
            foreach($agname in $stateTable.Keys)
            {
                if(-not $stateTable[$agname])
                {
                    New-AzureAffinityGroupWithRetry -Name $agname -Location $using:configLocation
                }
            }
        }

        $createAGJob = $jobArray[-1]
        Report-Progress -subProgress $ProvisionStepTable["CreateAffinityGroup"] -job $createAGJob
    }

    #### Step 2. Create storage account
    if($storageAccountStateTable.Values -contains $false)
    {
        $stgToCreate = @($storageAccountStateTable.Keys | ?{-not $storageAccountStateTable[$_]} | sort)
        $initScriptStr = $JobInitScript.ToString()
        $jobArray += Start-Job -Name "CreateStorageAccount" -InitializationScript $JobInitScript -ScriptBlock {
            $stgToCreate = $using:stgToCreate
            $location = $using:configLocation
            $initScript = [ScriptBlock]::Create($using:initScriptStr)
            $stgJobs = @()
            try
            {
                foreach($stg in $stgToCreate)
                {
                    TraceInfo "Start job to create the storage account $stg"
                    $stgJobs += Start-Job -Name $stg -InitializationScript $initScript -ScriptBlock {
                        param($stg, $location)
                        New-AzureStorageAccountWithRetry -StorageAccountName $stg -Location $location
                    } -ArgumentList @($stg, $location)
                }
            }
            catch
            {
                throw
            }

            $failedStgs = @()
            while($stgJobs.Count -gt 0)
            {
                $job = Wait-Job -Job $stgJobs -Any
                $job.ChildJobs[0].Verbose | Write-Verbose -Verbose
                if($job.State -ne "Completed")
                {
                    $failedStgs += $job.Name
                }
                $stgJobs = @($stgJobs | ?{$_ -ne $job})
            }

            if($failedStgs.Count -gt 0)
            {
                throw ("Failed to create the storage account(s): " + ($failedStgs -join ','))
            }
        }

        $createStorageJob = $jobArray[-1]
        Report-Progress -subProgress $ProvisionStepTable["CreateStorageAccount"] -job $createStorageJob
    }

    #### Step 3. Create VNet
    if(-not $vnetExists)
    {
        $jobArray += Start-Job -Name "CreateVNet" -InitializationScript $JobInitScript -ScriptBlock {
            CreateNewVNet -VNetName $using:configVNetName -SubnetName $using:configSubnetName -Location $using:configLocation
        }

        $createVNetJob = $jobArray[-1]
        Report-Progress -subProgress $ProvisionStepTable["CreateVNet"] -job $createVNetJob
    }

    ### Step 4. Create Services
    if($createAGJob -ne $null)
    {
        Wait-Progress -Jobs $createAGJob
        CheckJobStatus -Jobs $jobArray
    }
    if($serviceStateTable.Values -contains "NotCreated")
    {
        $jobArray += Start-Job -Name "CreateServices" -InitializationScript $JobInitScript -ScriptBlock {
            $stateTable = $using:serviceStateTable
            $svcAGMapTable = $using:serviceAGTable
            foreach($svc in $stateTable.Keys)
            {
                if($stateTable[$svc] -eq "NotCreated")
                {
                    if([String]::IsNullOrEmpty($svcAGMapTable[$svc]))
                    {
                        New-AzureServiceWithRetry -ServiceName $svc -Location $using:configLocation
                    }
                    else
                    {
                        New-AzureServiceWithRetry -ServiceName $svc -AffinityGroupName $svcAGMapTable[$svc]
                    }
                }
            }
        }

        $createServicesJob = $jobArray[-1]
        Report-Progress -subProgress $ProvisionStepTable["CreateServices"] -job $createServicesJob
    }

    $localAdminCred = New-Object -TypeName System.Management.Automation.PSCredential `
        -ArgumentList @($AdminUserName, (ConvertTo-SecureString -String $AdminPassword -AsPlainText -Force))

    $domainUserCred = New-Object -TypeName System.Management.Automation.PSCredential `
        -ArgumentList @("$domainNetBios\$AdminUserName", (ConvertTo-SecureString -String $AdminPassword -AsPlainText -Force))

    if($jobArray.Count -ne 0)
    {
        Wait-Progress -Jobs $jobArray
        CheckJobStatus -Jobs $jobArray
    }
    if(-not $dcExists)
    {
        ### Step 5. Create VM for DC
        $jobArray += Start-Job -Name "CreateVMForDC"  -InitializationScript $JobInitScript -ScriptBlock {
            TraceInfo "Creating the virtual machine $using:configDCName for domain controller in the cloud service $using:configDCSvcName"
            SetSubscriptionStorageAccount -SubscriptionName $using:configSubscriptionName -StorageAccountName $using:configDefaultStorageAccount
            CreateVMForDC -ServiceName $using:configDCSvcName `
                          -VMName $using:configDCName `
                          -VMSize $using:configDCSize `
                          -ImageName  $using:dcImageName `
                          -VNetName $using:configVNetName `
                          -SubNetName $using:configSubnetName `
                          -DeploymentLabel $using:deploymentLabel `
                          -AdminCredential $using:localAdminCred
            TraceInfo "The virtual machine $using:configDCName for domain controller created in the cloud service $using:configDCSvcName"
        }

        $createDCVMJob = $jobArray[-1]
        Report-Progress -subProgress $ProvisionStepTable["CreateVMForDC"] -job $createDCVMJob
        Wait-Progress -Jobs $createDCVMJob
        CheckJobStatus -Jobs $jobArray

        ### Step 6. Promote DC
        $jobArray += Start-Job -Name "PromoteDC" -InitializationScript $JobInitScript -ScriptBlock {
            PromoteDC -ServiceName $using:configDCSvcName `
                      -VMName $using:configDCName `
                      -DomainFQDN $using:configDomainFQDN `
                      -AdminCredential $using:localAdminCred
            TraceInfo "Domain controller $using:configDCName is now ready"
        } -ArgumentList @($configDCName, $configDCSvcName, $configDomainFQDN, $localAdminCred)
        
        $promoteDCJob = $jobArray[-1]
        Report-Progress -subProgress $ProvisionStepTable["PromoteDC"] -job $promoteDCJob
        Wait-Progress -Jobs $promoteDCJob
        CheckJobStatus -Jobs $jobArray
    }

    # Create HN VM and wait for vm ready
    $createHNJob = $null
    if(-not $hnExists)
    {
        if($configDCOption -ne "HeadNodeAsDC")
        {
            ### Step 7. Create head node VM
            $jobArray += Start-Job -Name "CreateVMForHN" -InitializationScript $JobInitScript -ScriptBlock {
                SetSubscriptionStorageAccount -SubscriptionName $using:configSubscriptionName -StorageAccountName $using:configDefaultStorageAccount
                $adUserCred = $using:domainUserCred
                CreateDomainJoinedVM -VMName $using:configHNName `
                                     -ServiceName $using:configHNSvcName `
                                     -VMSize $using:configHNSize `
                                     -ImageName $using:hnImageName `
                                     -VNetName $using:configVNetName `
                                     -SubnetName $using:configSubnetName `
                                     -LocalAdminCred $using:localAdminCred `
                                     -DomainUserCred $adUserCred `
                                     -DomainFQDN $using:configDomainFQDN `
                                     -DeploymentLabel $using:DeploymentLabel `
                                     -HostCaching ReadOnly `
                                     -Wait
                try
                {
                    AddADUserToLocalAdmin -VMName $using:configHNName -ServiceName $using:configHNSvcName -LocalAdminCred $using:localAdminCred -ADUserName $adUserCred.UserName
                }
                catch [System.Management.Automation.RemoteException]
                {
                    if($_.Exception.Message.Contains("is not domain joined"))
                    {
                        TraceWarning "$using:configHNName is not domain joined, trying to join domain $using:configDomainFQDN"
                        VMJoinToDomain -VMName $using:configHNName `
                                       -ServiceName $using:configHNSvcName `
                                       -LocalAdminCred $using:localAdminCred `
                                       -DomainUserCred $using:domainUserCred `
                                       -DomainFQDN $using:configDomainFQDN
                        # Join Domain succeeded, try to add again
                        AddADUserToLocalAdmin -VMName $using:configHNName -ServiceName $using:configHNSvcName -LocalAdminCred $using:localAdminCred -ADUserName $adUserCred.UserName
                    }
                    else
                    {
                        throw
                    }
                }          
            }
         
            $createHNJob = $jobArray[-1]
            Report-Progress -subProgress $ProvisionStepTable["CreateVMForHN"] -job $createHNJob
        }
    }

    ### Step 8. Create Remote DB Server
    $createDBJob = $null
    if(($configDBOption -eq "NewRemoteDB") -and (-not $dbExists))
    {
        $jobArray += Start-Job -Name "CreateDBServer" -InitializationScript $JobInitScript -ScriptBlock {
                TraceInfo "Creating remote DB server $using:configRemoteDBName in cloud service $using:configRemoteDBSvcName"
                SetSubscriptionStorageAccount -SubscriptionName $using:configSubscriptionName -StorageAccountName $using:configDefaultStorageAccount
                CreateDomainJoinedVM -VMName $using:configRemoteDBName `
                                     -ServiceName $using:configRemoteDBSvcName `
                                     -VMSize $using:configRemoteDBSize `
                                     -ImageName $using:dbImageName `
                                     -VNetName $using:configVNetName `
                                     -SubnetName $using:configSubnetName `
                                     -LocalAdminCred $using:localAdminCred `
                                     -DomainUserCred $using:domainUserCred `
                                     -DomainFQDN $using:configDomainFQDN `
                                     -DeploymentLabel $using:deploymentLabel `
                                     -HostCaching ReadOnly `
                                     -Wait

                TraceInfo "The remote DB server $using:configRemoteDBName created in cloud service $using:configRemoteDBSvcName"
                TraceInfo "Creating a data disk with $using:configRemoteDBSvcName GB for the remote DB server $using:configRemoteDBName ..."
                $vm = InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $using:configRemoteDBSvcName -Name $using:configRemoteDBName" | `
                      Add-AzureDataDisk -CreateNew -DiskSizeInGB $using:configRemoteDBDataDiskSize -DiskLabel "HPCDBData" -LUN 0 -HostCaching None
                Update-AzureVMWithRetry -ServiceName $using:configRemoteDBSvcName -Name $using:configRemoteDBName -VM $vm.VM
                TraceInfo "The data disk created for the remote DB server $using:configRemoteDBName."
        }
         
        $createDBJob = $jobArray[-1]
        Report-Progress -subProgress $ProvisionStepTable["CreateDBServer"] -job $createDBJob
    }

    ### Step 9. Create Compute Nodes
    $createCNsJob = $null
    if ($cnServiceVMTable.Count -gt 0)
    {
        
        $jobArray += Start-Job -Name "CreateComputeNodes" -InitializationScript $JobInitScript -ScriptBlock {
            CreateHPCNodeGroup -ServiceVMTable $using:cnServiceVMTable `
                               -ServiceStorageTable $using:configCNServiceStorageTable `
                               -VMSize $using:configCNSize `
                               -ClusterName $using:configHNName `
                               -HeadNodeServiceName $using:configHNSvcName `
                               -NodeType ComputeNode `
                               -VNetName $using:configVNetName `
                               -SubnetName $using:configSubnetName `
                               -DeploymentLabel $using:deploymentLabel `
                               -LocalAdminCred $using:localAdminCred `
                               -DomainUserCred $using:domainUserCred `
                               -DomainFQDN $using:configDomainFQDN `
                               -SourceImageName $using:configCNImage `
                               -SrcImageHPCInstalled:$using:configCNImageHPCInstalled
        }

        $createCNsJob = $jobArray[-1]
        Report-Progress -subProgress $ProvisionStepTable["CreateComputeNodes"] -job $createCNsJob
    }

    ### Step 10. Create Broker Nodes
    $createBNsJob = $null
    if ($bnServiceVMTable.Count -gt 0)
    {
        $jobArray += Start-Job -Name "CreateBrokerNodes" -InitializationScript $JobInitScript -ScriptBlock {
            CreateHPCNodeGroup -ServiceVMTable $using:bnServiceVMTable `
                               -ServiceStorageTable $using:configBNServiceStorageTable `
                               -VMSize $using:configBNSize `
                               -ClusterName $using:configHNName `
                               -HeadNodeServiceName $using:configHNSvcName `
                               -NodeType BrokerNode `
                               -VNetName $using:configVNetName `
                               -SubnetName $using:configSubnetName `
                               -DeploymentLabel $using:deploymentLabel `
                               -LocalAdminCred $using:localAdminCred `
                               -DomainUserCred $using:domainUserCred `
                               -DomainFQDN $using:configDomainFQDN `
                               -SourceImageName $using:configBNImage `
                               -SrcImageHPCInstalled:$using:configBNImageHPCInstalled
        }

        $createBNsJob = $jobArray[-1]
        Report-Progress -subProgress $ProvisionStepTable["CreateBrokerNodes"] -job $createBNsJob
    }

    ### Step 11. Config Remote DB Server
    if($createDBJob -ne $null)
    {
        if($createHNJob -ne $null)
        {
            Wait-Progress -Jobs @($createDBJob, $createHNJob)
            CheckJobStatus -Jobs @($createDBJob, $createHNJob)
        }
        else
        {
            Wait-Progress -Jobs $createDBJob
            CheckJobStatus -Jobs $createDBJob      
        }

        $jobArray += Start-Job -Name "ConfigDBServer" -InitializationScript $JobInitScript -ScriptBlock {
            $adUserCred = $using:domainUserCred
            try
            {
                ConfigRemoteDBServer -VMName $using:configRemoteDBName `
                                     -ServiceName $using:configRemoteDBSvcName `
                                     -DBAdminCred $using:localAdminCred `
                                     -HpcSetupUserName $adUserCred.UserName `
                                     -HNMachineAccount "$using:domainNetBios\$using:configHNName$"
            }
            catch [System.Management.Automation.RemoteException]
            {
                if($_.Exception.Message.Contains("is not domain joined"))
                {
                    TraceWarning "$using:configRemoteDBName is not domain joined, trying to join domain $using:configDomainFQDN"
                    VMJoinToDomain -VMName $using:configRemoteDBName `
                                   -ServiceName $using:configRemoteDBSvcName `
                                   -LocalAdminCred $using:localAdminCred `
                                   -DomainUserCred $adUserCred `
                                   -DomainFQDN $using:configDomainFQDN

                    # Join Domain succeeded, try to configure remote DB again
                    ConfigRemoteDBServer -VMName $using:configRemoteDBName `
                                         -ServiceName $using:configRemoteDBSvcName `
                                         -DBAdminCred $using:localAdminCred `
                                         -HpcSetupUserName $adUserCred.UserName `
                                         -HNMachineAccount "$using:domainNetBios\$using:configHNName$"
                }
                else
                {
                    throw
                }
            }
        }

        $configDBJob = $jobArray[-1]
            Report-Progress -subProgress $ProvisionStepTable["ConfigDBServer"] -job $configDBJob

        # Wait for its completion
        Wait-Progress -Jobs $configDBJob
        CheckJobStatus -Jobs $configDBJob
    }

    ### Step 12. Install head node
    if(-not $hnExists)
    {
        if($createHNJob -ne $null)
        {
            Wait-Progress -Jobs $createHNJob
            CheckJobStatus -Jobs $createHNJob
        }

        if($configDBOption -eq "LocalDB")
        {
            $jobArray += Start-Job -Name "PrepareHeadNode" -InitializationScript $JobInitScript -ScriptBlock {
                PrepareHeadNode -VMName $using:configHNName `
                                -ServiceName $using:configHNSvcName `
                                -SetupUserCred $using:domainUserCred `
                                -LocalDB `
                                -SubscriptionId  $using:subscriptionId `
                                -VNetName $using:configVNetName `
                                -SubNetName $using:configSubnetName `
                                -AffinityGroupName $using:configHNAGName `
                                -Location $using:configLocation `
                                -PreConfigScirptFile $using:configHNPreConfigScript
            }
        }
        elseif($configDBOption -eq "NewRemoteDB")
        {
            $jobArray += Start-Job -Name "PrepareHeadNode" -InitializationScript $JobInitScript -ScriptBlock {
                PrepareHeadNode -VMName $using:configHNName `
                                -ServiceName $using:configHNSvcName `
                                -SetupUserCred $using:domainUserCred `
                                -RemoteDB `
                                -DBServerInstance $using:configRemoteDBName `
                                -SubscriptionId  $using:subscriptionId `
                                -VNetName $using:configVNetName `
                                -SubNetName $using:configSubnetName `
                                -AffinityGroupName $using:configHNAGName `
                                -Location $using:configLocation `
                                -PreConfigScirptFile $using:configHNPreConfigScript
            }
        }
        elseif($configDBOption -eq "ExistingDB")
        {
            $jobArray += Start-Job -Name "PrepareHeadNode" -InitializationScript $JobInitScript -ScriptBlock {
                PrepareHeadNode -VMName $using:configHNName `
                                -ServiceName $using:configHNSvcName `
                                -SetupUserCred $using:domainUserCred `
                                -RemoteDB `
                                -DBServerInstance $using:configDBServerInstance `
                                -ManagementDB $using:configMgmtDBName `
                                -SchedulerDB $using:configSchdDBName `
                                -MonitoringDB $using:configMonDBName `
                                -ReportingDB $using:configRptDBName `
                                -DiagnosticsDB $using:configDiagDBName `
                                -SubscriptionId  $using:subscriptionId `
                                -VNetName $using:configVNetName `
                                -SubNetName $using:configSubnetName `
                                -AffinityGroupName $using:configHNAGName `
                                -Location $using:configLocation `
                                -PreConfigScirptFile $using:configHNPreConfigScript
            }
        }

        $prepareHNJob = $jobArray[-1]
        Report-Progress -subProgress $ProvisionStepTable["PrepareHeadNode"] -job $prepareHNJob        

        ### Step 13. Config head node
        Wait-Progress -Jobs $prepareHNJob
        CheckJobStatus -Jobs $prepareHNJob
        $jobArray += Start-Job -Name "ConfigHeadNode" -InitializationScript $JobInitScript -ScriptBlock {
            # If AzureBurst configured, the PostConfig script shall be scheduled in Azure Burst step.
            $hnPostConfigFile = $using:configHNPostConfigScript
            if($using:configAzureBurst)
            {
                $hnPostConfigFile = ""
            }
            ConfigHeadNode -VMName $using:configHNName `
                           -ServiceName $using:configHNSvcName `
                           -SetupUserCred $using:domainUserCred `
                           -NodeNamePattern $using:NodeNamePattern `
                           -EnableRESTAPI:$using:configHNEnableRESTAPI `
                           -EnableWebPortal:$using:configHNEnableWebPortal `
                           -PostConfigScirptFile $hnPostConfigFile
        }

        $configHNJob = $jobArray[-1]
        Report-Progress -subProgress $ProvisionStepTable["ConfigHeadNode"] -job $configHNJob
        Wait-Progress -Jobs $configHNJob
        CheckJobStatus -Jobs $configHNJob

        ### Step 14. Azure Burst
        if($configAzureBurst)
        {
            $burstConfigXmlStr = $config["AzureBurst"].OuterXml
            $jobArray += Start-Job -Name "CreatePaaSNodes" -InitializationScript $JobInitScript -ScriptBlock {
                try
                {
                    $burstConfig = GetAzureBustConfig -AzureBustConfigXml $using:burstConfigXmlStr -CertPasswords $using:certificatePasswords
                    CreateAzureNodeTemplates -VMName $using:configHNName `
                                             -ServiceName $using:configHNSvcName `
                                             -SetupUserCred $using:domainUserCred `
                                             -CertList $burstConfig["CertList"] `
                                             -NodeTemplates $burstConfig["TempList"] `
                                             -PostConfigScirptFile $using:configHNPostConfigScript
                }
                catch
                {
                    throw
                }
            }

            $createPaaSJob = $jobArray[-1]
            Report-Progress -subProgress $ProvisionStepTable["CreatePaaSNodes"] -job $createPaaSJob
            Wait-Progress -Jobs $createPaaSJob
        }
    }

    Wait-Progress
    CheckJobStatus -Jobs $jobArray
    $ProvisionState = "Succeeded"
    Write-Host

    $baseImages = @()
    if($createCNsJob -ne $null)
    {
        $baseImages += "HPC compute node image :  " + $createCNsJob.ChildJobs[0].Output[-1]
    }
    if($createBNsJob -ne $null)
    {
        $baseImages += "HPC broker node image  :  " + $createBNsJob.ChildJobs[0].Output[-1]
    }

    $deployCostTime = (Get-Date) - $StartDeployTime
    Write-Host ("The HPC cluster deployment completed. Total cost time: {0:N2} minutes." -f $deployCostTime.TotalMinutes) -ForegroundColor Green
    if(-not $hnExists -and ($configHNEnableRESTAPI -or $configHNEnableWebPortal))
    {
        $webCompCerFile = $configHNJob.ChildJobs[0].Output[-1]
        Write-Host "You have enabled REST API or web portal on HPC Pack head node. Please import the following certificate in the Trusted Root Certification Authorities certificate store on the computer where you are submitting job or accessing the HPC web portal: $webCompCerFile" -ForegroundColor Green
    }
    if($baseImages.Count -gt 0)
    {
        Write-Host "You can use the image name(s) as below to add more HPC Pack nodes later:" -ForegroundColor Green
        $baseImages | Write-Host -ForegroundColor Green
    }
}
catch
{
    $ProvisionState = "Failed"
    Write-Host "`nThe HPC cluster deployment failed: `n" -ForegroundColor Red
    $_ | Out-String | Write-Host -ForegroundColor Red
    exit 1
}
finally
{
    Uninitialize-Progress
    if($jobArray.Count -ne 0)
    {
        WriteJobLogsToFile $jobArray $LogFile 
    }

    # Remove the certificates from Store
    if((-not $PSSessionSkipCACheck.IsPresent) -and (Test-Path -Path $WinRMCertListFile -PathType Leaf))
    {
        Get-Content $WinRMCertListFile | Remove-Item -Force -ErrorAction SilentlyContinue
        Remove-Item  $WinRMCertListFile -Force
    }

    if($ProvisionState -eq "Running")
    {
        Write-Host "`nThe HPC cluster deployment was stopped.`n" -ForegroundColor Red
    }

    if(($ProvisionState -eq "Failed") -or ($ProvisionState -eq "Running"))
    {
        $jobArray | Stop-Job

        $queuedJobNames = @()
        $completedJobNames = @()
        $notCompletedJobNames = @()

        if($jobArray.Count -ne $ProvisionStepTable.Count)
        {
            $queuedJobNames = $ProvisionStepTable.Keys | ?{$jobArray.Name -notcontains $_}
        }

        foreach($job in $jobArray)
        {
            if($job.State -eq "Completed")
            {
                $completedJobNames += $job.Name
            }
            else
            {
                $notCompletedJobNames += $job.Name
            }
        }

        if($completedJobNames.Count -ne $ProvisionStepTable.Count)
        {
            $CleanUpVMTable = @{}
            if(($notCompletedJobNames -contains "CreateVMForDC") -or ($notCompletedJobNames -contains "PromoteDC") -or `
                (($completedJobNames -contains "CreateVMForDC") -and ($queuedJobNames -contains "PromoteDC")))
            {
                $CleanUpVMTable[$configDCName] = $configDCSvcName
            }

            if(($notCompletedJobNames -contains "CreateDBServer") -or ($notCompletedJobNames -contains "ConfigDBServer") -or `
                (($completedJobNames -contains "CreateDBServer") -and ($queuedJobNames -contains "ConfigDBServer")))
            {
                $CleanUpVMTable[$configRemoteDBName] = $configRemoteDBSvcName
            }

            if(($notCompletedJobNames -contains "CreateVMForHN") -or ($notCompletedJobNames -contains "ConfigHeadNode") -or `
                ($notCompletedJobNames -contains "PrepareHeadNode") -or `
                (($completedJobNames -contains "CreateVMForHN") -and ($queuedJobNames -contains "ConfigHeadNode")) -or `
                (($configDCOption -eq "HeadNodeAsDC") -and ($completedJobNames -contains "CreateVMForDC") -and ($queuedJobNames -contains "ConfigHeadNode")))
            {
                $CleanUpVMTable[$configHNName] = $configHNSvcName
            }

            if($notCompletedJobNames -contains "CreateComputeNodes")
            {
                if(-not $configCNImageHPCInstalled)
                {
                    $baseCNSvcName = $cnServiceVMTable.Keys | sort | select -First 1
                    $baseCNName = $cnServiceVMTable[$baseCNSvcName][0]
                    $baseCN = @(InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $baseCNSvcName -Name $baseCNName -WarningAction SilentlyContinue")
                    if(($baseCN -ne $null) -and -not (IsBaseHPCNodeImage -VMSrcImageName $baseCN.VM.OSVirtualHardDisk.SourceImageName -NodeType ComputeNode))
                    {
                        $CleanUpVMTable[$baseCN.Name] = $baseCNSvcName
                    }
                }
            }

            if($notCompletedJobNames -contains "CreateBrokerNodes")
            {
                if(-not $configBNImageHPCInstalled)
                {
                    $baseBNSvcName = $bnServiceVMTable.Keys | sort | select -First 1
                    $baseBNName = $bnServiceVMTable[$baseBNSvcName][0]
                    $baseBN = @(InvokeAzureCmdWithRetry -Command "Get-AzureVM -ServiceName $baseBNSvcName -Name $baseBNName -WarningAction SilentlyContinue")
                    if($baseBN -ne $null -and -not (IsBaseHPCNodeImage -VMSrcImageName $baseBN.VM.OSVirtualHardDisk.SourceImageName -NodeType BrokerNode))
                    {
                        $CleanUpVMTable[$baseBN.Name] = $baseBNSvcName
                    }
                }
            }

            if($CleanUpVMTable.Count -gt 0)
            {
                $msg = "The following virtual machines(s) were created but not successfully deployed, you must remove them before you re-run the script:`n" + ($CleanUpVMTable.Keys -join ",")
                Write-Host $msg -ForegroundColor Red

                if(-not $NoCleanOnFailure.IsPresent)
                {
                    if(-not $PsCmdlet.ShouldProcess("Remove the virtual machines which are not successfully deployed", "Do you want to remove them now?", "Confirm"))
                    {
                        Write-Warning "You must remove them manually before you re-run the script, or the deployment may fail"
                        $CleanUpVMTable.Clear()
                    }
                }
                else
                {
                    $CleanUpVMTable.Clear()
                }
            }

            foreach($vmname in $CleanUpVMTable.Keys)
            {
                try
                {
                    Remove-AzureVMWithRetry -VMName $vmname -ServiceName $CleanUpVMTable[$vmname]
                }
                catch
                {
                    TraceWarning "$_"
                }
            }
        }
    }
        
    # Keep the not completed jobs for investigation.
    $jobArray | ?{$_.State -eq "Completed"} | Remove-Job
}

# SIG # Begin signature block
# MIIavwYJKoZIhvcNAQcCoIIasDCCGqwCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUQxBu4QzL7n2hRXOXElmMjWeq
# x26gghV6MIIEuzCCA6OgAwIBAgITMwAAAF3JyvZpIzdoUAAAAAAAXTANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTQwNTIzMTcxMzE3
# WhcNMTUwODIzMTcxMzE3WjCBqzELMAkGA1UEBhMCVVMxCzAJBgNVBAgTAldBMRAw
# DgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24x
# DTALBgNVBAsTBE1PUFIxJzAlBgNVBAsTHm5DaXBoZXIgRFNFIEVTTjo3RDJFLTM3
# ODItQjBGNzElMCMGA1UEAxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZTCC
# ASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAKu6ei+vVB909A1zsuALhVjs
# q6u/RwD+jTcx8hyTPqa6zhvmUUBo1IAY11hjXQM3NvuDWHDexDM2sF8Hf33991ir
# WpduRc648Rs9Us+lPQVDm6CxNHLT7/MxvC5ojio1Cy50sO/kK1VyZZWPZ/SjyO8D
# EafcCzyhbA7/YYxCkuRWw5Gxj/au18Hj6KgZAaIXYL9DmHoqWmo0yJINP+BcHbd/
# sFodvcxsahFDHgINW0LpbUtjCzn7UqPseibSojBHY9L46hXYaS6eJulOTxyz0uJg
# Jj9uXZIUkF40bXgEfFJRmwsBdutBAFlTztuN1kmFbII7fsTph0ayOTzxXthvcv8C
# AwEAAaOCAQkwggEFMB0GA1UdDgQWBBTXVOYV2dNdyiTVARjweGSZe2hMMzAfBgNV
# HSMEGDAWgBQjNPjZUkZwCu1A+3b7syuwwzWzDzBUBgNVHR8ETTBLMEmgR6BFhkNo
# dHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNyb3Nv
# ZnRUaW1lU3RhbXBQQ0EuY3JsMFgGCCsGAQUFBwEBBEwwSjBIBggrBgEFBQcwAoY8
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNyb3NvZnRUaW1l
# U3RhbXBQQ0EuY3J0MBMGA1UdJQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEBBQUA
# A4IBAQCJMKxtI1M8/5xen1dwDBzeF903pNyVsaYkzAfuqYXRHNXZm0k3IZ8xPmmu
# v7UwG+ekLFDrwVCISqq17KiyCUCwau9zVEKtnimFVG2zD/GYfuP4wDAbTNKGKfCi
# 5V6bN41mnei8W2+6lcb20i5BNfnHvgyK5iZEdmgcnKQW7iJVpNZXK/gNgSAsoQ+a
# jvzHlcVeh4RB4tJvnM/QZSMW29B9hNAGFhKlZy17hYk+RZPYzSaZ+7Yq4VJCmhCX
# HzTakcGshvOSRG1Qu4mykew1qEj2weHE/sZlfUHv7W3KkeTmsNnlWFy0u5romkYc
# Zt0eS0SYyfRKWBF/kMshN4hDyN71MIIE7DCCA9SgAwIBAgITMwAAAMps1TISNcTh
# VQABAAAAyjANBgkqhkiG9w0BAQUFADB5MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSMwIQYDVQQDExpNaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QTAeFw0xNDA0MjIxNzM5MDBaFw0xNTA3MjIxNzM5MDBaMIGDMQswCQYDVQQGEwJV
# UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
# ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMQ0wCwYDVQQLEwRNT1BSMR4wHAYDVQQD
# ExVNaWNyb3NvZnQgQ29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAw
# ggEKAoIBAQCWcV3tBkb6hMudW7dGx7DhtBE5A62xFXNgnOuntm4aPD//ZeM08aal
# IV5WmWxY5JKhClzC09xSLwxlmiBhQFMxnGyPIX26+f4TUFJglTpbuVildGFBqZTg
# rSZOTKGXcEknXnxnyk8ecYRGvB1LtuIPxcYnyQfmegqlFwAZTHBFOC2BtFCqxWfR
# +nm8xcyhcpv0JTSY+FTfEjk4Ei+ka6Wafsdi0dzP7T00+LnfNTC67HkyqeGprFVN
# TH9MVsMTC3bxB/nMR6z7iNVSpR4o+j0tz8+EmIZxZRHPhckJRIbhb+ex/KxARKWp
# iyM/gkmd1ZZZUBNZGHP/QwytK9R/MEBnAgMBAAGjggFgMIIBXDATBgNVHSUEDDAK
# BggrBgEFBQcDAzAdBgNVHQ4EFgQUH17iXVCNVoa+SjzPBOinh7XLv4MwUQYDVR0R
# BEowSKRGMEQxDTALBgNVBAsTBE1PUFIxMzAxBgNVBAUTKjMxNTk1K2I0MjE4ZjEz
# LTZmY2EtNDkwZi05YzQ3LTNmYzU1N2RmYzQ0MDAfBgNVHSMEGDAWgBTLEejK0rQW
# WAHJNy4zFha5TJoKHzBWBgNVHR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1pY3Jv
# c29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNDb2RTaWdQQ0FfMDgtMzEtMjAx
# MC5jcmwwWgYIKwYBBQUHAQEETjBMMEoGCCsGAQUFBzAChj5odHRwOi8vd3d3Lm1p
# Y3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY0NvZFNpZ1BDQV8wOC0zMS0yMDEwLmNy
# dDANBgkqhkiG9w0BAQUFAAOCAQEAd1zr15E9zb17g9mFqbBDnXN8F8kP7Tbbx7Us
# G177VAU6g3FAgQmit3EmXtZ9tmw7yapfXQMYKh0nfgfpxWUftc8Nt1THKDhaiOd7
# wRm2VjK64szLk9uvbg9dRPXUsO8b1U7Brw7vIJvy4f4nXejF/2H2GdIoCiKd381w
# gp4YctgjzHosQ+7/6sDg5h2qnpczAFJvB7jTiGzepAY1p8JThmURdwmPNVm52Iao
# AP74MX0s9IwFncDB1XdybOlNWSaD8cKyiFeTNQB8UCu8Wfz+HCk4gtPeUpdFKRhO
# lludul8bo/EnUOoHlehtNA04V9w3KDWVOjic1O1qhV0OIhFeezCCBbwwggOkoAMC
# AQICCmEzJhoAAAAAADEwDQYJKoZIhvcNAQEFBQAwXzETMBEGCgmSJomT8ixkARkW
# A2NvbTEZMBcGCgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9z
# b2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5MB4XDTEwMDgzMTIyMTkzMloX
# DTIwMDgzMTIyMjkzMloweTELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0
# b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3Jh
# dGlvbjEjMCEGA1UEAxMaTWljcm9zb2Z0IENvZGUgU2lnbmluZyBQQ0EwggEiMA0G
# CSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCycllcGTBkvx2aYCAgQpl2U2w+G9Zv
# zMvx6mv+lxYQ4N86dIMaty+gMuz/3sJCTiPVcgDbNVcKicquIEn08GisTUuNpb15
# S3GbRwfa/SXfnXWIz6pzRH/XgdvzvfI2pMlcRdyvrT3gKGiXGqelcnNW8ReU5P01
# lHKg1nZfHndFg4U4FtBzWwW6Z1KNpbJpL9oZC/6SdCnidi9U3RQwWfjSjWL9y8lf
# RjFQuScT5EAwz3IpECgixzdOPaAyPZDNoTgGhVxOVoIoKgUyt0vXT2Pn0i1i8UU9
# 56wIAPZGoZ7RW4wmU+h6qkryRs83PDietHdcpReejcsRj1Y8wawJXwPTAgMBAAGj
# ggFeMIIBWjAPBgNVHRMBAf8EBTADAQH/MB0GA1UdDgQWBBTLEejK0rQWWAHJNy4z
# Fha5TJoKHzALBgNVHQ8EBAMCAYYwEgYJKwYBBAGCNxUBBAUCAwEAATAjBgkrBgEE
# AYI3FQIEFgQU/dExTtMmipXhmGA7qDFvpjy82C0wGQYJKwYBBAGCNxQCBAweCgBT
# AHUAYgBDAEEwHwYDVR0jBBgwFoAUDqyCYEBWJ5flJRP8KuEKU5VZ5KQwUAYDVR0f
# BEkwRzBFoEOgQYY/aHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJv
# ZHVjdHMvbWljcm9zb2Z0cm9vdGNlcnQuY3JsMFQGCCsGAQUFBwEBBEgwRjBEBggr
# BgEFBQcwAoY4aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNy
# b3NvZnRSb290Q2VydC5jcnQwDQYJKoZIhvcNAQEFBQADggIBAFk5Pn8mRq/rb0Cx
# MrVq6w4vbqhJ9+tfde1MOy3XQ60L/svpLTGjI8x8UJiAIV2sPS9MuqKoVpzjcLu4
# tPh5tUly9z7qQX/K4QwXaculnCAt+gtQxFbNLeNK0rxw56gNogOlVuC4iktX8pVC
# nPHz7+7jhh80PLhWmvBTI4UqpIIck+KUBx3y4k74jKHK6BOlkU7IG9KPcpUqcW2b
# Gvgc8FPWZ8wi/1wdzaKMvSeyeWNWRKJRzfnpo1hW3ZsCRUQvX/TartSCMm78pJUT
# 5Otp56miLL7IKxAOZY6Z2/Wi+hImCWU4lPF6H0q70eFW6NB4lhhcyTUWX92THUmO
# Lb6tNEQc7hAVGgBd3TVbIc6YxwnuhQ6MT20OE049fClInHLR82zKwexwo1eSV32U
# jaAbSANa98+jZwp0pTbtLS8XyOZyNxL0b7E8Z4L5UrKNMxZlHg6K3RDeZPRvzkbU
# 0xfpecQEtNP7LN8fip6sCvsTJ0Ct5PnhqX9GuwdgR2VgQE6wQuxO7bN2edgKNAlt
# HIAxH+IOVN3lofvlRxCtZJj/UBYufL8FIXrilUEnacOTj5XJjdibIa4NXJzwoq6G
# aIMMai27dmsAHZat8hZ79haDJLmIz2qoRzEvmtzjcT3XAH5iR9HOiMm4GPoOco3B
# oz2vAkBq/2mbluIQqBC0N1AI1sM9MIIGBzCCA++gAwIBAgIKYRZoNAAAAAAAHDAN
# BgkqhkiG9w0BAQUFADBfMRMwEQYKCZImiZPyLGQBGRYDY29tMRkwFwYKCZImiZPy
# LGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQDEyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZp
# Y2F0ZSBBdXRob3JpdHkwHhcNMDcwNDAzMTI1MzA5WhcNMjEwNDAzMTMwMzA5WjB3
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEwHwYDVQQDExhN
# aWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAw
# ggEKAoIBAQCfoWyx39tIkip8ay4Z4b3i48WZUSNQrc7dGE4kD+7Rp9FMrXQwIBHr
# B9VUlRVJlBtCkq6YXDAm2gBr6Hu97IkHD/cOBJjwicwfyzMkh53y9GccLPx754gd
# 6udOo6HBI1PKjfpFzwnQXq/QsEIEovmmbJNn1yjcRlOwhtDlKEYuJ6yGT1VSDOQD
# LPtqkJAwbofzWTCd+n7Wl7PoIZd++NIT8wi3U21StEWQn0gASkdmEScpZqiX5NMG
# gUqi+YSnEUcUCYKfhO1VeP4Bmh1QCIUAEDBG7bfeI0a7xC1Un68eeEExd8yb3zuD
# k6FhArUdDbH895uyAc4iS1T/+QXDwiALAgMBAAGjggGrMIIBpzAPBgNVHRMBAf8E
# BTADAQH/MB0GA1UdDgQWBBQjNPjZUkZwCu1A+3b7syuwwzWzDzALBgNVHQ8EBAMC
# AYYwEAYJKwYBBAGCNxUBBAMCAQAwgZgGA1UdIwSBkDCBjYAUDqyCYEBWJ5flJRP8
# KuEKU5VZ5KShY6RhMF8xEzARBgoJkiaJk/IsZAEZFgNjb20xGTAXBgoJkiaJk/Is
# ZAEZFgltaWNyb3NvZnQxLTArBgNVBAMTJE1pY3Jvc29mdCBSb290IENlcnRpZmlj
# YXRlIEF1dGhvcml0eYIQea0WoUqgpa1Mc1j0BxMuZTBQBgNVHR8ESTBHMEWgQ6BB
# hj9odHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9taWNy
# b3NvZnRyb290Y2VydC5jcmwwVAYIKwYBBQUHAQEESDBGMEQGCCsGAQUFBzAChjho
# dHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY3Jvc29mdFJvb3RD
# ZXJ0LmNydDATBgNVHSUEDDAKBggrBgEFBQcDCDANBgkqhkiG9w0BAQUFAAOCAgEA
# EJeKw1wDRDbd6bStd9vOeVFNAbEudHFbbQwTq86+e4+4LtQSooxtYrhXAstOIBNQ
# md16QOJXu69YmhzhHQGGrLt48ovQ7DsB7uK+jwoFyI1I4vBTFd1Pq5Lk541q1YDB
# 5pTyBi+FA+mRKiQicPv2/OR4mS4N9wficLwYTp2OawpylbihOZxnLcVRDupiXD8W
# mIsgP+IHGjL5zDFKdjE9K3ILyOpwPf+FChPfwgphjvDXuBfrTot/xTUrXqO/67x9
# C0J71FNyIe4wyrt4ZVxbARcKFA7S2hSY9Ty5ZlizLS/n+YWGzFFW6J1wlGysOUzU
# 9nm/qhh6YinvopspNAZ3GmLJPR5tH4LwC8csu89Ds+X57H2146SodDW4TsVxIxIm
# dgs8UoxxWkZDFLyzs7BNZ8ifQv+AeSGAnhUwZuhCEl4ayJ4iIdBD6Svpu/RIzCzU
# 2DKATCYqSCRfWupW76bemZ3KOm+9gSd0BhHudiG/m4LBJ1S2sWo9iaF2YbRuoROm
# v6pH8BJv/YoybLL+31HIjCPJZr2dHYcSZAI9La9Zj7jkIeW1sMpjtHhUBdRBLlCs
# lLCleKuzoJZ1GtmShxN1Ii8yqAhuoFuMJb+g74TKIdbrHk/Jmu5J4PcBZW+JC33I
# acjmbuqnl84xKf8OxVtc2E0bodj6L54/LlUWa8kTo/0xggSvMIIEqwIBATCBkDB5
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSMwIQYDVQQDExpN
# aWNyb3NvZnQgQ29kZSBTaWduaW5nIFBDQQITMwAAAMps1TISNcThVQABAAAAyjAJ
# BgUrDgMCGgUAoIHIMBkGCSqGSIb3DQEJAzEMBgorBgEEAYI3AgEEMBwGCisGAQQB
# gjcCAQsxDjAMBgorBgEEAYI3AgEVMCMGCSqGSIb3DQEJBDEWBBSI5ZYsiQX6aKJh
# MIZ1hcLCie9fUzBoBgorBgEEAYI3AgEMMVowWKA2gDQATQBpAGMAcgBvAHMAbwBm
# AHQAIABIAFAAQwAgAFAAYQBjAGsAIAAyADAAMQAyACAAUgAyoR6AHGh0dHA6Ly93
# d3cubWljcm9zb2Z0LmNvbS9IUEMwDQYJKoZIhvcNAQEBBQAEggEAa5e9171aINcd
# sXTrNdTqjohg3F/tFL1mxbdtMZZbD6APNgW3eU74+sXoIzRvV7Ah6kpKBB7QD6Tx
# D14+7NwZofv8XhcfA4+BN1aVeDAtH4WIyOhTGsVDQh9NaxWKJDcC7ulADSCy3WIL
# clGnAc2/ehyo/5f6D6kSOi254WtzQ8krkytVGPXYIp7sT1R0ymaQUJcGyd6WviDR
# 7ktu198vQQlIRxX0qPXbibQOn7LwVdr9ts2yciCLSzQ+hfnb/Vl55+Ahuc09YIhU
# 9OZtP7VQL5ASoJ5VtJQ3xFGj9/2z3SD3qya5A+v1ULsDQClvw9yy+u0hhT5cx7Bp
# FQeeogWtNqGCAigwggIkBgkqhkiG9w0BCQYxggIVMIICEQIBATCBjjB3MQswCQYD
# VQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEe
# MBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEwHwYDVQQDExhNaWNyb3Nv
# ZnQgVGltZS1TdGFtcCBQQ0ECEzMAAABdycr2aSM3aFAAAAAAAF0wCQYFKw4DAhoF
# AKBdMBgGCSqGSIb3DQEJAzELBgkqhkiG9w0BBwEwHAYJKoZIhvcNAQkFMQ8XDTE1
# MDIxMjE4MTQ1NFowIwYJKoZIhvcNAQkEMRYEFKrf8CTY/04atBfGWlb4X749vj5M
# MA0GCSqGSIb3DQEBBQUABIIBABhnFJ2Jb72Ii93Qin6Cb/mhdbNVLO9NW+o6pkpl
# fZUkmK88RidGVNLXJ4BaVMzR+RyBuXlweiK1eu8Z2xpTXvQ7MCUY4qkbw8Ywl5YE
# 7Np/ELJvlsvshWFx24xx1D89L9HU8rLiSr3/wP4HmxswiC61jBRISgfKeqr2gj7v
# xLG51QzgyZ8G26YSvrKbjlotbuDlK3QK4H2XNsjubSrSV6iXnk+Qr1s1BAvMcyj7
# sI6bxjWKnieslSDpUJy4legELIx8jfztOY46j6wgnJbGpxP9fgKcC6Et67nD5blO
# B3g1NZAyuemwm+qLIc6riar7VSX7AFa1ojQqtEK8b6OcYiI=
# SIG # End signature block
