# private
# Clear all reported progress.
function Clear-Progress()
{
    [Tracing]::states.Clear();
}

# private
# Add a datetime prefix before $log
function AppendDateTime($log)
{
    $date = Get-Date -Format 'MM/dd/yyyy HH:mm:ss' 
    "{0} - {1}" -f $date, $log
}    

# private
# Set the cursor to the saved end.
function Set-CursorEnd()
{
    if ([Tracing]::MaxLine -gt $host.UI.RawUI.CursorPosition.Y)
    {
        $old = $host.UI.RawUI.CursorPosition
        $old.Y = [Tracing]::MaxLine
        $host.UI.RawUI.CursorPosition = $old
    }
}

# private
# Save the current cursor end to a variable.
function Save-CursorEnd()
{
    if ([Tracing]::MaxLine -lt $host.UI.RawUI.CursorPosition.Y)
    {
        [Tracing]::MaxLine = $host.UI.RawUI.CursorPosition.Y
    }
}

# public
# Wait for all reported progress to complete, display the progress bar, progress notification to screen.
# This function will block until all progress complete.
function Wait-Progress
{
    [CmdletBinding(DefaultParametersetName="Jobs")] 
    Param
    (
        # No Wait
        [Parameter(Mandatory=$false, Position=0, ParameterSetName="NoWait")]
        [Switch] $NoWait,

        # Wait jobs
        [Parameter(Mandatory=$false, Position=0, ParameterSetName="Jobs")]
        [System.Management.Automation.Job[]] $Jobs,

        [Parameter(Mandatory=$false, Position=1, ParameterSetName="Jobs")]
        [Switch] $Any
    )

    Set-CursorEnd

    $windowWidth = $host.UI.RawUI.WindowSize.Width
    $dateWidth = 17
    $progressWidth = 5
    $statusWidth = 10
    $nameWidth = 38

    $textWidth = $windowWidth - $progressWidth - $statusWidth - $dateWidth - $nameWidth - 1

    $scroll = "/-\|/-\|"

    if ($NoWait)
    {
        $Jobs = @();
    }

    if ($Jobs -eq $null)
    {
        $Jobs = @([Tracing]::states.GetEnumerator() | Where-Object { $_.Value.Job -ne $null } | %{ $_.Value.Job })
    }

    $jobNum = $Jobs.Count

    do
    {
        Start-Sleep -Milliseconds 300

        $exitLoop = $false
        $doneJobs = @($Jobs | Where-Object { ($_.State -eq "Completed") -or ($_.State -eq "Failed") })
        if(($doneJobs.Count -eq $Jobs.Count) -or ($Any -and ($doneJobs.Count -gt 0)))
        {
            $exitLoop = $true
        }
        
        foreach($subProgress in ([Tracing]::states.GetEnumerator() | Where-Object { !$_.Value.Ended }))
        {
            $currentState = $subProgress.Value;

            $messageColor = [ConsoleColor]::White

            if ($currentState.Job -ne $null)
            {
                $job = $currentState.Job;
                if ($job.JobStateInfo.State -eq [System.Management.Automation.JobState]::Completed -or
                    $job.JobStateInfo.State -eq [System.Management.Automation.JobState]::Failed)
                {
                    $currentState.Ended = $true;
                }

                $currentState.Message = $null

                $messageColor = [ConsoleColor]::White

                if ($job.ChildJobs[0].JobStateInfo.Reason -ne $null)
                {
                    $messageColor = [ConsoleColor]::Red
                    $currentState.Message = $job.ChildJobs[0].JobStateInfo.Reason.Message
                }

                if (!$currentState.Message)
                {
                    $messageColor = [ConsoleColor]::Red
                    $errors = $job.ChildJobs[0].Error | ?{ $_.Exception -ne $null }

                    if ($errors)
                    {
                        $currentState.Message = $errors[-1]
                    }

                    if ($currentState.Message)
                    {
                        $start = $currentState.Message.IndexOf("-");
                        if ($start -gt 0 -and $currentState.Message.Length -gt $start + 2)
                        {
                            $currentState.Message = $currentState.Message.Substring($start + 2);
                        }
                    }
                }

                if (!$currentState.Message)
                {
                    $messageColor = [ConsoleColor]::White

                    $msgs = $job.ChildJobs[0].Verbose | ?{ $_.Message -ne $null -and
                        ($_.Message.StartsWith("[Information]") -or $_.Message.StartsWith("[Warning]")) }

                    if ($msgs)
                    {
                        $currentState.Message = $msgs[-1]
                    }

                    if ($currentState.Message)
                    {
                        if ($currentState.Message.StartsWith("[Warning]"))
                        {
                            $messageColor = [ConsoleColor]::Yellow
                        }

                        $start = $currentState.Message.IndexOf("-");
                        if ($start -gt 0 -and $currentState.Message.Length -gt $start + 2)
                        {
                            $currentState.Message = $currentState.Message.Substring($start + 2);
                        }
                    }
                }

                $currentState.State = $job.State;
            }
            else
            {
                if(($currentState.State -eq "Passed") -or ($currentState.State -eq "Failed"))
                {
                    $currentState.Ended = $true;
                }
            }
                
            if (!$currentState.Message)
            {
                $messageColor = [ConsoleColor]::White
                $currentState.Message = "---"
            }

            if ($currentState.State -eq "Completed")
            {
                $currentState.State = "Passed";
            }

            $currentState.Message = $currentState.Message -replace "`t|`n|`r", " "

            if ($currentState.Message.Length -gt $textWidth)
            {
                $currentState.Message = $currentState.Message.Substring(0, $textWidth - 5) + " ... "
            }

            if ($currentState.CursorPosition -eq $null)
            {
                # new item, move to end
                Set-CursorEnd

                # save cursor
                $currentState.CursorPosition = $host.UI.RawUI.CursorPosition;
            }
            else
            {
                # old item, move to saved place.
                $host.UI.RawUI.CursorPosition = $currentState.CursorPosition;
            }

            # OUTPUT Starts
            # Date
            Write-Host $currentState.StartTime.ToString('MM/dd HH:mm:ss') -f White -NoNewline
            if ($currentState.State -eq "Running")
            {
                $progressSign = $scroll[$currentState.ScrollIndex];
                Write-Host " $progressSign " -f Cyan -NoNewline
            }
            else
            {
                Write-Host " - " -f White -NoNewline
            }

            # Progress
            if ($currentState.State -eq "Queued")
            {
                Write-Host "     " -f White -NoNewline
            }
            else
            {
                $now = Get-Date
                $totalSeconds = ($now - $currentState.StartTime).TotalSeconds
                if ($totalSeconds -gt 86400)
                {
                    Write-Host (" > 1d") -NoNewline
                }
                elseif ($totalSeconds -ge 600)
                {
                    # only show m.
                    $totalSeconds /= 60
                    Write-Host ("{0,4:0}m" -f $totalSeconds) -NoNewline
                }
                elseif ($totalSeconds -ge 120)
                {
                    Write-Host ("{0,5}" -f ("{0:0}m{1:0}s" -f ([Math]::Floor($totalSeconds / 60)), ($totalSeconds % 60))) -NoNewline
                }
                else
                {
                    Write-Host ("{0,4:0}s" -f $totalSeconds) -NoNewline
                }
            }

            $currentState.ScrollIndex ++
            if ($currentState.ScrollIndex -ge $scroll.Length)
            {
                $currentState.ScrollIndex = 0
            }

            # Status
            Write-Host "[" -f White -NoNewline
            $color = [ConsoleColor]::Cyan
            if ($currentState.State -eq "Failed")
            {
                $color = [ConsoleColor]::Red
            }
            elseif ($currentState.State -eq "Running")
            {
                $color = [ConsoleColor]::Cyan
            }
            elseif ($currentState.State -eq "Passed")
            {
                $color = [ConsoleColor]::Green
            }

            $statusTextWidth = $statusWidth - 3
            Write-Host ("{0, -$statusTextWidth}" -f $currentState.State) -f $color -NoNewline
            Write-Host '] ' -f White -NoNewline

            # Name
            Write-Host ("{0, -$nameWidth}" -f $currentState.Name) -f Gray -NoNewline

            # Message
            Write-Host ("{0, -$textWidth}" -f $currentState.Message) -f $messageColor -NoNewline

            Write-Host

            Save-CursorEnd
        }

        $completed = ([Tracing]::states.GetEnumerator() | Where-Object { $_.Value.Ended } | measure).Count
        $total = [Tracing]::SubProgressCount
        $mainProgress = [Tracing]::MainProgress

        # old item, move to saved place.
        # "$host.UI.RawUI.CursorPosition = [Tracing]::CursorPosition" throws object unbox exception in Win10 Preview
        $savedCurP = [Tracing]::CursorPosition
        $host.UI.RawUI.CursorPosition = $savedCurP

        AppendDateTime $mainProgress | Write-Host -f White -NoNewline
        Write-Host " - ( $completed / $total )" -f Gray

        Set-CursorEnd
    } while (-not $exitLoop)
    
    # return the done jobs if the switch 'Any' is specified
    if($Any)
    {
        return $doneJobs    
    }
}

# public
# Register a main progress and set its sub progress count
# This will clear all registered progress if the previous main progress is not the same as the current.
# It will do nothing if the previous is the same as current.
function Start-Progress($mainProgress, $subProgressCount)
{
    Clear-Progress
    [Tracing]::MainProgress = $mainProgress
    $host.UI.RawUI.WindowTitle = $mainProgress
    Set-CursorEnd
    Write-Host

    [Tracing]::CursorPosition = $host.UI.RawUI.CursorPosition

    # Date
    AppendDateTime $mainProgress | Write-Host -f White -NoNewline
    Write-Host " - ( 0 / $subProgressCount )" -f Gray

    $width = $host.UI.RawUI.WindowSize.Width - 2
    Write-Host (New-Object System.String '-', $width) -f Gray

    [Tracing]::SubProgressCount = $subProgressCount

    Save-CursorEnd
}

# Public
# Report a sub progress with sub progress name, status, and message.
# If the status is reported as string, the progress will be displayed without the progress notification sign, and it
# won't cause the Wait-Progress to block.
# If the status is reported as Job, the progress will be displayed with a progress notification sign.
# and it will block the Wait-Progress call until the Job is completed or failed.
function Report-Progress
{
    [CmdletBinding(DefaultParametersetName="Jobs")] 
    Param
    (
        [Parameter(Position = 0, Mandatory = $true)]
        [System.String]$subProgress,

        [Parameter(Position = 1, Mandatory = $true, ParameterSetName = "Status")]
        [ValidateSet("Queued", "Failed", "Passed", "Running")]
        [System.String]$status,

        [Parameter(Position = 1, Mandatory = $true, ParameterSetName = "Job")]
        [System.Management.Automation.Job]$job,

        [Parameter(Position = 2, Mandatory = $false)]
        [System.String]$message
    )

    if ($message -eq $null)
    {
        $message = "---";
    }

    if ($PSCmdlet.ParameterSetName -eq "Job")
    {
        [Tracing]::ReportProgress($subProgress, $job, $message);
    }
    else
    {
        [Tracing]::ReportProgress($subProgress, $status, $message);
    }

    # false means not block, just output to the screen once.
    Wait-Progress -NoWait
}

# Public
# Write a string to standard warning stream 3>
function Warning($log)
{
    Set-CursorEnd
    AppendDateTime $log | Write-Warning
    Save-CursorEnd
}

# Public
# Write a string to standard Error stream 2>
function Error($log)
{
    Set-CursorEnd
    AppendDateTime $log | Write-Error
    Save-CursorEnd
}

# Public
# Write a string to standard verbose stream 4>
function Verbose($log)
{
    Set-CursorEnd
    AppendDateTime $log | Write-Verbose
    Save-CursorEnd
}

# Public
# Initialize before reporting progress.
function Initialize-Progress()
{
    $TraceSource = @"
    using System;
    using System.Collections.Generic;
    using System.Management.Automation.Host;
    using System.Management.Automation;

    public static class Tracing
    {
        public static string OrigConsoleTitle
        {
            get;
            set;
        }

        public static int SubProgressCount
        {
            get;
            set;
        }

        public static int MaxLine
        {
            get;
            set;
        }

        public static string MainProgress
        {
            get;
            set;
        }

        public static Coordinates? CursorPosition { get; set; }

        public class Progress
        {
            public DateTime StartTime { get; set; }
            public string Name { get; set; }
            public string Message { get; set; }
            public string State   { get; set; }
            public int ScrollIndex { get; set; }
            public Coordinates? CursorPosition { get; set; }
            public Job Job { get; set; }
            public bool Ended { get; set; }
        }

        public static Dictionary<string, Progress> states = new Dictionary<string, Progress>();
        public static void ReportProgress(string subProgress, Job job, string message)
        {
            Progress progress = null;
            if (!states.TryGetValue(subProgress, out progress))
            {
                progress = new Progress ()
                {
                    Name = subProgress, 
                    Message = message,
                    Job = job,
                    StartTime = DateTime.Now,
                    ScrollIndex = 0,
                };

                states[subProgress] = progress;
            }
            else
            {
                // If the previous state is Queued, reset the StartTime
                if(progress.State.Equals("Queued", StringComparison.OrdinalIgnoreCase))
                {
                    progress.StartTime = DateTime.Now;
                }
            }

            progress.Job = job;
            progress.Ended = false;
            progress.Message = message;
        }

        public static void ReportProgress(string subProgress, string state, string message)
        {
            Progress progress = null;
            if (!states.TryGetValue(subProgress, out progress))
            {
                progress = new Progress ()
                {
                    Name = subProgress,
                    Message = message,
                    State = state,
                    StartTime = DateTime.Now,
                    ScrollIndex = 0,
                };

                states[subProgress] = progress;
            }
            else
            {
                // If the previous state is Queued, reset the StartTime
                if(progress.State.Equals("Queued", StringComparison.OrdinalIgnoreCase))
                {
                    progress.StartTime = DateTime.Now;
                }
            }

            progress.Ended = false;
            progress.State = state;
            progress.Message = message;
        }
    }
"@

    Add-Type -TypeDefinition $TraceSource
    Clear-Progress
    Save-CursorEnd

    # Save the original Title
    [Tracing]::OrigConsoleTitle = $host.UI.RawUI.WindowTitle

    $preferWidth = 160
    $preferHeight = 60
    $bufferSize = $host.UI.RawUI.BufferSize
    # Set the buffer width to a large value to test whether the maximum supported window width can be 160
    $bufferSize.Width = $preferWidth + 15 
    $host.UI.RawUI.BufferSize = $bufferSize
    $maxWinSize = $host.UI.RawUI.Get_MaxWindowSize()

    $width = $maxWinSize.Width - 10
    $height = $maxWinSize.Height - 4
    if($width -gt $preferWidth)
    {
        $width = $preferWidth
    }
    if($height -gt $preferHeight )
    {
        $height = $preferHeight 
    }

    $bufferSize.Width = $width

    # Set the buffer height to the maximum value to minimize the probability of window scroll.
    $bufferSize.Height = 9999
    $host.UI.RawUI.BufferSize = $bufferSize
    
    try
    {
        # Clear the buffer, and reset the saved [Tracing]::MaxLine to let the next line start from the beginning.
        # The .Net call is faster according to some post, so don't use cls or Clear-Host
        if($host.UI.RawUI.BufferSize.Height -lt ($host.UI.RawUI.CursorPosition.Y + 200))
        {
            [System.Console]::Clear()
            [Tracing]::MaxLine = $host.UI.RawUI.CursorPosition.Y
        }
    }
    catch
    {
        # Ignore the exception
    }

    $windowSize = $bufferSize
    $windowSize.Height = $height
    $host.UI.RawUI.WindowSize = $windowSize
}

# Public 
# Un-initialize after done.
function Uninitialize-Progress()
{
    Clear-Progress
    [Tracing]::MainProgress = $null
    $host.UI.RawUI.WindowTitle = [Tracing]::OrigConsoleTitle
    Set-CursorEnd
}

# SIG # Begin signature block
# MIIavwYJKoZIhvcNAQcCoIIasDCCGqwCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUKbrL5Vcf9KSSp+VpaejF7NVT
# p8SgghV6MIIEuzCCA6OgAwIBAgITMwAAAFwJq3ADEfxcFQAAAAAAXDANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTQwNTIzMTcxMzE2
# WhcNMTUwODIzMTcxMzE2WjCBqzELMAkGA1UEBhMCVVMxCzAJBgNVBAgTAldBMRAw
# DgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24x
# DTALBgNVBAsTBE1PUFIxJzAlBgNVBAsTHm5DaXBoZXIgRFNFIEVTTjpCQkVDLTMw
# Q0EtMkRCRTElMCMGA1UEAxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZTCC
# ASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBANYJa/7h+ToIHrYOs0PLI95O
# bOJZcc75TxmjL8QLy/H1Xyl5ielMpmU8xUtMXQHp/i9QvCCDC5hHW5nqTZ/eqcsj
# lYzk07XcemMgs/L/r8dp/5K7jQpF7punPfdkmj6O6rRaABbNqeXVTJUmQH8DmFyk
# 6dBH1jblvqVRQb9b9uyuNN9K8gXa/fw9YY6yEa11unZrY2JOJSHPpz88ub9uBEMk
# HrvplbnGydEgydhVI1Xnsr+vLVxathcwcTMQJH5xnrFl4ma2BdMyV0nEYhPYkoyB
# kY8zQ5EZbBL0eRTm5Z/F5anhM0C22y9X05U2It3yQqmUtGTTgSMUlMPZRhdm3r0C
# AwEAAaOCAQkwggEFMB0GA1UdDgQWBBR/kUhDg4fJeBzpVcW521b4F9FHBDAfBgNV
# HSMEGDAWgBQjNPjZUkZwCu1A+3b7syuwwzWzDzBUBgNVHR8ETTBLMEmgR6BFhkNo
# dHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNyb3Nv
# ZnRUaW1lU3RhbXBQQ0EuY3JsMFgGCCsGAQUFBwEBBEwwSjBIBggrBgEFBQcwAoY8
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNyb3NvZnRUaW1l
# U3RhbXBQQ0EuY3J0MBMGA1UdJQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEBBQUA
# A4IBAQCCCVb4b+5GzEwp9r7Tx2pf+V2EABkXgYloZBU/wUQ2OxhtBMusE7eOlOMt
# 66P0QnxE0QnsWr7nk9OEcuKP1R3VOVP7ciILZpI7ysqr+s/MZE7GRGyv4IOOFBG4
# sJBJn8OmlM+D4BFiWqtYo7hhlyR2rG1D947qHfOI/ipL7h/0HW2sIEyIAEAutLlI
# 2tHEbsGt9DOFYQuJLF5rygNjUkox7r1VwxCnUaiSrxJgxZxdRmPTzOqfNcatCSB0
# hTUbRRKzllHhye2bgOV25fdWhesgnCiRVPtoulaFg1Vh/4hAXLENJgejxVfsFXkP
# lvSOdqGKZ1qNf05i9H7fEKDWFLqXMIIE7DCCA9SgAwIBAgITMwAAAMps1TISNcTh
# VQABAAAAyjANBgkqhkiG9w0BAQUFADB5MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSMwIQYDVQQDExpNaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QTAeFw0xNDA0MjIxNzM5MDBaFw0xNTA3MjIxNzM5MDBaMIGDMQswCQYDVQQGEwJV
# UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
# ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMQ0wCwYDVQQLEwRNT1BSMR4wHAYDVQQD
# ExVNaWNyb3NvZnQgQ29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAw
# ggEKAoIBAQCWcV3tBkb6hMudW7dGx7DhtBE5A62xFXNgnOuntm4aPD//ZeM08aal
# IV5WmWxY5JKhClzC09xSLwxlmiBhQFMxnGyPIX26+f4TUFJglTpbuVildGFBqZTg
# rSZOTKGXcEknXnxnyk8ecYRGvB1LtuIPxcYnyQfmegqlFwAZTHBFOC2BtFCqxWfR
# +nm8xcyhcpv0JTSY+FTfEjk4Ei+ka6Wafsdi0dzP7T00+LnfNTC67HkyqeGprFVN
# TH9MVsMTC3bxB/nMR6z7iNVSpR4o+j0tz8+EmIZxZRHPhckJRIbhb+ex/KxARKWp
# iyM/gkmd1ZZZUBNZGHP/QwytK9R/MEBnAgMBAAGjggFgMIIBXDATBgNVHSUEDDAK
# BggrBgEFBQcDAzAdBgNVHQ4EFgQUH17iXVCNVoa+SjzPBOinh7XLv4MwUQYDVR0R
# BEowSKRGMEQxDTALBgNVBAsTBE1PUFIxMzAxBgNVBAUTKjMxNTk1K2I0MjE4ZjEz
# LTZmY2EtNDkwZi05YzQ3LTNmYzU1N2RmYzQ0MDAfBgNVHSMEGDAWgBTLEejK0rQW
# WAHJNy4zFha5TJoKHzBWBgNVHR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1pY3Jv
# c29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNDb2RTaWdQQ0FfMDgtMzEtMjAx
# MC5jcmwwWgYIKwYBBQUHAQEETjBMMEoGCCsGAQUFBzAChj5odHRwOi8vd3d3Lm1p
# Y3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY0NvZFNpZ1BDQV8wOC0zMS0yMDEwLmNy
# dDANBgkqhkiG9w0BAQUFAAOCAQEAd1zr15E9zb17g9mFqbBDnXN8F8kP7Tbbx7Us
# G177VAU6g3FAgQmit3EmXtZ9tmw7yapfXQMYKh0nfgfpxWUftc8Nt1THKDhaiOd7
# wRm2VjK64szLk9uvbg9dRPXUsO8b1U7Brw7vIJvy4f4nXejF/2H2GdIoCiKd381w
# gp4YctgjzHosQ+7/6sDg5h2qnpczAFJvB7jTiGzepAY1p8JThmURdwmPNVm52Iao
# AP74MX0s9IwFncDB1XdybOlNWSaD8cKyiFeTNQB8UCu8Wfz+HCk4gtPeUpdFKRhO
# lludul8bo/EnUOoHlehtNA04V9w3KDWVOjic1O1qhV0OIhFeezCCBbwwggOkoAMC
# AQICCmEzJhoAAAAAADEwDQYJKoZIhvcNAQEFBQAwXzETMBEGCgmSJomT8ixkARkW
# A2NvbTEZMBcGCgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9z
# b2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5MB4XDTEwMDgzMTIyMTkzMloX
# DTIwMDgzMTIyMjkzMloweTELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0
# b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3Jh
# dGlvbjEjMCEGA1UEAxMaTWljcm9zb2Z0IENvZGUgU2lnbmluZyBQQ0EwggEiMA0G
# CSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCycllcGTBkvx2aYCAgQpl2U2w+G9Zv
# zMvx6mv+lxYQ4N86dIMaty+gMuz/3sJCTiPVcgDbNVcKicquIEn08GisTUuNpb15
# S3GbRwfa/SXfnXWIz6pzRH/XgdvzvfI2pMlcRdyvrT3gKGiXGqelcnNW8ReU5P01
# lHKg1nZfHndFg4U4FtBzWwW6Z1KNpbJpL9oZC/6SdCnidi9U3RQwWfjSjWL9y8lf
# RjFQuScT5EAwz3IpECgixzdOPaAyPZDNoTgGhVxOVoIoKgUyt0vXT2Pn0i1i8UU9
# 56wIAPZGoZ7RW4wmU+h6qkryRs83PDietHdcpReejcsRj1Y8wawJXwPTAgMBAAGj
# ggFeMIIBWjAPBgNVHRMBAf8EBTADAQH/MB0GA1UdDgQWBBTLEejK0rQWWAHJNy4z
# Fha5TJoKHzALBgNVHQ8EBAMCAYYwEgYJKwYBBAGCNxUBBAUCAwEAATAjBgkrBgEE
# AYI3FQIEFgQU/dExTtMmipXhmGA7qDFvpjy82C0wGQYJKwYBBAGCNxQCBAweCgBT
# AHUAYgBDAEEwHwYDVR0jBBgwFoAUDqyCYEBWJ5flJRP8KuEKU5VZ5KQwUAYDVR0f
# BEkwRzBFoEOgQYY/aHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJv
# ZHVjdHMvbWljcm9zb2Z0cm9vdGNlcnQuY3JsMFQGCCsGAQUFBwEBBEgwRjBEBggr
# BgEFBQcwAoY4aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNy
# b3NvZnRSb290Q2VydC5jcnQwDQYJKoZIhvcNAQEFBQADggIBAFk5Pn8mRq/rb0Cx
# MrVq6w4vbqhJ9+tfde1MOy3XQ60L/svpLTGjI8x8UJiAIV2sPS9MuqKoVpzjcLu4
# tPh5tUly9z7qQX/K4QwXaculnCAt+gtQxFbNLeNK0rxw56gNogOlVuC4iktX8pVC
# nPHz7+7jhh80PLhWmvBTI4UqpIIck+KUBx3y4k74jKHK6BOlkU7IG9KPcpUqcW2b
# Gvgc8FPWZ8wi/1wdzaKMvSeyeWNWRKJRzfnpo1hW3ZsCRUQvX/TartSCMm78pJUT
# 5Otp56miLL7IKxAOZY6Z2/Wi+hImCWU4lPF6H0q70eFW6NB4lhhcyTUWX92THUmO
# Lb6tNEQc7hAVGgBd3TVbIc6YxwnuhQ6MT20OE049fClInHLR82zKwexwo1eSV32U
# jaAbSANa98+jZwp0pTbtLS8XyOZyNxL0b7E8Z4L5UrKNMxZlHg6K3RDeZPRvzkbU
# 0xfpecQEtNP7LN8fip6sCvsTJ0Ct5PnhqX9GuwdgR2VgQE6wQuxO7bN2edgKNAlt
# HIAxH+IOVN3lofvlRxCtZJj/UBYufL8FIXrilUEnacOTj5XJjdibIa4NXJzwoq6G
# aIMMai27dmsAHZat8hZ79haDJLmIz2qoRzEvmtzjcT3XAH5iR9HOiMm4GPoOco3B
# oz2vAkBq/2mbluIQqBC0N1AI1sM9MIIGBzCCA++gAwIBAgIKYRZoNAAAAAAAHDAN
# BgkqhkiG9w0BAQUFADBfMRMwEQYKCZImiZPyLGQBGRYDY29tMRkwFwYKCZImiZPy
# LGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQDEyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZp
# Y2F0ZSBBdXRob3JpdHkwHhcNMDcwNDAzMTI1MzA5WhcNMjEwNDAzMTMwMzA5WjB3
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEwHwYDVQQDExhN
# aWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAw
# ggEKAoIBAQCfoWyx39tIkip8ay4Z4b3i48WZUSNQrc7dGE4kD+7Rp9FMrXQwIBHr
# B9VUlRVJlBtCkq6YXDAm2gBr6Hu97IkHD/cOBJjwicwfyzMkh53y9GccLPx754gd
# 6udOo6HBI1PKjfpFzwnQXq/QsEIEovmmbJNn1yjcRlOwhtDlKEYuJ6yGT1VSDOQD
# LPtqkJAwbofzWTCd+n7Wl7PoIZd++NIT8wi3U21StEWQn0gASkdmEScpZqiX5NMG
# gUqi+YSnEUcUCYKfhO1VeP4Bmh1QCIUAEDBG7bfeI0a7xC1Un68eeEExd8yb3zuD
# k6FhArUdDbH895uyAc4iS1T/+QXDwiALAgMBAAGjggGrMIIBpzAPBgNVHRMBAf8E
# BTADAQH/MB0GA1UdDgQWBBQjNPjZUkZwCu1A+3b7syuwwzWzDzALBgNVHQ8EBAMC
# AYYwEAYJKwYBBAGCNxUBBAMCAQAwgZgGA1UdIwSBkDCBjYAUDqyCYEBWJ5flJRP8
# KuEKU5VZ5KShY6RhMF8xEzARBgoJkiaJk/IsZAEZFgNjb20xGTAXBgoJkiaJk/Is
# ZAEZFgltaWNyb3NvZnQxLTArBgNVBAMTJE1pY3Jvc29mdCBSb290IENlcnRpZmlj
# YXRlIEF1dGhvcml0eYIQea0WoUqgpa1Mc1j0BxMuZTBQBgNVHR8ESTBHMEWgQ6BB
# hj9odHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9taWNy
# b3NvZnRyb290Y2VydC5jcmwwVAYIKwYBBQUHAQEESDBGMEQGCCsGAQUFBzAChjho
# dHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY3Jvc29mdFJvb3RD
# ZXJ0LmNydDATBgNVHSUEDDAKBggrBgEFBQcDCDANBgkqhkiG9w0BAQUFAAOCAgEA
# EJeKw1wDRDbd6bStd9vOeVFNAbEudHFbbQwTq86+e4+4LtQSooxtYrhXAstOIBNQ
# md16QOJXu69YmhzhHQGGrLt48ovQ7DsB7uK+jwoFyI1I4vBTFd1Pq5Lk541q1YDB
# 5pTyBi+FA+mRKiQicPv2/OR4mS4N9wficLwYTp2OawpylbihOZxnLcVRDupiXD8W
# mIsgP+IHGjL5zDFKdjE9K3ILyOpwPf+FChPfwgphjvDXuBfrTot/xTUrXqO/67x9
# C0J71FNyIe4wyrt4ZVxbARcKFA7S2hSY9Ty5ZlizLS/n+YWGzFFW6J1wlGysOUzU
# 9nm/qhh6YinvopspNAZ3GmLJPR5tH4LwC8csu89Ds+X57H2146SodDW4TsVxIxIm
# dgs8UoxxWkZDFLyzs7BNZ8ifQv+AeSGAnhUwZuhCEl4ayJ4iIdBD6Svpu/RIzCzU
# 2DKATCYqSCRfWupW76bemZ3KOm+9gSd0BhHudiG/m4LBJ1S2sWo9iaF2YbRuoROm
# v6pH8BJv/YoybLL+31HIjCPJZr2dHYcSZAI9La9Zj7jkIeW1sMpjtHhUBdRBLlCs
# lLCleKuzoJZ1GtmShxN1Ii8yqAhuoFuMJb+g74TKIdbrHk/Jmu5J4PcBZW+JC33I
# acjmbuqnl84xKf8OxVtc2E0bodj6L54/LlUWa8kTo/0xggSvMIIEqwIBATCBkDB5
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSMwIQYDVQQDExpN
# aWNyb3NvZnQgQ29kZSBTaWduaW5nIFBDQQITMwAAAMps1TISNcThVQABAAAAyjAJ
# BgUrDgMCGgUAoIHIMBkGCSqGSIb3DQEJAzEMBgorBgEEAYI3AgEEMBwGCisGAQQB
# gjcCAQsxDjAMBgorBgEEAYI3AgEVMCMGCSqGSIb3DQEJBDEWBBS4gm11D5k2GwGT
# idsNCtA6tPghVzBoBgorBgEEAYI3AgEMMVowWKA2gDQATQBpAGMAcgBvAHMAbwBm
# AHQAIABIAFAAQwAgAFAAYQBjAGsAIAAyADAAMQAyACAAUgAyoR6AHGh0dHA6Ly93
# d3cubWljcm9zb2Z0LmNvbS9IUEMwDQYJKoZIhvcNAQEBBQAEggEAc+CS4mL6/XMC
# +MRF/MCjraLjjO+QjutkiDCM8XGIUTDf8Rx2VV8Rd8EX3z+dd8jV1Ran/y4o27LM
# 1XeQkVqthcGaE74IOpDRoPWMqsNpgcusAaBNRwGAp/SNNRP4AWWkbi8mYquCJPMX
# 83+0owaY/34WIvn8eV8dwN+E5Or1+406oenI87+Nh8Caxik8nwYaK0OZyYPXTMnR
# OXDWeJGamScg7uEgOtxHGuNE7Ib7APIJZ5AUtiWOpdQkqFTu1SpZvol5WbvC+zzR
# 4igqD4RmtEUVj/BzluKVMKzHhpSrJNtGxVaYBnMnalrF4R6K3AxeFVpmeTE+eWzQ
# JmEsKcXCZKGCAigwggIkBgkqhkiG9w0BCQYxggIVMIICEQIBATCBjjB3MQswCQYD
# VQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEe
# MBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEwHwYDVQQDExhNaWNyb3Nv
# ZnQgVGltZS1TdGFtcCBQQ0ECEzMAAABcCatwAxH8XBUAAAAAAFwwCQYFKw4DAhoF
# AKBdMBgGCSqGSIb3DQEJAzELBgkqhkiG9w0BBwEwHAYJKoZIhvcNAQkFMQ8XDTE1
# MDIxMjE4MTQ1M1owIwYJKoZIhvcNAQkEMRYEFNGxIIyhXWamxRk+XHYuSY4ebcOF
# MA0GCSqGSIb3DQEBBQUABIIBAFWIfQzp6+bZlHlNgMzSjRqbtzchXP5++RYpmA1r
# olUti7p36R0k5r5mPbvR9JwlWRsXWaq5nMn0uFt8QL5JzatwmzfnnXbuwInTJpdP
# ABqLCyJAzUl5jtysEaJkJbxZH9BpYaMTdgPhhke5rXPzPUdSOYK3LrKrX4XN8IIr
# djVa9WMZALVEghbo/xrF1sX2UW02BPHY6EkvfGLDt7foVbSsLq/1yP5tALquHjdT
# BnVKNYtDESQJB+tlkSfcXwLNCGm7PCNvlpau10PEvOaA+C7COLTsHwUAc3t0MBKf
# pwkfJb2Dj+hd86l+EPqOyA5nGZ+ZD1sy/MzwK6K+Bbm5udE=
# SIG # End signature block
