﻿<#
.SYNOPSIS
   Removes a Microsoft HPC Pack cluster from Microsoft Azure
.DESCRIPTION
   This script removes a Microsoft HPC cluster from Microsoft Azure

   Author :  Microsoft HPC Pack team
   Version:  4.3.1
.EXAMPLE
   .\Remove-HPCIaasCluster.ps1 -ConfigFile ".\HPCIaaSCluster.xml"
.EXAMPLE
   .\Remove-HPCIaasCluster.ps1 -ConfigFile ".\HPCIaaSCluster.xml" -KeepImage
#>
[CmdletBinding(SupportsShouldProcess = $true, ConfirmImpact = "High")]
Param
(
    [Parameter(Mandatory=$true)]
    [String] $ConfigFile,

    [Parameter(Mandatory=$false)]
    [Switch] $KeepImage,

    [Parameter(Mandatory=$false)]
    [Switch] $KeepVNet,

    [Parameter(Mandatory=$false)]
    [Switch] $KeepAffinityGroup,

    [Parameter(Mandatory=$false)]
    [Switch] $KeepStorageAccount,

    [Parameter(Mandatory=$false)]
    [Switch] $Force,

    [Parameter(Mandatory=$false)]
    [ValidateRange(0, 900)]
    [Int] $SecondsWaitingForDiskDeletion = 180
)

$removeImage = -not $KeepImage.IsPresent
$removeVNet = -not $KeepVNet.IsPresent
$removeAG = -not $KeepAffinityGroup.IsPresent
$removeStoreAcc = -not $KeepStorageAccount.IsPresent

"Checking config file"
$clusterConfig = [xml](Get-Content $ConfigFile -ErrorAction Stop)
$iaasConfig = $clusterConfig.IaaSClusterConfig

$subName = $iaasConfig.Subscription.SubscriptionName
$storeAcc = $iaasConfig.Subscription.StorageAccount
$AGName = $iaasConfig.AffinityGroup
$VNetName = $iaasConfig.VNet.VNetName

$roleList = @()
$DNSName = $null

$newDC = $iaasConfig.Domain.DCOption -eq 'NewDC'
if ($newDC)
{
    # used to remove the DNS server setting
    $DNSName = $iaasConfig.Domain.DomainController.VMName

    $DCSvc = $iaasConfig.Domain.DomainController.ServiceName
    $DCAG = $iaasConfig.Domain.DomainController.AffinityGroup
    $roleList += @{ RoleName = "Domain Controller"; Svcs = @($DCSvc); AG = $DCAG }
}

$newDB = $iaasConfig.Database.DBOption -eq 'NewRemoteDB'
if ($newDB)
{
    $DBSvc = $iaasConfig.Database.DBServer.ServiceName
    $DBAG = $iaasConfig.Database.DBServer.AffinityGroup
    $roleList += @{ RoleName = "DB Server"; Svcs = @($DBSvc); AG = $DBAG }
}

$HNSvc = $iaasConfig.HeadNode.ServiceName
$HNAG = $iaasConfig.HeadNode.AffinityGroup
$roleList += @{ RoleName = "Head Node"; Svcs = @($HNSvc); AG = $HNAG }

$hasCN = $iaasConfig.ComputeNodes -ne $null -and $clusterConfig.IaaSClusterConfig.ComputeNodes.NodeCount -ne 0
if ($hasCN)
{
    $CNAG = $iaasConfig.ComputeNodes.AffinityGroup

    $CNSvc = $iaasConfig.ComputeNodes.ServiceName
    if ($CNSvc -ne $null)
    {
        # single service
        $roleList += @{ RoleName = "Compute Node"; Svcs = @($CNSvc); AG = $CNAG }
    }
    else
    {
        # service pattern
        $count = [Math]::Ceiling($iaasConfig.ComputeNodes.NodeCount / $iaasConfig.ComputeNodes.MaxNodeCountPerService)

        $svcPattern = $iaasConfig.ComputeNodes.ServiceNamePattern.Trim(" %").Split("%") | ? { -not [String]::IsNullOrWhiteSpace($_) }
        $svcRoot = $svcPattern[0]
        $svcStartsFrom = [Int]::Parse($svcPattern[1])
        $svcDigitWidth = $svcPattern[1].Length

        $storePattern = $iaasConfig.ComputeNodes.StorageAccountNamePattern.Trim(" %").Split("%") | ? { -not [String]::IsNullOrWhiteSpace($_) }
        $storeRoot = $storePattern[0]
        $storeStartsFrom = [Int]::Parse($storePattern[1])
        $storeDigitWidth = $storePattern[1].Length

        $CNSvcs = @()
        $CNStores = @()

        for ($i = 0; $i -lt $count; $i++)
        {
            $CNSvcs += "$svcRoot$(($svcStartsFrom + $i).ToString("D$svcDigitWidth"))"
            $CNStores += "$storeRoot$(($storeStartsFrom + $i).ToString("D$storeDigitWidth"))"
        }

        $roleList += @{ RoleName = "Compute Node"; Svcs = $CNSvcs; Stores = $CNStores; AG = $CNAG }
    }
}

$hasBN = $iaasConfig.BrokerNodes -ne $null -and $clusterConfig.IaaSClusterConfig.BrokerNodes.NodeCount -ne 0
if ($hasBN)
{
    $BNSvc = $iaasConfig.BrokerNodes.ServiceName
    $BNAG = $iaasConfig.BrokerNodes.AffinityGroup
    $roleList += @{ RoleName = "Broker Node"; Svcs = @($BNSvc); AG = $BNAG }
}

"Scanning Azure deployment"
Select-AzureSubscription -SubscriptionName $subName | Out-Null

$listAGs = @()
$listSvcs = @()
$listImgs = @()
$listStores = @()
$VNetExisting = $false
$DNSExisting = $false

$createdSvcs = @(Get-AzureService | % { $_.ServiceName })
$createdStores = $null
if ($removeStoreAcc) { $createdStores = @(Get-AzureStorageAccount -WarningAction SilentlyContinue | % { $_.StorageAccountName }) }
$createdAGs = $null
if ($removeAG) { $createdAGs = @(Get-AzureAffinityGroup | % { $_.Name }) }

foreach ($role in $roleList)
{
    foreach ($svcName in $role.Svcs)
    {
        if ($svcName -notin $createdSvcs)
        {
            "$($role.RoleName)'s Cloud Service $svcName does not exist"
        }
        else
        {
            if ($listSvcs -notcontains $svcName)
            {
                $listSvcs += $svcName
            }
        }
    }

    if ($removeStoreAcc)
    {
        foreach ($storeName in $role.Stores)
        {
            if ($storeName -notin $createdStores)
            {
                "$($role.RoleName)'s Storage Account $storeName does not exist"
            }
            else
            {
                if ($listStores -notcontains $storeName)
                {
                    $listStores += $storeName
                }
            }
        }
    }

    if ($removeAG -and -not [string]::IsNullOrEmpty($role.AG))
    {
        if ($role.AG -notin $createdAGs)
        {
            "$($role.RoleName)'s Affinity Group $($role.AG) does not exist"
        }
        else
        {
            if ($listAGs -notcontains $role.AG)
            {
                $listAGs += $role.AG
            }
        }
    }
}

if ($removeImage)
{
    $imgs = Get-AzureVMImage
    if($hasCN)
    {
        $svcList = @()
        if ($CNSvc -ne $null)
        {
            $svcList += $CNSvc
        }
        else
        {
            $svcList += $CNSvcs
        }

        $rootImages = @($imgs | ? { $_.ImageName -like "HPC-Pack-ComputeNode_*" -and $_.ServiceName -in $svcList })

        if ($rootImages.Count -eq 0)
        {
            "Compute Node Image does not exist"
        }
        else
        {
            $svcsHostingMultiImages = @($rootImages | % { $_.ServiceName } | Group-Object | ? { $_.Count -gt 1 })
            if ($svcsHostingMultiImages.Count -gt 0)
            {
                Write-Warning "More than one Compute Node Images captured from the same Cloud Service are found. (e.g. from $($svcsHostingMultiImages[0].Name))"
                Write-Warning "Skipping Compute Node Image"
            }
            else
            {
                $imgNames = @($rootImages | % { $_.ImageName })
                foreach ($name in $imgNames)
                {
                    $listImgs += $imgs | % { $_.ImageName } | ? { $_.StartsWith($name) }
                }
            }
        }
    }

    if($hasBN)
    {
        $BNImage = @($imgs | ? { $_.ImageName -like "HPC-Pack-BrokerNode_*" -and $_.ServiceName -eq $BNSvc })
        if ($BNImage.Count -gt 1)
        {
            Write-Warning "More than one Broker Node Images are found. Skipping Broker Node Image"
        }
        elseif ($BNImage.Count -eq 0)
        {
            "Broker Node Image does not exist"
        }
        else
        {
            $listImgs += $BNImage[0].ImageName
        }
    }
}

if ($removeStoreAcc)
{
    if ($storeAcc -notin $createdStores)
    {
        "Config-wide Storage Account $storeAcc does not exist"
    }
    else
    {
        if ($listStores -notcontains $storeAcc)
        {
            $listStores += $storeAcc
        }
    }
}

if ($removeAG -and -not [string]::IsNullOrEmpty($AGName))
{
    if ($AGName -notin $createdAGs)
    {
        "Config-wide Affinity Group $AGName does not exist"
    }
    else
    {
        if ($listAGs -notcontains $AGName)
        {
            $listAGs += $AGName
        }
    }
}

if ($removeVNet -or $DNSName -ne $null)
{
    $VNConfig = [xml](Get-AzureVNetConfig).XMLConfiguration
}
if ($removeVNet)
{
    $sites = $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.VirtualNetworkSites
    if ($sites -ne $null)
    {
        $targetSite = $sites.ChildNodes | ? { $_.Name -eq $VNetName }
        if ($targetSite -eq $null)
        {
            "Virtual Network $VNetName does not exist"
        }
        else
        {
            $VNetExisting = $true
        }
    }
}

if ($DNSName -ne $null)
{
    $dnsServers = $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.Dns.DnsServers
    if ($dnsServers -ne $null)
    {
        $targetDns = $dnsServers.ChildNodes | ? { $_.Name -eq $DNSName }
        if ($targetDns -eq $null)
        {
            "DNS Server setting for $DNSName does not exist"
        }
        else
        {
            $ns = New-Object System.Xml.XmlNamespaceManager($VNConfig.NameTable)
            $ns.AddNamespace("ns", $VNConfig.NetworkConfiguration.NamespaceURI)
            $sitesUsingThisDNS = @($VNConfig.SelectNodes("//ns:NetworkConfiguration/ns:VirtualNetworkConfiguration/ns:VirtualNetworkSites/ns:VirtualNetworkSite[ns:DnsServersRef/ns:DnsServerRef/@name='$DNSName']", $ns)) | % { $_.name }
            $okToRemoveDNS = $true
            foreach ($siteName in $sitesUsingThisDNS)
            {
                if ($siteName -ne $VNetName)
                {
                    $okToRemoveDNS = $false
                    Write-Warning "DNS Server $DNSName (Domain Controller of this cluster) is referenced by non-HPC VNet(s). (e.g. $siteName)"
                    Write-Warning "Skipping DNS Server setting"
                    break;
                }
            }
            if ($okToRemoveDNS)
            {
                $DNSExisting = $true
            }
        }
    }
}

if (-not $VNetExisting -and -not $DNSExisting -and $listStores.Count -eq 0 -and $listImgs.Count -eq 0 -and $listAGs.Count -eq 0 -and $listSvcs.Count -eq 0)
{
    "All the resources have already been removed"
    return
}

"The following Azure resource(s) will be removed:"
if ($listSvcs.Count -gt 0)
{
    "Cloud Service: $($listSvcs -join ',')"
}
if ($listImgs.Count -gt 0)
{
    "Image: $($listImgs -join ',')"
}
if ($listAGs.Count -gt 0)
{
    "Affinity Group: $($listAGs -join ',')"
}
if ($listStores.Count -gt 0)
{
    "Storage Account: $($listStores -join ',')"
}
if ($VNetExisting)
{
    "Virtual Network: $VNetName"
}
if ($DNSExisting)
{
    "DNS Server Setting: $DNSName"
}

if (-not $Force.IsPresent)
{
    if ($PSCmdlet.ShouldProcess("The HPC Cluster will be removed from Azure", "Continue?", "Confirm:") -eq $false)
    {
        "Remove operation aborted"
        return
    }
}

if ($listSvcs.Count -gt 0)
{
    "Removing Cloud Services"
    $listSvcs | % { Remove-AzureService -ServiceName $_ -Force -DeleteAll -ErrorAction Stop | Out-Null }

    "Checking Cloud Services Removed"
    $timeout = 900
    $beginTime = [DateTime]::Now
    while ($true)
    {
        if (([DateTime]::Now - $beginTime).TotalSeconds -ge $timeout)
        {
            throw "Failed to remove Cloud Services"
        }
    
        $remainingSvcs = @(Get-AzureService | % { $_.ServiceName })

        Start-Sleep -Seconds 15
        $listSvcs | % { if ($svc -in $remainingSvcs) { continue } }

        break
    }
}

if ($listImgs.Count -gt 0)
{
    "Removing VM Images"
    $listImgs | % { Remove-AzureVMImage -ImageName $_ -DeleteVHD -ErrorAction Stop | Out-Null }
}

if (($listSvcs.Count -gt 0 -or $listImgs.Count -gt 0) -and $SecondsWaitingForDiskDeletion -gt 0)
{
    "Waiting $SecondsWaitingForDiskDeletion seconds for Azure to delete the Services/Images' associated vhds"
    Start-Sleep -Seconds $SecondsWaitingForDiskDeletion
}

"Checking Azure Disks"
$retryLimit = 5
$retryCount = 0

while ($retryCount -lt $retryLimit)
{
    $allDisks = Get-AzureDisk
    $listDisks = @()

    if ($newDC)
    {
        $DCDisks = @($allDisks | ? { $_.DiskName.StartsWith("$DCSvc-") })
        if ($DCDisks.Count -gt 0)
        {
            $listDisks += $DCDisks
        }
    }

    if ($newDB)
    {
        $DBDisks = @($allDisks | ? { $_.DiskName.StartsWith("$DBSvc-") })
        if ($DBDisks.Count -gt 0)
        {
            $listDisks += $DBDisks
        }
    }

    $HNDisks = @($allDisks | ? { $_.DiskName.StartsWith("$HNSvc-") })
    if ($HNDisks.Count -gt 0)
    {
        $listDisks += $HNDisks
    }

    if ($hasCN)
    {
        foreach ($svc in $svcList)
        {
            $CNDisks = @($allDisks | ? { $_.DiskName.StartsWith("$svc-") })
            if ($CNDisks.Count -gt 0)
            {
                $listDisks += $CNDisks
            }
        }
    }

    if ($hasBN)
    {
        $BNDisks = @($allDisks | ? { $_.DiskName.StartsWith("$BNSvc-") })
        if ($BNDisks.Count -gt 0)
        {
            $listDisks += $BNDisks
        }
    }
    
    if ($listDisks.Count -eq 0)
    {
        break
    }

    Start-Sleep -Seconds 30
    $retryCount++
}

if ($listDisks.Count -gt 0)
{
    "Removing Azure Disks"
    $listDisks | Remove-AzureDisk -DeleteVHD -ErrorAction Stop | Out-Null
    "Waiting $SecondsWaitingForDiskDeletion seconds for Azure to delete the remaining VHDs detected in the last step"
    Start-Sleep -Seconds $SecondsWaitingForDiskDeletion
}

$storesFailedToRemove = @()

if ($listStores.Count -gt 0)
{
    "Removing Storage Account"
    foreach ($store in $listStores)
    {
        Remove-AzureStorageAccount -StorageAccountName $store | Out-Null
        if (-not $?)
        {
            $storesFailedToRemove += $store
        }
    }
}

$agsFailedToRemove = @()

if ($listAGs.Count -gt 0)
{
    "Removing Affinity Groups"
    foreach ($agToRemove in $listAGs)
    {
        Remove-AzureAffinityGroup -Name $agToRemove | Out-Null
        if (-not $?)
        {
            $agsFailedToRemove += $agToRemove
        }
    }
}

$netCfgUpdated = $true

if ($DNSExisting -or $VNetExisting)
{
    "Retrieving Azure Virtual Network Configuration"
    $VNConfig = [xml](Get-AzureVNetConfig).XMLConfiguration

    if ($DNSExisting)
    {
        "Removing DNS Server settings from the configuration"
        $dnsServers = $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.Dns.DnsServers
        $targetDns = $dnsServers.ChildNodes | ? { $_.Name -eq $DNSName }
        $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.Dns.DnsServers.RemoveChild($targetDns) | Out-Null
    }

    if ($VNetExisting)
    {
        "Removing Virtual Network from the configuration"
        $sites = $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.VirtualNetworkSites
        $targetSite = $sites.ChildNodes | ? { $_.Name -eq $VNetName }
        $VNConfig.NetworkConfiguration.VirtualNetworkConfiguration.VirtualNetworkSites.RemoveChild($targetSite) | Out-Null
    }

    $fileName = "$PSScriptRoot\RemoveCluster_NetCfg_$(Get-Random -Minimum 100000 -Maximum 1000000).xml"
    "Temp NetCfg file is saved to $fileName"
    $VNConfig.Save($fileName)
    "Updating the configuration to Azure"
    Set-AzureVNetConfig -ConfigurationPath $fileName | Out-Null
    if (-not $?)
    {
        $netCfgUpdated = $false
    }
    Remove-Item -Path $fileName
}

if ($storesFailedToRemove.Count -gt 0 -or $agsFailedToRemove.Count -gt 0 -or $netCfgUpdated -eq $false)
{
    Write-Warning "Errors occured while removing the following resources:"
    if ($storesFailedToRemove.Count -gt 0) { Write-Warning "Storage Account: $($storesFailedToRemove -join ',')" }
    if ($agsFailedToRemove.Count -gt 0) { Write-Warning "Affinity Group: $($agsFailedToRemove -join ',')" }
    if (-not $netCfgUpdated) { Write-Warning "VNet and/or DNS settings" }
    Write-Warning "Check whether they are held by other Azure resources (e.g. images) and try again"
}

"Finished removing IaaS cluster"
# SIG # Begin signature block
# MIIavwYJKoZIhvcNAQcCoIIasDCCGqwCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUtqj71wTkjP/pC6SyXrQW6Si0
# 8+CgghV6MIIEuzCCA6OgAwIBAgITMwAAAFwJq3ADEfxcFQAAAAAAXDANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTQwNTIzMTcxMzE2
# WhcNMTUwODIzMTcxMzE2WjCBqzELMAkGA1UEBhMCVVMxCzAJBgNVBAgTAldBMRAw
# DgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24x
# DTALBgNVBAsTBE1PUFIxJzAlBgNVBAsTHm5DaXBoZXIgRFNFIEVTTjpCQkVDLTMw
# Q0EtMkRCRTElMCMGA1UEAxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZTCC
# ASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBANYJa/7h+ToIHrYOs0PLI95O
# bOJZcc75TxmjL8QLy/H1Xyl5ielMpmU8xUtMXQHp/i9QvCCDC5hHW5nqTZ/eqcsj
# lYzk07XcemMgs/L/r8dp/5K7jQpF7punPfdkmj6O6rRaABbNqeXVTJUmQH8DmFyk
# 6dBH1jblvqVRQb9b9uyuNN9K8gXa/fw9YY6yEa11unZrY2JOJSHPpz88ub9uBEMk
# HrvplbnGydEgydhVI1Xnsr+vLVxathcwcTMQJH5xnrFl4ma2BdMyV0nEYhPYkoyB
# kY8zQ5EZbBL0eRTm5Z/F5anhM0C22y9X05U2It3yQqmUtGTTgSMUlMPZRhdm3r0C
# AwEAAaOCAQkwggEFMB0GA1UdDgQWBBR/kUhDg4fJeBzpVcW521b4F9FHBDAfBgNV
# HSMEGDAWgBQjNPjZUkZwCu1A+3b7syuwwzWzDzBUBgNVHR8ETTBLMEmgR6BFhkNo
# dHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNyb3Nv
# ZnRUaW1lU3RhbXBQQ0EuY3JsMFgGCCsGAQUFBwEBBEwwSjBIBggrBgEFBQcwAoY8
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNyb3NvZnRUaW1l
# U3RhbXBQQ0EuY3J0MBMGA1UdJQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEBBQUA
# A4IBAQCCCVb4b+5GzEwp9r7Tx2pf+V2EABkXgYloZBU/wUQ2OxhtBMusE7eOlOMt
# 66P0QnxE0QnsWr7nk9OEcuKP1R3VOVP7ciILZpI7ysqr+s/MZE7GRGyv4IOOFBG4
# sJBJn8OmlM+D4BFiWqtYo7hhlyR2rG1D947qHfOI/ipL7h/0HW2sIEyIAEAutLlI
# 2tHEbsGt9DOFYQuJLF5rygNjUkox7r1VwxCnUaiSrxJgxZxdRmPTzOqfNcatCSB0
# hTUbRRKzllHhye2bgOV25fdWhesgnCiRVPtoulaFg1Vh/4hAXLENJgejxVfsFXkP
# lvSOdqGKZ1qNf05i9H7fEKDWFLqXMIIE7DCCA9SgAwIBAgITMwAAAMps1TISNcTh
# VQABAAAAyjANBgkqhkiG9w0BAQUFADB5MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSMwIQYDVQQDExpNaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QTAeFw0xNDA0MjIxNzM5MDBaFw0xNTA3MjIxNzM5MDBaMIGDMQswCQYDVQQGEwJV
# UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
# ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMQ0wCwYDVQQLEwRNT1BSMR4wHAYDVQQD
# ExVNaWNyb3NvZnQgQ29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAw
# ggEKAoIBAQCWcV3tBkb6hMudW7dGx7DhtBE5A62xFXNgnOuntm4aPD//ZeM08aal
# IV5WmWxY5JKhClzC09xSLwxlmiBhQFMxnGyPIX26+f4TUFJglTpbuVildGFBqZTg
# rSZOTKGXcEknXnxnyk8ecYRGvB1LtuIPxcYnyQfmegqlFwAZTHBFOC2BtFCqxWfR
# +nm8xcyhcpv0JTSY+FTfEjk4Ei+ka6Wafsdi0dzP7T00+LnfNTC67HkyqeGprFVN
# TH9MVsMTC3bxB/nMR6z7iNVSpR4o+j0tz8+EmIZxZRHPhckJRIbhb+ex/KxARKWp
# iyM/gkmd1ZZZUBNZGHP/QwytK9R/MEBnAgMBAAGjggFgMIIBXDATBgNVHSUEDDAK
# BggrBgEFBQcDAzAdBgNVHQ4EFgQUH17iXVCNVoa+SjzPBOinh7XLv4MwUQYDVR0R
# BEowSKRGMEQxDTALBgNVBAsTBE1PUFIxMzAxBgNVBAUTKjMxNTk1K2I0MjE4ZjEz
# LTZmY2EtNDkwZi05YzQ3LTNmYzU1N2RmYzQ0MDAfBgNVHSMEGDAWgBTLEejK0rQW
# WAHJNy4zFha5TJoKHzBWBgNVHR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1pY3Jv
# c29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNDb2RTaWdQQ0FfMDgtMzEtMjAx
# MC5jcmwwWgYIKwYBBQUHAQEETjBMMEoGCCsGAQUFBzAChj5odHRwOi8vd3d3Lm1p
# Y3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY0NvZFNpZ1BDQV8wOC0zMS0yMDEwLmNy
# dDANBgkqhkiG9w0BAQUFAAOCAQEAd1zr15E9zb17g9mFqbBDnXN8F8kP7Tbbx7Us
# G177VAU6g3FAgQmit3EmXtZ9tmw7yapfXQMYKh0nfgfpxWUftc8Nt1THKDhaiOd7
# wRm2VjK64szLk9uvbg9dRPXUsO8b1U7Brw7vIJvy4f4nXejF/2H2GdIoCiKd381w
# gp4YctgjzHosQ+7/6sDg5h2qnpczAFJvB7jTiGzepAY1p8JThmURdwmPNVm52Iao
# AP74MX0s9IwFncDB1XdybOlNWSaD8cKyiFeTNQB8UCu8Wfz+HCk4gtPeUpdFKRhO
# lludul8bo/EnUOoHlehtNA04V9w3KDWVOjic1O1qhV0OIhFeezCCBbwwggOkoAMC
# AQICCmEzJhoAAAAAADEwDQYJKoZIhvcNAQEFBQAwXzETMBEGCgmSJomT8ixkARkW
# A2NvbTEZMBcGCgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9z
# b2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5MB4XDTEwMDgzMTIyMTkzMloX
# DTIwMDgzMTIyMjkzMloweTELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0
# b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3Jh
# dGlvbjEjMCEGA1UEAxMaTWljcm9zb2Z0IENvZGUgU2lnbmluZyBQQ0EwggEiMA0G
# CSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCycllcGTBkvx2aYCAgQpl2U2w+G9Zv
# zMvx6mv+lxYQ4N86dIMaty+gMuz/3sJCTiPVcgDbNVcKicquIEn08GisTUuNpb15
# S3GbRwfa/SXfnXWIz6pzRH/XgdvzvfI2pMlcRdyvrT3gKGiXGqelcnNW8ReU5P01
# lHKg1nZfHndFg4U4FtBzWwW6Z1KNpbJpL9oZC/6SdCnidi9U3RQwWfjSjWL9y8lf
# RjFQuScT5EAwz3IpECgixzdOPaAyPZDNoTgGhVxOVoIoKgUyt0vXT2Pn0i1i8UU9
# 56wIAPZGoZ7RW4wmU+h6qkryRs83PDietHdcpReejcsRj1Y8wawJXwPTAgMBAAGj
# ggFeMIIBWjAPBgNVHRMBAf8EBTADAQH/MB0GA1UdDgQWBBTLEejK0rQWWAHJNy4z
# Fha5TJoKHzALBgNVHQ8EBAMCAYYwEgYJKwYBBAGCNxUBBAUCAwEAATAjBgkrBgEE
# AYI3FQIEFgQU/dExTtMmipXhmGA7qDFvpjy82C0wGQYJKwYBBAGCNxQCBAweCgBT
# AHUAYgBDAEEwHwYDVR0jBBgwFoAUDqyCYEBWJ5flJRP8KuEKU5VZ5KQwUAYDVR0f
# BEkwRzBFoEOgQYY/aHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJv
# ZHVjdHMvbWljcm9zb2Z0cm9vdGNlcnQuY3JsMFQGCCsGAQUFBwEBBEgwRjBEBggr
# BgEFBQcwAoY4aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNy
# b3NvZnRSb290Q2VydC5jcnQwDQYJKoZIhvcNAQEFBQADggIBAFk5Pn8mRq/rb0Cx
# MrVq6w4vbqhJ9+tfde1MOy3XQ60L/svpLTGjI8x8UJiAIV2sPS9MuqKoVpzjcLu4
# tPh5tUly9z7qQX/K4QwXaculnCAt+gtQxFbNLeNK0rxw56gNogOlVuC4iktX8pVC
# nPHz7+7jhh80PLhWmvBTI4UqpIIck+KUBx3y4k74jKHK6BOlkU7IG9KPcpUqcW2b
# Gvgc8FPWZ8wi/1wdzaKMvSeyeWNWRKJRzfnpo1hW3ZsCRUQvX/TartSCMm78pJUT
# 5Otp56miLL7IKxAOZY6Z2/Wi+hImCWU4lPF6H0q70eFW6NB4lhhcyTUWX92THUmO
# Lb6tNEQc7hAVGgBd3TVbIc6YxwnuhQ6MT20OE049fClInHLR82zKwexwo1eSV32U
# jaAbSANa98+jZwp0pTbtLS8XyOZyNxL0b7E8Z4L5UrKNMxZlHg6K3RDeZPRvzkbU
# 0xfpecQEtNP7LN8fip6sCvsTJ0Ct5PnhqX9GuwdgR2VgQE6wQuxO7bN2edgKNAlt
# HIAxH+IOVN3lofvlRxCtZJj/UBYufL8FIXrilUEnacOTj5XJjdibIa4NXJzwoq6G
# aIMMai27dmsAHZat8hZ79haDJLmIz2qoRzEvmtzjcT3XAH5iR9HOiMm4GPoOco3B
# oz2vAkBq/2mbluIQqBC0N1AI1sM9MIIGBzCCA++gAwIBAgIKYRZoNAAAAAAAHDAN
# BgkqhkiG9w0BAQUFADBfMRMwEQYKCZImiZPyLGQBGRYDY29tMRkwFwYKCZImiZPy
# LGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQDEyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZp
# Y2F0ZSBBdXRob3JpdHkwHhcNMDcwNDAzMTI1MzA5WhcNMjEwNDAzMTMwMzA5WjB3
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEwHwYDVQQDExhN
# aWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAw
# ggEKAoIBAQCfoWyx39tIkip8ay4Z4b3i48WZUSNQrc7dGE4kD+7Rp9FMrXQwIBHr
# B9VUlRVJlBtCkq6YXDAm2gBr6Hu97IkHD/cOBJjwicwfyzMkh53y9GccLPx754gd
# 6udOo6HBI1PKjfpFzwnQXq/QsEIEovmmbJNn1yjcRlOwhtDlKEYuJ6yGT1VSDOQD
# LPtqkJAwbofzWTCd+n7Wl7PoIZd++NIT8wi3U21StEWQn0gASkdmEScpZqiX5NMG
# gUqi+YSnEUcUCYKfhO1VeP4Bmh1QCIUAEDBG7bfeI0a7xC1Un68eeEExd8yb3zuD
# k6FhArUdDbH895uyAc4iS1T/+QXDwiALAgMBAAGjggGrMIIBpzAPBgNVHRMBAf8E
# BTADAQH/MB0GA1UdDgQWBBQjNPjZUkZwCu1A+3b7syuwwzWzDzALBgNVHQ8EBAMC
# AYYwEAYJKwYBBAGCNxUBBAMCAQAwgZgGA1UdIwSBkDCBjYAUDqyCYEBWJ5flJRP8
# KuEKU5VZ5KShY6RhMF8xEzARBgoJkiaJk/IsZAEZFgNjb20xGTAXBgoJkiaJk/Is
# ZAEZFgltaWNyb3NvZnQxLTArBgNVBAMTJE1pY3Jvc29mdCBSb290IENlcnRpZmlj
# YXRlIEF1dGhvcml0eYIQea0WoUqgpa1Mc1j0BxMuZTBQBgNVHR8ESTBHMEWgQ6BB
# hj9odHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9taWNy
# b3NvZnRyb290Y2VydC5jcmwwVAYIKwYBBQUHAQEESDBGMEQGCCsGAQUFBzAChjho
# dHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY3Jvc29mdFJvb3RD
# ZXJ0LmNydDATBgNVHSUEDDAKBggrBgEFBQcDCDANBgkqhkiG9w0BAQUFAAOCAgEA
# EJeKw1wDRDbd6bStd9vOeVFNAbEudHFbbQwTq86+e4+4LtQSooxtYrhXAstOIBNQ
# md16QOJXu69YmhzhHQGGrLt48ovQ7DsB7uK+jwoFyI1I4vBTFd1Pq5Lk541q1YDB
# 5pTyBi+FA+mRKiQicPv2/OR4mS4N9wficLwYTp2OawpylbihOZxnLcVRDupiXD8W
# mIsgP+IHGjL5zDFKdjE9K3ILyOpwPf+FChPfwgphjvDXuBfrTot/xTUrXqO/67x9
# C0J71FNyIe4wyrt4ZVxbARcKFA7S2hSY9Ty5ZlizLS/n+YWGzFFW6J1wlGysOUzU
# 9nm/qhh6YinvopspNAZ3GmLJPR5tH4LwC8csu89Ds+X57H2146SodDW4TsVxIxIm
# dgs8UoxxWkZDFLyzs7BNZ8ifQv+AeSGAnhUwZuhCEl4ayJ4iIdBD6Svpu/RIzCzU
# 2DKATCYqSCRfWupW76bemZ3KOm+9gSd0BhHudiG/m4LBJ1S2sWo9iaF2YbRuoROm
# v6pH8BJv/YoybLL+31HIjCPJZr2dHYcSZAI9La9Zj7jkIeW1sMpjtHhUBdRBLlCs
# lLCleKuzoJZ1GtmShxN1Ii8yqAhuoFuMJb+g74TKIdbrHk/Jmu5J4PcBZW+JC33I
# acjmbuqnl84xKf8OxVtc2E0bodj6L54/LlUWa8kTo/0xggSvMIIEqwIBATCBkDB5
# MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVk
# bW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSMwIQYDVQQDExpN
# aWNyb3NvZnQgQ29kZSBTaWduaW5nIFBDQQITMwAAAMps1TISNcThVQABAAAAyjAJ
# BgUrDgMCGgUAoIHIMBkGCSqGSIb3DQEJAzEMBgorBgEEAYI3AgEEMBwGCisGAQQB
# gjcCAQsxDjAMBgorBgEEAYI3AgEVMCMGCSqGSIb3DQEJBDEWBBTWJ4ZQVj4pL+rn
# zEAc6MxR31hQzTBoBgorBgEEAYI3AgEMMVowWKA2gDQATQBpAGMAcgBvAHMAbwBm
# AHQAIABIAFAAQwAgAFAAYQBjAGsAIAAyADAAMQAyACAAUgAyoR6AHGh0dHA6Ly93
# d3cubWljcm9zb2Z0LmNvbS9IUEMwDQYJKoZIhvcNAQEBBQAEggEAblbpRLzHyCOR
# 00xOjePCeoKJNZWB3JLL+jjUD9L2fYAaGDXUVTiRzmYRIk2Qi3U53sEqBhQPjL7W
# wiUdwiKSY3cD7IAY/9/WthjE8zD/IrygsOGyuRq1GhSLUplWHbbnZcgOxEOOL6SZ
# IiIiTEJlIn7fWjKR4eJUqc7DTn5Bn15P4FoaLVAuhkfoj+X5WHJreXBqONY6y0sh
# yYBcLgVFEg7HIuNnKRheEifHJmiX2fWHAlP9TmDaXfSHLpTkM2EWmYwCmTuZ4R3h
# 8HNY6+8abmjS5OG6sskqG4miQXHw6Cxo4Cjqmrn9Tbe9O9syEAYGNk4AQSGl3IpJ
# z7DhLIjQ86GCAigwggIkBgkqhkiG9w0BCQYxggIVMIICEQIBATCBjjB3MQswCQYD
# VQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEe
# MBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEwHwYDVQQDExhNaWNyb3Nv
# ZnQgVGltZS1TdGFtcCBQQ0ECEzMAAABcCatwAxH8XBUAAAAAAFwwCQYFKw4DAhoF
# AKBdMBgGCSqGSIb3DQEJAzELBgkqhkiG9w0BBwEwHAYJKoZIhvcNAQkFMQ8XDTE1
# MDIxMjE4MTQ1NlowIwYJKoZIhvcNAQkEMRYEFNBLVYVpvE68DZpptplZQFCBfBQ1
# MA0GCSqGSIb3DQEBBQUABIIBABXXnNSVtS3wyB287kRjx5ZMnTdl4xl0NSCWLXXR
# eY5Kecu3FvD3ONnK+yMIZPzHq74VnA/b4p+oBL33BhqEo9YHsZoO+yvrRl5RYJ3i
# Kw193/ddIc/Y2QWxsr8Vg71GxmLut2zIOtXQ1LTKLJbTYMQGOsm0uWrkJDzm5hDO
# qXM0Sd9nS9oH+uoce2xnRfbmBrPwst4tUt9HCbsTA2+l42StgvF/ZxFGgnmSoECu
# se1WLJ5W3ckNO0UJacz/+V4lKnTthum4+U7wlyGDnNLi9307VlOEL/4Su/Ag7MMS
# /phITAinRZTjBTP1K1O0BDwQsv7UsAWxy5UUcUbtVehBfLw=
# SIG # End signature block
